/*
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <common.h>
#include <command.h>
#include <rt_mmap.h>
#include <configs/rt2880.h>
#include "ralink_spi.h"

#if (CONFIG_COMMANDS & CFG_CMD_SPI) 
#define DBG

#define	RT2880_SPI_DUMP_STR		"dump"	/* Dump Content Command Prompt    */
#define	RT2880_SPI_READ_STR		"read"	/* SPI read operation */
#define	RT2880_SPI_WRITE_STR		"write"	/* SPI read operation */

#define RT2880_SPI_DUMP        2
#define RT2880_SPI_READ        3
#define RT2880_SPI_WRITE       5
#define RT2880_SPI_RD_VTSS     7
#define RT2880_SPI_WR_VTSS     9


/*---------------------------------------------------------------------*/
/* Symbol & Macro Definitions                                          */
/*---------------------------------------------------------------------*/
#define RT2880_REG(x)		(*((volatile u32 *)(x)))

#define IS_BUSY		(RT2880_REG(RT2880_SPISTAT_REG) & 0x01)

#define spi_busy_loop 3000
#define max_ee_busy_loop 500


/*
 * ATMEL AT25XXXX Serial EEPROM 
 * access type
 */

/* Instruction codes */
#define WREN_CMD	0x06
#define WRDI_CMD	0x04
#define RDSR_CMD	0x05
#define WRSR_CMD	0x01
#define READ_CMD	0x03
#define WRITE_CMD	0x02

/* STATUS REGISTER BIT */
#define RDY 0	/*  Busy Indicator Bit */
#define WEN 1	/*  Write Enable Bit   */
#define BP0 2	/* Block Write Protect Bit */
#define BP1 3	/* Block Write Protect Bit */
#define WPEN 7	/* Software Write Protect Enable Bit */


#define ENABLE	1
#define DISABLE	0


//vlan
#if 0
static unsigned char lutonu [] = {
	0x2, 0xa, 0xfd, 0x8f, 0x2e, 0x12, 0x1e, 0x86, 0x12, 0x8, 0xe1, 0x50, 0x3, 0x7f, 0x1, 0x22, 
	0xe5, 0x2e, 0xb4, 0x6, 0x3, 0x7f, 0x1, 0x22, 0x7f, 00, 0x22, 0x2, 0x18, 0xa1, 0x75, 0x22, 
	0x1, 0x80, 0xfe, 0x2, 0x14, 0x10, 0xae, 0x5, 0xad, 0x7, 0x75, 0x39, 0x6, 0xee, 0x70, 0x4, 
	0x7f, 0x1, 0x80, 0x2, 0x7f, 00, 0x8f, 0x3a, 0x7b, 0x19, 0x7f, 0x20, 0x2, 0x16, 0x52, 0x22, 
	00, 0x1, 0xc1, 00, 00, 00, 0x8f, 0x3f, 0x8d, 0x40, 0x8b, 0x41, 0xe5, 0x41, 0x7b, 0xff, 
	0x60, 0x6, 0x7a, 0xb, 0x79, 0xbe, 0x80, 0x4, 0x7a, 0xb, 0x79, 0xa1, 0x12, 0x18, 0xc6, 0x7b, 
	0xff, 0x7a, 0xb, 0x79, 0xdb, 0x12, 0x18, 0xc6, 0xe5, 0x3f, 0xc3, 0x94, 0x7, 0x40, 0x8, 0xe4, 
	0xf5, 0x3f, 0x75, 0x42, 0x6, 0x80, 0x11, 0xe5, 0x41, 0x60, 0xa, 0x53, 0x3f, 0xfe, 0xe5, 0x3f, 
	0x4, 0xf5, 0x42, 0x80, 0x3, 0x85, 0x3f, 0x42, 0xe5, 0x40, 0xb4, 0xff, 0x5, 0x75, 0x44, 0x7, 
	0x80, 0xf, 0xaf, 0x40, 0x74, 0x1, 0xa8, 0x7, 0x8, 0x80, 0x2, 0xc3, 0x33, 0xd8, 0xfc, 0xf5, 
	0x44, 0x7b, 0xd, 0x12, 0x3, 0x2b, 0x7b, 0xe, 0x12, 0x3, 0x2b, 0x12, 0x3, 0x66, 0x12, 0x1a, 
	0x3c, 0x12, 0x8, 0x65, 0xe5, 0x3f, 0x90, 0x1e, 0x9b, 0x93, 0xf5, 0x4b, 0xe5, 0x41, 0x60, 0x1e, 
	0xe5, 0x3f, 0x24, 0x1, 0xff, 0xe4, 0x33, 0xfe, 0xc3, 0xef, 0x94, 0x7, 0xee, 0x64, 0x80, 0x94, 
	0x80, 0x50, 0x8, 0xe5, 0x3f, 0xa3, 0x93, 0xf5, 0x4c, 0x80, 0x3, 0x75, 0x4c, 0x7, 0x7e, 00, 
	0x7f, 0x2, 0x7d, 0xff, 0x7b, 00, 0x7a, 00, 0x79, 0x45, 0x12, 0x4, 0x60, 0x75, 0x43, 0x2, 
	0xe5, 0x43, 0xf4, 0x70, 0x3, 0x2, 0x2, 0x50, 0x74, 0x1, 0x7e, 00, 0xa8, 0x43, 0x8, 0x80, 
	0x5, 0xc3, 0x33, 0xce, 0x33, 0xce, 0xd8, 0xf9, 0xff, 0xef, 0x55, 0x44, 0x70, 0x3, 0x2, 0x2, 
	0x4b, 0xe5, 0x43, 0xd3, 0x94, 0x2, 0x40, 0x3, 0x2, 0x1, 0xec, 0xaf, 0x4b, 0x12, 0x1d, 0x5d, 
	0x40, 0x3, 0x2, 0x2, 0x4b, 0xaf, 0x4c, 0x12, 0x1d, 0x5d, 0x50, 0x7, 0xe5, 0x4c, 0xc3, 0x94, 
	0x7, 0x40, 0x7, 0xe5, 0x41, 0x60, 0x3, 0x2, 0x2, 0x4b, 0xe5, 0x4b, 0x64, 0x6, 0x70, 0x1b, 
	0xe5, 0x43, 0x70, 0x9, 0x12, 0x3, 0x5d, 0x20, 0xe4, 0x3, 0x2, 0x2, 0x4b, 0xe5, 0x43, 0xb4, 
	0x1, 0x9, 0x12, 0x3, 0x5d, 0x20, 0xe6, 0x3, 0x2, 0x2, 0x4b, 0xe5, 0x41, 0x60, 0x67, 0xe5, 
	0x43, 0xc3, 0x94, 0x2, 0x50, 0x6f, 0xe5, 0x4c, 0x64, 0x6, 0x70, 0x25, 0xe5, 0x43, 0x70, 0xe, 
	0x7d, 0x1, 0xaf, 0x4c, 0x12, 0x11, 0xca, 0xee, 0x20, 0xe4, 0x3, 0x2, 0x2, 0x4b, 0xe5, 0x43, 
	0xb4, 0x1, 0xe, 0x7d, 0x1, 0xaf, 0x4c, 0x12, 0x11, 0xca, 0xee, 0x20, 0xe6, 0x3, 0x2, 0x2, 
	0x4b, 0x12, 0x3, 0x70, 0xe4, 0xf5, 0x49, 0xf5, 0x4a, 0xaf, 0x4b, 0x12, 0x1d, 0x79, 0x50, 0x7, 
	0xaf, 0x4c, 0x12, 0x1d, 0x79, 0x40, 0x2e, 0xe5, 0x4a, 0xc3, 0x94, 0x3, 0x50, 0x27, 0x7f, 0xa, 
	0x12, 0x1c, 0x37, 0x5, 0x49, 0xe5, 0x49, 0xc3, 0x94, 0xfa, 0x40, 0xdd, 0x12, 0x3, 0x70, 0x5, 
	0x4a, 0xe4, 0xf5, 0x49, 0x80, 0xd3, 0xe5, 0x43, 0x44, 0x90, 0xfd, 0xaf, 0x4b, 0x12, 0x12, 0xbf, 
	0x7f, 0xa, 0x12, 0x1d, 0xd9, 0xad, 0x43, 0xaf, 0x4b, 0x12, 0x10, 0x45, 0xe5, 0x41, 0x60, 0x7, 
	0xad, 0x43, 0xaf, 0x4c, 0x12, 0x10, 0x45, 0x7f, 0x5, 0x12, 0x1d, 0xd9, 0xe5, 0x41, 0x60, 0x41, 
	0xad, 0x4c, 0x12, 0x3, 0x4e, 0x80, 0x2, 0xc3, 0x33, 0xd8, 0xfc, 0x12, 0x1a, 0x53, 0x80, 0x2, 
	0xc3, 0x33, 0xd8, 0xfc, 0xf4, 0x55, 0x45, 0x4f, 0xf5, 0x45, 0xad, 0x4b, 0xaf, 0x4c, 0x12, 0xe, 
	0x9d, 0xe5, 0x43, 0x25, 0xe0, 0xfe, 0xef, 0xa8, 0x6, 0x8, 0x80, 0x2, 0xc3, 0x33, 0xd8, 0xfc, 
	0x12, 0x1a, 0x53, 0x80, 0x2, 0xc3, 0x33, 0xd8, 0xfc, 0xf4, 0x55, 0x46, 0x4f, 0xf5, 0x46, 0x80, 
	0x1a, 0xad, 0x4b, 0x12, 0x3, 0x4e, 0x80, 0x2, 0xc3, 0x33, 0xd8, 0xfc, 0x12, 0x1a, 0x53, 0x80, 
	0x2, 0xc3, 0x33, 0xd8, 0xfc, 0xf4, 0x55, 0x45, 0x4f, 0xf5, 0x45, 0x15, 0x43, 0x2, 00, 0xf0, 
	0xe4, 0xf5, 0x48, 0xe5, 0x41, 0x70, 0x7, 0xe5, 0x48, 0xc3, 0x94, 0x1, 0x40, 0x11, 0xe5, 0x41, 
	0x70, 0x3, 0x2, 0x2, 0xfd, 0xe5, 0x48, 0xc3, 0x94, 0x2, 0x40, 0x3, 0x2, 0x2, 0xfd, 0x12, 
	0x1d, 0xf7, 0x7f, 0x4, 0x12, 00, 00, 0xaf, 0x3f, 0x12, 00, 00, 0x7f, 0x4, 0x12, 00, 
	00, 0xe4, 0xf5, 0x43, 0xe5, 0x43, 0x25, 0xe0, 0xff, 0x74, 0x45, 0x25, 0x48, 0xf8, 0xe6, 0xfe, 
	0xa8, 0x7, 0x8, 0x80, 0x2, 0xc3, 0x13, 0xd8, 0xfc, 0x54, 0x3, 0xf5, 0x47, 0xb4, 0x3, 0x7, 
	0x7f, 0x8, 0x12, 00, 00, 0x80, 0x24, 0xe5, 0x47, 0x7b, 0xff, 0x70, 0x9, 0x7a, 0xb, 0x79, 
	0xfd, 0x12, 0x18, 0xc6, 0x80, 0x15, 0x7a, 0xc, 0x79, 0x6, 0x12, 0x18, 0xc6, 0xaf, 0x47, 0xe4, 
	0xfc, 0xfd, 0xfe, 0x12, 00, 00, 0x7f, 0x20, 0x12, 00, 00, 0x5, 0x43, 0xe5, 0x43, 0xc3, 
	0x94, 0x3, 0x40, 0xb0, 0xe5, 0x48, 0x70, 0xb, 0xaf, 0x4b, 0x12, 0x1d, 0x5d, 0x50, 0x10, 0xaf, 
	0x4b, 0x80, 0x9, 0xaf, 0x4c, 0x12, 0x1d, 0x5d, 0x50, 0x5, 0xaf, 0x4c, 0x12, 0x1e, 0x1, 0x5, 
	0x3f, 0xe5, 0x3f, 0xd3, 0x95, 0x42, 0x50, 0x5, 0x5, 0x48, 0x2, 0x2, 0x53, 0xe5, 0x3f, 0xd3, 
	0x95, 0x42, 0x50, 0x3, 0x2, 00, 0xb4, 0x7b, 0xff, 0x7a, 0xc, 0x79, 0xd, 0x12, 0x18, 0xc6, 
	0x7b, 0xd, 0x12, 0x3, 0x3e, 0x12, 0x8, 0x65, 0x7b, 0xe, 0x12, 0x3, 0x3e, 0x12, 0x8, 0x65, 
	0x12, 0x3, 0x66, 0x7f, 0x5f, 0x12, 0x1a, 0x3e, 0x2, 0x8, 0x65, 0xe4, 0xfd, 0x7f, 0x40, 0x12, 
	0x8, 0x5e, 0x7f, 00, 0x7e, 00, 0x7d, 00, 0x7c, 00, 0x12, 0x8, 0x65, 0x22, 0xe4, 0xfd, 
	0x7f, 0x40, 0x12, 0x8, 0x5e, 0x7f, 0x5f, 0x7e, 00, 0x7d, 00, 0x7c, 00, 0x22, 0xaf, 0x4b, 
	0x12, 0xe, 0x9d, 0xe5, 0x43, 0x25, 0xe0, 0xfe, 0xef, 0xa8, 0x6, 0x8, 0x22, 0x7d, 0x1, 0xaf, 
	0x4b, 0x12, 0x11, 0xca, 0xee, 0x22, 0x7b, 0xf, 0xe4, 0xfd, 0x7f, 0x40, 0x12, 0x8, 0x5e, 0x22, 
	0xab, 0x43, 0xad, 0x4c, 0xaf, 0x4b, 0x12, 0x17, 0xbc, 0x22, 0xbb, 0x1, 0x6, 0x89, 0x82, 0x8a, 
	0x83, 0xe0, 0x22, 0x50, 0x2, 0xe7, 0x22, 0xbb, 0xfe, 0x2, 0xe3, 0x22, 0x89, 0x82, 0x8a, 0x83, 
	0xe4, 0x93, 0x22, 0xbb, 0x1, 0x6, 0x89, 0x82, 0x8a, 0x83, 0xf0, 0x22, 0x50, 0x2, 0xf7, 0x22, 
	0xbb, 0xfe, 0x1, 0xf3, 0x22, 0xe8, 0x8f, 0xf0, 0xa4, 0xcc, 0x8b, 0xf0, 0xa4, 0x2c, 0xfc, 0xe9, 
	0x8e, 0xf0, 0xa4, 0x2c, 0xfc, 0x8a, 0xf0, 0xed, 0xa4, 0x2c, 0xfc, 0xea, 0x8e, 0xf0, 0xa4, 0xcd, 
	0xa8, 0xf0, 0x8b, 0xf0, 0xa4, 0x2d, 0xcc, 0x38, 0x25, 0xf0, 0xfd, 0xe9, 0x8f, 0xf0, 0xa4, 0x2c, 
	0xcd, 0x35, 0xf0, 0xfc, 0xeb, 0x8e, 0xf0, 0xa4, 0xfe, 0xa9, 0xf0, 0xeb, 0x8f, 0xf0, 0xa4, 0xcf, 
	0xc5, 0xf0, 0x2e, 0xcd, 0x39, 0xfe, 0xe4, 0x3c, 0xfc, 0xea, 0xa4, 0x2d, 0xce, 0x35, 0xf0, 0xfd, 
	0xe4, 0x3c, 0xfc, 0x22, 0xcf, 0xf4, 0xcf, 0xce, 0xf4, 0xce, 0xcd, 0xf4, 0xcd, 0xcc, 0xf4, 0xcc, 
	0x22, 0xeb, 0x9f, 0xf5, 0xf0, 0xea, 0x9e, 0x42, 0xf0, 0xe9, 0x9d, 0x42, 0xf0, 0xe8, 0x9c, 0x45, 
	0xf0, 0x22, 0xe8, 0x60, 0xf, 0xec, 0xc3, 0x13, 0xfc, 0xed, 0x13, 0xfd, 0xee, 0x13, 0xfe, 0xef, 
	0x13, 0xff, 0xd8, 0xf1, 0x22, 0xe8, 0x60, 0xf, 0xef, 0xc3, 0x33, 0xff, 0xee, 0x33, 0xfe, 0xed, 
	0x33, 0xfd, 0xec, 0x33, 0xfc, 0xd8, 0xf1, 0x22, 0xe6, 0xfc, 0x8, 0xe6, 0xfd, 0x8, 0xe6, 0xfe, 
	0x8, 0xe6, 0xff, 0x22, 0xe4, 0x93, 0xfc, 0x74, 0x1, 0x93, 0xfd, 0x74, 0x2, 0x93, 0xfe, 0x74, 
	0x3, 0x93, 0xff, 0x22, 0xec, 0xf6, 0x8, 0xed, 0xf6, 0x8, 0xee, 0xf6, 0x8, 0xef, 0xf6, 0x22, 
	0xef, 0x4e, 0x60, 0x12, 0xef, 0x60, 0x1, 0xe, 0xed, 0xbb, 0x1, 0xb, 0x89, 0x82, 0x8a, 0x83, 
	0xf0, 0xa3, 0xdf, 0xfc, 0xde, 0xfa, 0x22, 0x89, 0xf0, 0x50, 0x7, 0xf7, 0x9, 0xdf, 0xfc, 0xa9, 
	0xf0, 0x22, 0xbb, 0xfe, 0xfc, 0xf3, 0x9, 0xdf, 0xfc, 0xa9, 0xf0, 0x22, 0xa8, 0x6a, 0xe6, 0xf5, 
	0x24, 0x64, 0x20, 0x60, 0x9, 0xe5, 0x24, 0x64, 0xd, 0x60, 0x3, 0x7f, 0x1, 0x22, 0xe4, 0xf5, 
	0x69, 0xfe, 0x7f, 0x10, 0x7d, 0xff, 0xfb, 0x7a, 00, 0x79, 0x7c, 0x12, 0x4, 0x60, 0x12, 0x1d, 
	0xc3, 0x75, 0x25, 0xa, 0xe4, 0xf5, 0x27, 0xa8, 0x6a, 0xe6, 0x64, 0x30, 0x70, 0x1b, 0xe8, 0x4, 
	0xf8, 0xe6, 0xff, 0x12, 0x1b, 0x31, 0xbf, 0x58, 0x10, 0x75, 0x25, 0x10, 0x5, 0x6a, 0x5, 0x6a, 
	0xa8, 0x6a, 0xe6, 0xb4, 0x20, 0x3, 0x7f, 0x1, 0x22, 0xa8, 0x6a, 0xe6, 0xf5, 0x24, 0x64, 0x20, 
	0x60, 0x6, 0xe5, 0x24, 0x64, 0xd, 0x70, 0x7c, 0xe5, 0x27, 0xd3, 0x94, 00, 0x40, 0x6a, 0xe4, 
	0xf5, 0x35, 0xf5, 0x34, 0xf5, 0x33, 0xf5, 0x32, 0xf5, 0x26, 0xe5, 0x26, 0xc3, 0x95, 0x27, 0x50, 
	0x3a, 0xaf, 0x25, 0xe4, 0xfc, 0xfd, 0xfe, 0xab, 0x35, 0xaa, 0x34, 0xa9, 0x33, 0xa8, 0x32, 0x12, 
	0x3, 0xa5, 0xc0, 0x4, 0xa9, 0x5, 0xaa, 0x6, 0xab, 0x7, 0x74, 0x28, 0x25, 0x26, 0xf8, 0xe6, 
	0xff, 0xe4, 0xfc, 0xfd, 0xfe, 0xd0, 00, 0xeb, 0x2f, 0xf5, 0x35, 0xee, 0x3a, 0xf5, 0x34, 0xed, 
	0x39, 0xf5, 0x33, 0xec, 0x38, 0xf5, 0x32, 0x5, 0x26, 0x80, 0xbf, 0xe5, 0x69, 0xc3, 0x94, 0x4, 
	0x50, 0x17, 0xaf, 0x35, 0xae, 0x34, 0xad, 0x33, 0xac, 0x32, 0xab, 0x69, 0x5, 0x69, 0xeb, 0x25, 
	0xe0, 0x25, 0xe0, 0x24, 0x7c, 0xf8, 0x12, 0x4, 0x54, 0xe5, 0x24, 0x64, 0xd, 0x60, 0x3, 0x2, 
	0x4, 0xae, 0xff, 0x22, 0xaf, 0x24, 0x12, 0x18, 0x7b, 0x8f, 0x24, 0xe5, 0x24, 0xf4, 0x60, 0x1d, 
	0xe5, 0x27, 0xc3, 0x94, 0xa, 0x50, 0x19, 0xaf, 0x27, 0x5, 0x27, 0x74, 0x28, 0x2f, 0xf8, 0xa6, 
	0x24, 0xe5, 0x24, 0xd3, 0x94, 0x9, 0x40, 0x8, 0x75, 0x25, 0x10, 0x80, 0x3, 0x7f, 0x1, 0x22, 
	0x5, 0x6a, 0x2, 0x4, 0xd9, 0x8f, 0x2a, 0x7f, 0x1, 0xed, 0x54, 0x30, 0xfe, 0xbe, 0x30, 0x4, 
	0x7e, 0x1, 0x80, 0x2, 0x7e, 00, 0xed, 0x54, 0x3, 0xfc, 0xbc, 0x2, 0x4, 0x7c, 0x1, 0x80, 
	0x2, 0x7c, 00, 0x8c, 0x2c, 0xed, 0x20, 0xe4, 0x7, 0xef, 0x60, 0x4, 0x7d, 0x1, 0x80, 0x2, 
	0x7d, 00, 0xef, 0x60, 0x3, 0xee, 0x70, 0x3, 0xed, 0x60, 0x4, 0x7f, 0x1, 0x80, 0x2, 0x7f, 
	00, 0x8f, 0x2b, 0x7d, 0x30, 0xaf, 0x2a, 0x12, 0x17, 0x6a, 0x7b, 0x8, 0xad, 0x2a, 0x7f, 0x20, 
	0x12, 0x8, 0x5e, 0xe5, 0x31, 0xfe, 0xe5, 0x32, 0xff, 0xad, 0x30, 0x7c, 00, 0x12, 0x8, 0x7b, 
	0x12, 0x8, 0x65, 0x7b, 0xc, 0xad, 0x2a, 0x7f, 0x20, 0x12, 0x8, 0x5e, 0xe5, 0x34, 0xfe, 0xe5, 
	0x35, 0xff, 0xad, 0x33, 0x7c, 00, 0x12, 0x8, 0x7b, 0x12, 0x8, 0x65, 0xe5, 0x2b, 0x60, 0x1e, 
	0x75, 0x2d, 0x16, 0x75, 0x2e, 0x9c, 0x12, 0x6, 0x81, 0x7f, 0xff, 0x7e, 00, 0x7d, 0x3, 0x7c, 
	00, 0x12, 0x8, 0x65, 0x85, 0x2a, 0x39, 0xe4, 0xf5, 0x3a, 0x7b, 0x12, 0x80, 0x1d, 0x75, 0x2d, 
	0x16, 0x75, 0x2e, 0x84, 0x12, 0x6, 0x81, 0x7f, 00, 0x7e, 00, 0x7d, 00, 0x7c, 00, 0x12, 
	0x8, 0x65, 0x85, 0x2a, 0x39, 0x75, 0x3a, 0x1, 0x7b, 0x12, 0xe4, 0xfd, 0x7f, 0xa0, 0x12, 0x16, 
	0x52, 0xe4, 0xf5, 0x2f, 0xe5, 0x2f, 0x24, 0xde, 0xfb, 0xad, 0x2a, 0x7f, 0x20, 0x12, 0x8, 0x5e, 
	0x85, 0x2e, 0x82, 0x85, 0x2d, 0x83, 0x12, 0x4, 0x44, 0x12, 0x8, 0x65, 0x74, 0x4, 0x25, 0x2e, 
	0xf5, 0x2e, 0xe4, 0x35, 0x2d, 0xf5, 0x2d, 0x5, 0x2f, 0xe5, 0x2f, 0xc3, 0x94, 0x6, 0x40, 0xd4, 
	0x22, 0x7b, 0x4, 0xad, 0x2a, 0x7f, 0x20, 0x12, 0x8, 0x5e, 0x22, 0x8f, 0x2b, 0x8c, 0x2c, 0x8d, 
	0x2d, 0xe5, 0x2d, 0xf4, 0x45, 0x2c, 0x70, 0xa, 0xf5, 0x31, 0xf5, 0x30, 0xf5, 0x2f, 0xf5, 0x2e, 
	0x80, 0x12, 0xe5, 0x2d, 0x54, 0x13, 0xfd, 0xaf, 0x2b, 0x12, 0x11, 0xf, 0x8f, 0x31, 0x8e, 0x30, 
	0x8d, 0x2f, 0x8c, 0x2e, 0xe4, 0xf5, 0x39, 0xf5, 0x38, 0xf5, 0x37, 0x75, 0x36, 0x80, 0xf5, 0x3d, 
	0xf5, 0x3c, 0xf5, 0x3b, 0x75, 0x3a, 0x80, 0x7b, 0xdf, 0xad, 0x2b, 0x7f, 0x20, 0x12, 0xf, 0xdd, 
	0x12, 0x7, 0x69, 0xe5, 0x31, 0x54, 0xf8, 0xff, 0xe5, 0x30, 0xfe, 0xe5, 0x2f, 0xfd, 0xe5, 0x2e, 
	0xfc, 0xef, 0x44, 0x31, 0xff, 0xec, 0x44, 0x20, 0xfc, 0x12, 0x8, 0x65, 0xe5, 0x2d, 0xf4, 0x45, 
	0x2c, 0x70, 0x23, 0xfd, 0xaf, 0x2b, 0x12, 0x1d, 0x94, 0x12, 0x7, 0x69, 0x7f, 0x30, 0x7e, 00, 
	0x7d, 00, 0x7c, 0x20, 0x12, 0x8, 0x65, 0x85, 0x2b, 0x39, 0x75, 0x3a, 0x1, 0x7b, 0x12, 0xe4, 
	0xfd, 0x7f, 0xa0, 0x2, 0x16, 0x52, 0xaf, 0x2b, 0x12, 0x1e, 0x52, 0xe4, 0xfc, 0xfd, 0xfe, 0x78, 
	0x13, 0x12, 0x4, 0x25, 0xec, 0x44, 0x8, 0xfc, 0xe5, 0x31, 0x4f, 0xf5, 0x31, 0xe5, 0x30, 0x4e, 
	0xf5, 0x30, 0xe5, 0x2f, 0x4d, 0xf5, 0x2f, 0xe5, 0x2e, 0x4c, 0xf5, 0x2e, 0xe4, 0xfb, 0xad, 0x2b, 
	0x7f, 0x20, 0x12, 0x8, 0x5e, 0xaf, 0x31, 0xae, 0x30, 0xad, 0x2f, 0xac, 0x2e, 0x12, 0x8, 0x65, 
	0xaf, 0x2b, 0x12, 0x1a, 0x5a, 0xe5, 0x2d, 0x54, 0x13, 0xfd, 0xaf, 0x2b, 0x12, 0x17, 0xe3, 0xe5, 
	0x2b, 0xb4, 0x6, 0x4, 0xff, 0x12, 0x12, 0x73, 0x22, 0xe4, 0xfb, 0xad, 0x2b, 0x7f, 0x20, 0x12, 
	0x8, 0x5e, 0x22, 0x8f, 0x2a, 0xe5, 0x2a, 0xb4, 0x6, 0x5, 0x75, 0x2e, 0x1, 0x80, 0x3, 0xe4, 
	0xf5, 0x2e, 0xe5, 0x2a, 0x70, 0x7, 0xf5, 0x2b, 0x75, 0x2c, 0x2, 0x80, 0x12, 0xe5, 0x2a, 0xb4, 
	0x6, 0x7, 0xe4, 0xf5, 0x2b, 0xf5, 0x2c, 0x80, 0x6, 0x75, 0x2b, 00, 0x75, 0x2c, 0x1, 0xe5, 
	0x2c, 0x45, 0x2b, 0x70, 0x4, 0x7f, 0x1, 0x80, 0x2, 0x7f, 00, 0x8f, 0x2d, 0x75, 0x39, 0xe, 
	0x75, 0x3a, 0x1, 0xe4, 0xfb, 0xad, 0x2a, 0x7f, 0x20, 0x12, 0x16, 0x52, 0x75, 0x39, 0x7, 0xe5, 
	0x2e, 0x70, 0x4, 0x7f, 0x1, 0x80, 0x2, 0x7f, 00, 0x12, 0x8, 0x3e, 0x75, 0x39, 0x8, 0xe5, 
	0x2e, 0x70, 0x4, 0x7f, 0x1, 0x80, 0x2, 0x7f, 00, 0x12, 0x8, 0x3e, 0xe5, 0x2d, 0x70, 0x4, 
	0x7f, 0x1, 0x80, 0x2, 0x7f, 00, 0xef, 0x75, 0xf0, 0x8, 0xa4, 0xfd, 0xe5, 0x2c, 0xc4, 0xf8, 
	0x54, 0xf, 0xc8, 0x68, 0xff, 0xe5, 0x2b, 0xc4, 0x54, 0xf0, 0x48, 0x45, 0xf0, 0xfe, 0xef, 0x4d, 
	0xff, 0xef, 0x45, 0x2e, 0xff, 0xe4, 0x8f, 0x39, 0x8e, 0x38, 0xf5, 0x37, 0xf5, 0x36, 0x75, 0x3d, 
	0xf9, 0x75, 0x3c, 0xff, 0xf5, 0x3b, 0xf5, 0x3a, 0x7b, 0x24, 0xad, 0x2a, 0x7f, 0x20, 0x12, 0xf, 
	0xdd, 0xe5, 0x2d, 0x60, 0x6, 0x75, 0x2b, 00, 0x75, 0x2c, 0x1, 0x7b, 0x7a, 0xad, 0x2a, 0x7f, 
	0x20, 0x12, 0x8, 0x5e, 0xae, 0x2b, 0xaf, 0x2c, 0xe4, 0xfc, 0xfd, 0x2, 0x8, 0x65, 0x8f, 0x3a, 
	0x7b, 0x79, 0xad, 0x2a, 0x7f, 0x20, 0x12, 0x16, 0x52, 0x22, 0xef, 0x2d, 0xf5, 0xf9, 0x8b, 0xfa, 
	0x30, 0xf8, 0xfd, 0xd2, 0xf8, 0xac, 0xff, 0xad, 0xfe, 0xae, 0xfd, 0xaf, 0xfc, 0x22, 0xef, 0x2d, 
	0xf5, 0x73, 0x8b, 0x74, 0x22, 0x8c, 0xff, 0x8d, 0xfe, 0x8e, 0xfd, 0x8f, 0xfc, 0xe5, 0x73, 0xf5, 
	0xf9, 0xe5, 0x74, 0xf5, 0xfb, 0x30, 0xf8, 0xfd, 0xd2, 0xf8, 0x22, 0x22, 0xec, 0xfe, 0xed, 0xff, 
	0x22, 0x1, 0x2, 0x4, 0x8, 0x10, 0x20, 0x40, 0x80, 0xe4, 0xfe, 0xcf, 0x90, 0x8, 0x81, 0xb4, 
	0x8, 00, 0x50, 0x3, 0x93, 0xff, 0x22, 0x94, 0x8, 0x93, 0xfe, 0x22, 0xef, 0x90, 0x8, 0x81, 
	0x93, 0xff, 0x22, 0x90, 0x8, 0x81, 0xef, 0x93, 0xff, 0xf4, 0xfe, 0xbb, 0x1, 0xe, 0x89, 0x82, 
	0x8a, 0x83, 0xe0, 0x5e, 0xbd, 00, 0x2, 0x80, 0x1, 0x4f, 0xf0, 0x22, 0xbb, 00, 0xb, 0xe9, 
	0xf8, 0xe6, 0x5e, 0xbd, 00, 0x2, 0x80, 0x1, 0x4f, 0xf6, 0x22, 0x74, 0x1f, 0xc3, 0x9f, 0x13, 
	0x13, 0x13, 0x54, 0x3, 0x29, 0xf9, 0xe4, 0x3a, 0xfa, 0x74, 0x7, 0x5f, 0xff, 0x12, 0x8, 0xa3, 
	0x22, 0x90, 0x8, 0x81, 0xef, 0x93, 0xff, 0xbb, 0x1, 0x7, 0x89, 0x82, 0x8a, 0x83, 0xe0, 0x80, 
	0x6, 0xbb, 00, 0x8, 0xe9, 0xf8, 0xe6, 0x5f, 0x60, 0x2, 0xd3, 0x22, 0xc3, 0x22, 0x74, 0x1f, 
	0xc3, 0x9f, 0x13, 0x13, 0x13, 0x54, 0x3, 0x29, 0xf9, 0xe4, 0x3a, 0xfa, 0x74, 0x7, 0x5f, 0xff, 
	0x12, 0x8, 0xe1, 0x22, 0xa8, 0x6a, 0xe6, 0xf5, 0x23, 0x5, 0x6a, 0x12, 0x4, 0x8c, 0xef, 0x60, 
	0x3, 0x7f, 0x1, 0x22, 0xaf, 0x23, 0x12, 0x1b, 0x31, 0xef, 0x24, 0xb7, 0x60, 0x59, 0x24, 0xfa, 
	0x60, 0x6f, 0x24, 0xfd, 0x60, 0x10, 0x24, 0xfb, 0x60, 0x28, 0x4, 0x60, 0x3, 0x2, 0x9, 0xcb, 
	0x12, 0x1d, 0xa1, 0x2, 0x9, 0xce, 0x12, 0x1d, 0xed, 0x12, 0x18, 0xfb, 0xc0, 0x7, 0x12, 0x18, 
	0xeb, 0xc0, 0x5, 0x12, 0x18, 0xf3, 0xd0, 0x5, 0xd0, 0x7, 0x12, 0x8, 0x4a, 0x12, 0x1a, 0xff, 
	0x80, 0x3a, 0xe5, 0x69, 0xc3, 0x94, 0x4, 0x40, 0x65, 0x12, 0x18, 0xfb, 0xc0, 0x7, 0x12, 0x18, 
	0xeb, 0xc0, 0x5, 0x12, 0x18, 0xf3, 0xd0, 0x5, 0xd0, 0x7, 0x12, 0x8, 0x5e, 0x78, 0x88, 0x12, 
	0x4, 0x38, 0x12, 0x8, 0x65, 0x80, 0x47, 0x12, 0x1d, 0xed, 0x78, 0x7c, 0x12, 0x4, 0x38, 0xc0, 
	0x7, 0x12, 0x18, 0xeb, 0xd0, 0x7, 0x12, 0x11, 0xca, 0x12, 0x1c, 0xe7, 0x12, 0x1d, 0xf7, 0x80, 
	0x2d, 0xe5, 0x69, 0xc3, 0x94, 0x3, 0x40, 0x26, 0x78, 0x7c, 0x12, 0x4, 0x38, 0xc0, 0x7, 0x12, 
	0x18, 0xeb, 0xc0, 0x5, 0x12, 0x18, 0xf3, 0xaa, 0x6, 0x78, 0x88, 0x12, 0x4, 0x38, 0x8e, 0x2c, 
	0x8f, 0x2d, 0xd0, 0x5, 0xd0, 0x7, 0x12, 0x18, 0x2f, 0x80, 0x3, 0x7f, 0x1, 0x22, 0x7f, 00, 
	0x22, 0x8f, 0x23, 0xe5, 0x23, 0x64, 0x6, 0x70, 0x54, 0x7b, 0x22, 0x7a, 0x1a, 0x7d, 0x17, 0x12, 
	0x16, 0xb4, 0xaf, 0x23, 0x12, 0x16, 0xe4, 0x12, 0xa, 0x5e, 0x7a, 0x2, 0x12, 0xa, 0x54, 0x7b, 
	0xb5, 0x7a, 0x52, 0x7d, 0x1f, 0xaf, 0x23, 0x12, 0x16, 0xb4, 0x7b, 0xf, 0x7a, 00, 0x7d, 0x2, 
	0xaf, 0x23, 0x12, 0x16, 0xb4, 0x7b, 0x2a, 0x7a, 0x47, 0x7d, 0x1, 0xaf, 0x23, 0x12, 0x16, 0xb4, 
	0x7b, 0xa4, 0x7a, 0x8f, 0xe4, 0xfd, 0xaf, 0x23, 0x12, 0x16, 0xb4, 0x12, 0xa, 0x5e, 0x7a, 00, 
	0x12, 0xa, 0x54, 0xe4, 0xfb, 0xfa, 0x7d, 0x1f, 0xaf, 0x23, 0x2, 0x16, 0xb4, 0xaf, 0x23, 0x12, 
	0xe, 0x26, 0xaf, 0x23, 0x12, 0x1b, 0x90, 0x7b, 0x64, 0x7a, 00, 0x7d, 0x1b, 0xaf, 0x23, 0x12, 
	0x16, 0xb4, 0x75, 0x2c, 00, 0x75, 0x2d, 0xf, 0x7b, 0x3, 0x7a, 00, 0x7d, 0x1e, 0xaf, 0x23, 
	0x12, 0x18, 0x2f, 0x22, 0x7b, 0x12, 0x7d, 0x8, 0xaf, 0x23, 0x12, 0x16, 0xb4, 0x22, 0x7b, 0x30, 
	0x7a, 0x2a, 0x7d, 0x1f, 0xaf, 0x23, 0x12, 0x16, 0xb4, 0x22, 0x8f, 0x25, 0x12, 0x1e, 0x39, 0x12, 
	0x19, 0x62, 0x12, 0xa, 0xe9, 0x12, 0x1e, 0x30, 0x7a, 0xaf, 0x12, 0xa, 0xcb, 0x75, 0x2c, 00, 
	0x75, 0x2d, 0xc, 0x7d, 0x11, 0x12, 0xa, 0xdf, 0x7a, 0x8f, 0x12, 0xa, 0xcb, 0x7a, 0xaf, 0x12, 
	0xa, 0xd5, 0x75, 0x2c, 00, 0x75, 0x2d, 0xc, 0x7d, 0x12, 0x12, 0xa, 0xdf, 0x7a, 0x8f, 0x12, 
	0xa, 0xd5, 0x7a, 0xaf, 0x12, 0xa, 0xf3, 0x75, 0x2c, 0x1, 0x75, 0x2d, 0x80, 0x7b, 00, 0x7a, 
	0x1, 0x7d, 0x11, 0xaf, 0x25, 0x12, 0x18, 0x2f, 0x7a, 0x8f, 0x12, 0xa, 0xf3, 0xaf, 0x25, 0x12, 
	0x1e, 0x39, 0x12, 0x19, 0x6d, 0x12, 0xa, 0xe9, 0x2, 0x1e, 0x4a, 0x7b, 0x8a, 0x7d, 0x10, 0xaf, 
	0x25, 0x12, 0x16, 0xb4, 0x22, 0x7b, 0x86, 0x7d, 0x10, 0xaf, 0x25, 0x12, 0x16, 0xb4, 0x22, 0x7b, 
	0x8, 0x7a, 00, 0xaf, 0x25, 0x12, 0x18, 0x2f, 0x22, 0x7d, 0x8, 0xaf, 0x25, 0x12, 0x18, 0x2f, 
	0xaf, 0x25, 0x22, 0x7b, 0x82, 0x7d, 0x10, 0xaf, 0x25, 0x12, 0x16, 0xb4, 0x22, 0x78, 0xfe, 0xe4, 
	0xf6, 0xd8, 0xfd, 0x75, 0x77, 00, 0x75, 0x78, 00, 0x75, 0x81, 0xd1, 0x2, 0xb, 0x4a, 0x2, 
	0xd, 0xa8, 0xe4, 0x93, 0xa3, 0xf8, 0xe4, 0x93, 0xa3, 0x40, 0x3, 0xf6, 0x80, 0x1, 0xf2, 0x8, 
	0xdf, 0xf4, 0x80, 0x29, 0xe4, 0x93, 0xa3, 0xf8, 0x54, 0x7, 0x24, 0xc, 0xc8, 0xc3, 0x33, 0xc4, 
	0x54, 0xf, 0x44, 0x20, 0xc8, 0x83, 0x40, 0x4, 0xf4, 0x56, 0x80, 0x1, 0x46, 0xf6, 0xdf, 0xe4, 
	0x80, 0xb, 0x1, 0x2, 0x4, 0x8, 0x10, 0x20, 0x40, 0x80, 0x90, 0x17, 0x13, 0xe4, 0x7e, 0x1, 
	0x93, 0x60, 0xbc, 0xa3, 0xff, 0x54, 0x3f, 0x30, 0xe5, 0x9, 0x54, 0x1f, 0xfe, 0xe4, 0x93, 0xa3, 
	0x60, 0x1, 0xe, 0xcf, 0x54, 0xc0, 0x25, 0xe0, 0x60, 0xa8, 0x40, 0xb8, 0xe4, 0x93, 0xa3, 0xfa, 
	0xe4, 0x93, 0xa3, 0xf8, 0xe4, 0x93, 0xa3, 0xc8, 0xc5, 0x82, 0xc8, 0xca, 0xc5, 0x83, 0xca, 0xf0, 
	0xa3, 0xc8, 0xc5, 0x82, 0xc8, 0xca, 0xc5, 0x83, 0xca, 0xdf, 0xe9, 0xde, 0xe7, 0x80, 0xbe, 0x49, 
	0x6e, 0x76, 0x61, 0x6c, 0x69, 0x64, 0x20, 0x63, 0x6f, 0x6d, 0x6d, 0x61, 0x6e, 0x64, 0xd, 0xa, 
	00, 0x49, 0x6e, 0x74, 0x65, 0x72, 0x6e, 0x61, 0x6c, 0x20, 0x6c, 0x6f, 0x6f, 0x70, 0x62, 0x61, 
	0x63, 0x6b, 0x20, 0x6f, 0x6e, 0x20, 0x50, 0x48, 0x59, 0x73, 0x3a, 0xd, 0xa, 00, 0x45, 0x78, 
	0x74, 0x65, 0x72, 0x6e, 0x61, 0x6c, 0x20, 0x6c, 0x6f, 0x6f, 0x70, 0x62, 0x61, 0x63, 0x6b, 0x20, 
	0x6f, 0x6e, 0x20, 0x50, 0x48, 0x59, 0x73, 0x3a, 0xd, 0xa, 00, 0xd, 0xa, 0x20, 0x20, 0x20, 
	0x20, 0x50, 0x6f, 0x72, 0x74, 0x20, 0x20, 0x40, 0x31, 0x30, 0x20, 0x20, 0x20, 0x20, 0x20, 0x40, 
	0x31, 0x30, 0x30, 0x20, 0x20, 0x20, 0x20, 0x40, 0x31, 0x30, 0x30, 0x30, 00, 0x4f, 0x6b, 0x20, 
	0x20, 0x20, 0x20, 0x20, 0x20, 00, 0x45, 0x72, 0x72, 0x6f, 0x72, 0x20, 00, 0xd, 0xa, 0x54, 
	0x65, 0x73, 0x74, 0x20, 0x63, 0x6f, 0x6d, 0x70, 0x6c, 0x65, 0x74, 0x65, 0x64, 0xd, 0xa, 00, 
	0x12, 0x15, 0x4, 0x40, 0x3, 0x2, 0xc, 0xa5, 0x75, 0x6a, 0x8c, 0x12, 0x1d, 0xc3, 0x12, 0x19, 
	0x5, 0x75, 0x27, 0x1d, 0x75, 0x28, 0x4f, 0x12, 0x14, 0x4f, 0xef, 0x70, 0x4, 0xf5, 0x22, 0x80, 
	0x4a, 0x12, 0x19, 0x5, 0x75, 0x27, 0x1d, 0x75, 0x28, 0x55, 0x12, 0x14, 0x4f, 0xef, 0x70, 0x36, 
	0x74, 0x7, 0x25, 0x6a, 0xf5, 0x6a, 0x12, 0x4, 0x8c, 0xef, 0x60, 0x2, 0x80, 0x6, 0xe5, 0x69, 
	0x64, 0x2, 0x60, 0x5, 0x75, 0x22, 0x1, 0x80, 0x18, 0x78, 0x7c, 0x12, 0x4, 0x38, 0xc0, 0x6, 
	0xc0, 0x7, 0x12, 0x18, 0xeb, 0xac, 0x6, 0xd0, 0x7, 0xd0, 0x6, 0x12, 0xf, 0x74, 0x12, 0x1c, 
	0x4a, 0xe4, 0xf5, 0x22, 0x80, 0x5, 0x12, 0x9, 0x14, 0x8f, 0x22, 0xe5, 0x22, 0x60, 0x13, 0x12, 
	0x1d, 0xad, 0x50, 0x5, 0x12, 0x1c, 0xd8, 0x80, 0xf6, 0x7b, 0xff, 0x7a, 0xb, 0x79, 0x8f, 0x12, 
	0x18, 0xc6, 0xe4, 0xf5, 0x68, 0x22, 0x12, 00, 0x3f, 0xe4, 0x75, 0x26, 0xc0, 0xf5, 0x25, 0xf5, 
	0x24, 0xf5, 0x23, 0xf5, 0x22, 0x7b, 00, 0x7a, 00, 0x79, 0x23, 0xaf, 0x22, 0x12, 0x8, 0xfe, 
	0x40, 0x21, 0xe4, 0xfb, 0x7d, 0x2, 0x7f, 0x60, 0x12, 0x8, 0x5e, 0xe4, 0x25, 0x22, 0xff, 0xe4, 
	0x34, 0x4, 0xfe, 0xe4, 0x34, 0x1, 0xfd, 0xe4, 0x34, 0x1, 0xfc, 0x12, 0x8, 0x65, 0x7f, 0x2, 
	0x12, 0x1c, 0x37, 0x5, 0x22, 0xe5, 0x22, 0xd3, 0x94, 0xf, 0x40, 0xc9, 0x7f, 0x3, 0x12, 0x1d, 
	0xd9, 0x12, 0x1e, 0x78, 0x12, 0x1e, 0x42, 0x7f, 0x4, 0x12, 0x1d, 0xd9, 0x12, 0x13, 0x8f, 0x75, 
	0x39, 0x1, 0x75, 0x3a, 0x1, 0x7b, 0x14, 0xe4, 0xfd, 0x7f, 0xe0, 0x12, 0x16, 0x52, 0x7f, 0x14, 
	0x12, 0x1c, 0x37, 0x7b, 0x4, 0xe4, 0xfd, 0x7f, 0x40, 0x12, 0x8, 0x5e, 0x7f, 0x5f, 0x12, 0x1e, 
	0x7f, 0x12, 0x8, 0x65, 0x2, 0x13, 0x4d, 0x8f, 0x23, 0x74, 0xca, 0x2f, 0xf8, 0xe6, 0x24, 0xfe, 
	0x60, 0x2a, 0x14, 0x60, 0x5c, 0x24, 0x2, 0x70, 0x6e, 0x7b, 0xe1, 0x7a, 0x5, 0x7d, 0x4, 0xaf, 
	0x23, 0x12, 0x16, 0xb4, 0x7b, 00, 0x7a, 0x6, 0x7d, 0x9, 0xaf, 0x23, 0x12, 0x16, 0xb4, 0xaf, 
	0x23, 0x12, 0x1d, 0x5, 0x74, 0xca, 0x25, 0x23, 0xf8, 0x76, 0x2, 0x22, 0x30, 0x9, 0x48, 0xaf, 
	0x23, 0x12, 0x1d, 0x79, 0x50, 0x41, 0x12, 0x1e, 0x86, 0x7d, 0x1, 0xaf, 0x23, 0x12, 0x8, 0xa3, 
	0xaf, 0x23, 0x12, 0x1b, 0x4a, 0x8f, 0x24, 0xad, 0x24, 0xaf, 0x23, 0x12, 0x1c, 0x6f, 0xad, 0x24, 
	0xaf, 0x23, 0x12, 0x15, 0x7a, 0x74, 0xca, 0x25, 0x23, 0xf8, 0x76, 0x3, 0xaf, 0x23, 0x2, 0x1e, 
	0xad, 0x30, 0x9, 0x13, 0xaf, 0x23, 0x12, 0x1d, 0x79, 0x40, 0xc, 0x74, 0xca, 0x25, 0x23, 0xf8, 
	0x76, 0x2, 0xaf, 0x23, 0x12, 0x1a, 0x77, 0x22, 0x12, 0x1e, 0xb0, 0x12, 0x17, 0x94, 0xd2, 0xa1, 
	0xc2, 0x91, 0xe4, 0xf5, 0x39, 0x75, 0x38, 0x1, 0xf5, 0x37, 0xf5, 0x36, 0xf5, 0x3d, 0x75, 0x3c, 
	0x1f, 0xf5, 0x3b, 0xf5, 0x3a, 0x7b, 0x10, 0xfd, 0x7f, 0xe0, 0x12, 0xf, 0xdd, 0x12, 0x1c, 0x24, 
	0x12, 0x1b, 0x62, 0xd2, 0xaf, 0x7f, 0x14, 0x12, 0x1c, 0x37, 0x12, 0xc, 0xa6, 0x12, 0x1e, 0xb1, 
	0x12, 0x19, 0x98, 0xef, 0x70, 0x3, 0x30, 0x7, 0x3, 0x12, 00, 0x1e, 0x12, 0x19, 0xda, 0x12, 
	0x1b, 0xbe, 0xef, 0x60, 0x3, 0x12, 00, 0x1e, 0x12, 0x16, 0x1e, 0x12, 0x1b, 0x18, 0x30, 0x7, 
	0x5, 0xc2, 0x7, 0x12, 0x1b, 0xd4, 0x12, 0xc, 0x20, 0x30, 00, 0x8, 0xc2, 00, 0x12, 0x1c, 
	0xf6, 0x12, 0x1a, 0xcb, 0x30, 0x1, 0xe7, 0xc2, 0x1, 0x12, 0x1c, 0xc8, 0x12, 0x1b, 0x79, 0xa2, 
	0x81, 0xb3, 0x92, 0x91, 0x80, 0xd8, 0x8f, 0x24, 0x12, 0xa, 0x6a, 0xaf, 0x24, 0x12, 0x11, 0x6e, 
	0x75, 0x2c, 00, 0x75, 0x2d, 0x40, 0x7b, 0x40, 0x7d, 0x12, 0x12, 0xe, 0x93, 0x12, 0x1e, 0x39, 
	0x75, 0x2c, 0xf, 0x75, 0x2d, 0xc0, 0x7b, 0x40, 0x7a, 0x2, 0x7d, 0x16, 0xaf, 0x24, 0x12, 0x18, 
	0x2f, 0x75, 0x2c, 0x60, 0x75, 0x2d, 00, 0x7b, 00, 0x7a, 0x40, 0x7d, 0x14, 0xaf, 0x24, 0x12, 
	0x18, 0x2f, 0x75, 0x2c, 00, 0x75, 0x2d, 0x38, 0x7b, 0x30, 0x7d, 0x18, 0x12, 0xe, 0x93, 0x12, 
	0x1e, 0x27, 0x75, 0x2c, 0xe0, 0x12, 0x19, 0x58, 0xaf, 0x24, 0x12, 0x18, 0x2f, 0xaf, 0x24, 0x12, 
	0x1e, 0x4a, 0x75, 0x2c, 0xe0, 0x75, 0x2d, 0xe, 0x7b, 0x2, 0x7a, 0x80, 0x7d, 0x18, 0xaf, 0x24, 
	0x2, 0x18, 0x2f, 0x7a, 00, 0xaf, 0x24, 0x12, 0x18, 0x2f, 0xaf, 0x24, 0x22, 0x8f, 0x65, 0x8d, 
	0x66, 0x7b, 0x24, 0x7f, 0x20, 0x12, 0x8, 0x5e, 0x7f, 0x6, 0x12, 0x1a, 0x3e, 0x12, 0x8, 0x65, 
	0xe5, 0x65, 0x24, 0x80, 0xfb, 0x75, 0x39, 0x1b, 0x75, 0x3a, 0x1, 0xe4, 0xfd, 0x7f, 0x40, 0x12, 
	0x16, 0x52, 0x7f, 0x2, 0x12, 0x1d, 0xd9, 0xaf, 0x66, 0x12, 0x19, 0x77, 0x7f, 0xa, 0x12, 0x1e, 
	0x6a, 0x12, 0x1d, 0xb8, 0x40, 0x5, 0x12, 0x1e, 0xa2, 0x50, 0xf6, 0x12, 0x1e, 0xa2, 0x40, 0x13, 
	0xad, 0x66, 0xaf, 0x65, 0x12, 0x12, 0x23, 0xef, 0x70, 0x4, 0xf5, 0x67, 0x80, 0x8, 0x75, 0x67, 
	0x1, 0x80, 0x3, 0x75, 0x67, 0x2, 0xe5, 0x65, 0x24, 0x80, 0xfb, 0x75, 0x39, 0x1b, 0xe4, 0xf5, 
	0x3a, 0xfd, 0x7f, 0x40, 0x12, 0x16, 0x52, 0xaf, 0x67, 0x22, 0x8f, 0x5c, 0xc3, 0xed, 0x94, 0x40, 
	0xec, 0x94, 00, 0x50, 0x16, 0xc3, 0x74, 0x40, 0x9d, 0xff, 0xe4, 0x9c, 0xfe, 0xef, 0x78, 0x2, 
	0xce, 0xc3, 0x13, 0xce, 0x13, 0xd8, 0xf9, 0xf5, 0x5d, 0x80, 0x1c, 0xed, 0x24, 0x3, 0xff, 0xe4, 
	0x3c, 0xfe, 0xef, 0x78, 0x2, 0xce, 0xc3, 0x13, 0xce, 0x13, 0xd8, 0xf9, 0x30, 0xe0, 0x5, 0x75, 
	0x5d, 0x1, 0x80, 0x3, 0xe4, 0xf5, 0x5d, 0xc3, 0xed, 0x94, 0x40, 0xec, 0x94, 00, 0x40, 0xe, 
	0xae, 0x4, 0xaf, 0x5, 0xe4, 0xfc, 0xfd, 0x78, 0x10, 0x12, 0x4, 0x25, 0x80, 0x8, 0x7f, 00, 
	0x7e, 00, 0x7d, 0x40, 0x7c, 00, 0x12, 0x14, 0x8c, 0x7f, 0x20, 0x7e, 0x5, 0x7d, 00, 0x7c, 
	00, 0x2, 0x14, 0x8c, 0x8e, 0x23, 0x8f, 0x24, 0x8c, 0x25, 0x8d, 0x26, 0x7f, 0xa, 0x12, 0x1c, 
	0x10, 0x12, 0x1e, 0xb, 0x40, 0x52, 0x12, 0x1d, 0x23, 0xe4, 0xff, 0x12, 0x1c, 0x93, 0x7f, 0x2, 
	0x12, 0x19, 0x32, 0xe5, 0x23, 0xff, 0x12, 0x19, 0x32, 0xaf, 0x24, 0x12, 0x19, 0x32, 0xe4, 0xf5, 
	0x27, 0x12, 0x1e, 0xb, 0x40, 0x24, 0x12, 0x1d, 0xad, 0x50, 0xfb, 0x12, 0x1c, 0xd8, 0x12, 0x19, 
	0x32, 0x5, 0x24, 0xe5, 0x24, 0x70, 0x2, 0x5, 0x23, 0xe5, 0x26, 0x15, 0x26, 0x70, 0x2, 0x15, 
	0x25, 0x5, 0x27, 0xe5, 0x27, 0xc3, 0x94, 0x20, 0x40, 0xd7, 0x7f, 0x1, 0x12, 0x1c, 0x93, 0x12, 
	0x1b, 0xa7, 0xef, 0x30, 0xe0, 0xab, 0x80, 0xf7, 0xe4, 0xff, 0x2, 0x1c, 0x10, 0x8f, 0x33, 0x8d, 
	0x34, 0x8b, 0x35, 0xe5, 0x39, 0x55, 0x3d, 0xf5, 0x39, 0xe5, 0x38, 0x55, 0x3c, 0xf5, 0x38, 0xe5, 
	0x37, 0x55, 0x3b, 0xf5, 0x37, 0xe5, 0x36, 0x55, 0x3a, 0xf5, 0x36, 0x12, 0x8, 0x4a, 0xa8, 0x4, 
	0xa9, 0x5, 0xaa, 0x6, 0xab, 0x7, 0xaf, 0x3d, 0xae, 0x3c, 0xad, 0x3b, 0xac, 0x3a, 0x12, 0x3, 
	0xf4, 0xeb, 0x5f, 0xff, 0xea, 0x5e, 0xfe, 0xe9, 0x5d, 0xfd, 0xe8, 0x5c, 0xfc, 0xe5, 0x39, 0x4f, 
	0xf5, 0x39, 0xe5, 0x38, 0x4e, 0xf5, 0x38, 0xe5, 0x37, 0x4d, 0xf5, 0x37, 0xe5, 0x36, 0x4c, 0xf5, 
	0x36, 0xab, 0x35, 0xad, 0x34, 0xaf, 0x33, 0x12, 0x8, 0x5e, 0xaf, 0x39, 0xae, 0x38, 0xad, 0x37, 
	0xac, 0x36, 0x2, 0x8, 0x65, 0x8f, 0x8, 0x8d, 0x9, 0x7b, 0x19, 0xad, 0x8, 0x7f, 0x20, 0x12, 
	0x8, 0x5e, 0x12, 0x1a, 0x3c, 0x12, 0x8, 0x65, 0xe5, 0x9, 0x44, 0x10, 0xfd, 0xaf, 0x8, 0x12, 
	0x11, 0xf, 0xef, 0xf5, 0xd, 0xee, 0xf5, 0xc, 0xed, 0x44, 0x1, 0xf5, 0xb, 0xec, 0x44, 0x10, 
	0xf5, 0xa, 0x12, 0x10, 0xa0, 0xe5, 0xd, 0x44, 0x30, 0xff, 0xe5, 0xc, 0xfe, 0xe5, 0xb, 0xfd, 
	0xe5, 0xa, 0x44, 0x20, 0xfc, 0x12, 0x8, 0x65, 0x12, 0x10, 0xa0, 0xaf, 0xd, 0xae, 0xc, 0xad, 
	0xb, 0xac, 0xa, 0x12, 0x8, 0x65, 0xe5, 0x8, 0xb4, 0x6, 0x4, 0xff, 0x12, 0x12, 0x73, 0x22, 
	0xe4, 0xfb, 0xad, 0x8, 0x7f, 0x20, 0x12, 0x8, 0x5e, 0x22, 0x7f, 0xa, 0x12, 0x1c, 0x10, 0xe4, 
	0xff, 0xfe, 0x12, 0x18, 0x55, 0xef, 0x7f, 00, 0xfe, 0xc0, 0x7, 0xc0, 0x6, 0xf, 0x7e, 00, 
	0x12, 0x18, 0x55, 0xef, 0xfd, 0xd0, 0xe0, 0xf5, 0x6b, 0xd0, 0xe0, 0x4d, 0xf5, 0x6c, 0x24, 0x1, 
	0xff, 0xe4, 0x35, 0x6b, 0xfe, 0x12, 0x18, 0x55, 0x8f, 0x6d, 0xe4, 0xff, 0x12, 0x1c, 0x10, 0xe5, 
	0x6c, 0x15, 0x6c, 0x70, 0x2, 0x15, 0x6b, 0xe4, 0xff, 0xfe, 0xc3, 0xef, 0x95, 0x6c, 0xee, 0x95, 
	0x6b, 0x50, 0x11, 0x8f, 0x82, 0x8e, 0x83, 0xe4, 0x93, 0x25, 0x6d, 0xf5, 0x6d, 0xf, 0xbf, 00, 
	0x1, 0xe, 0x80, 0xe6, 0xe5, 0x6d, 0x70, 0x4, 0x7f, 0x1, 0x80, 0x2, 0x7f, 00, 0x22, 0xef, 
	0x64, 0x6, 0xef, 0x64, 0x6, 0x70, 0x24, 0xbd, 0x10, 0x8, 0x7f, 0x43, 0x7e, 0x4, 0x7d, 0x4, 
	0xfc, 0x22, 0xed, 0x70, 0x4, 0x7f, 0x43, 0x80, 0x1e, 0xbd, 0x11, 0x4, 0x7f, 0x42, 0x80, 0x22, 
	0xbd, 0x1, 0x4, 0x7f, 0x42, 0x80, 0x27, 0x7f, 0x81, 0x80, 0x2c, 0xbd, 0x10, 0x4, 0x7f, 0x44, 
	0x80, 0x10, 0xed, 0x70, 0x8, 0x7f, 0x44, 0x7e, 0x4, 0xfd, 0x7c, 0x80, 0x22, 0xbd, 0x11, 0x9, 
	0x7f, 0x44, 0x7e, 0x4, 0x7d, 0x4, 0x7c, 00, 0x22, 0xbd, 0x1, 0x9, 0x7f, 0x44, 0x7e, 0x4, 
	0x7d, 00, 0x7c, 0x80, 0x22, 0x7f, 0x84, 0x7e, 0x1, 0x7d, 0x6, 0x7c, 00, 0x22, 0x8f, 0x25, 
	0x12, 0x1e, 0x39, 0x12, 0x19, 0x62, 0x12, 0x11, 0xb6, 0x12, 0x1e, 0x30, 0x7a, 0xb6, 0x12, 0x11, 
	0xc0, 0x75, 0x2c, 0xff, 0x75, 0x2d, 0x7, 0x7b, 0x3, 0x7a, 00, 0x7d, 0x12, 0xaf, 0x25, 0x12, 
	0x18, 0x2f, 0x75, 0x2c, 00, 0x75, 0x2d, 0xff, 0x7b, 0xa2, 0x7a, 00, 0x7d, 0x11, 0xaf, 0x25, 
	0x12, 0x18, 0x2f, 0x7a, 0x96, 0x12, 0x11, 0xc0, 0xaf, 0x25, 0x12, 0x1e, 0x39, 0x12, 0x19, 0x6d, 
	0x12, 0x11, 0xb6, 0x2, 0x1e, 0x4a, 0x7d, 0x8, 0xaf, 0x25, 0x12, 0x18, 0x2f, 0xaf, 0x25, 0x22, 
	0x7b, 0x8a, 0x7d, 0x10, 0xaf, 0x25, 0x12, 0x16, 0xb4, 0x22, 0x8f, 0x2e, 0x8d, 0x2f, 0x12, 0x1e, 
	0x8d, 0x8f, 0x34, 0xaf, 0x2e, 0x12, 0x1e, 0x94, 0x8f, 0x35, 0x7b, 0x1, 0xad, 0x34, 0x7f, 0x60, 
	0x12, 0x8, 0x5e, 0xe5, 0x35, 0x75, 0xf0, 0x20, 0xa4, 0xff, 0xe5, 0xf0, 0x44, 0x4, 0xfc, 0xef, 
	0x45, 0x2f, 0xfd, 0xe4, 0xff, 0xfe, 0x12, 0x8, 0x7b, 0x12, 0x8, 0x65, 0xaf, 0x34, 0x12, 0x1a, 
	0x1c, 0x7b, 0x2, 0xad, 0x34, 0x7f, 0x60, 0x12, 0x8, 0x4a, 0x8f, 0x33, 0x8e, 0x32, 0x8d, 0x31, 
	0x8c, 0x30, 0xe5, 0x31, 0x54, 0x1, 0xfd, 0xed, 0x70, 0x5, 0xae, 0x32, 0xaf, 0x33, 0x22, 0xe4, 
	0xfe, 0xff, 0x22, 0x8f, 0x18, 0x8d, 0x19, 0xe4, 0xf5, 0x1b, 0xfb, 0x7a, 00, 0x79, 0x1c, 0x12, 
	0x13, 0xd0, 0xef, 0x64, 0x80, 0x4e, 0x60, 0x3, 0x75, 0x1b, 0x1, 0xe5, 0x1c, 0x65, 0x18, 0x60, 
	0x3, 0x75, 0x1b, 0x1, 0xe5, 0x1b, 0x70, 0x25, 0xf5, 0x1a, 0x12, 0x1b, 0xe8, 0xa8, 0x4, 0xa9, 
	0x5, 0xaa, 0x6, 0xab, 0x7, 0xe4, 0x25, 0x19, 0x12, 0x1a, 0x45, 0xc3, 0x12, 0x4, 0x1, 0x60, 
	0x3, 0x75, 0x1b, 0x1, 0x5, 0x1a, 0xe5, 0x1a, 0xc3, 0x94, 0x1f, 0x40, 0xdd, 0x12, 0x1b, 0xfc, 
	0xaf, 0x1b, 0x22, 0x8f, 0x32, 0x75, 0x39, 0x5, 0x12, 0x12, 0xb2, 0xe5, 0x32, 0x64, 0x6, 0x60, 
	0x7, 0xe5, 0x32, 0xb4, 0x6, 0x8, 0x80, 0x6, 0x75, 0x39, 0x3, 0x12, 0x12, 0xb2, 0xe5, 0x32, 
	0xb4, 0x6, 0xe, 0x75, 0x39, 0x2, 0x75, 0x3a, 0x1, 0x7b, 0x19, 0xfd, 0x7f, 0x20, 0x12, 0x16, 
	0x52, 0x7b, 0x5, 0xe4, 0xfd, 0x7f, 0xe0, 0x12, 0x8, 0x5e, 0x7f, 0x20, 0x12, 0x1e, 0x7f, 0x2, 
	0x8, 0x65, 0x75, 0x3a, 0x1, 0x7b, 0x19, 0xad, 0x32, 0x7f, 0x20, 0x12, 0x16, 0x52, 0x22, 0x8f, 
	0x5e, 0x8d, 0x5f, 0xe4, 0xfb, 0xfa, 0xfd, 0x12, 0x16, 0xb4, 0xe5, 0x5f, 0x54, 0x3, 0xff, 0x70, 
	0x6, 0xf5, 0x60, 0xf5, 0x61, 0x80, 0x14, 0xbf, 0x1, 0x8, 0x75, 0x60, 0x20, 0x75, 0x61, 00, 
	0x80, 0x9, 0xbf, 0x2, 0x6, 0x75, 0x60, 00, 0x75, 0x61, 0x40, 0xe5, 0x5f, 0x30, 0xe4, 0x3, 
	0x43, 0x60, 0x1, 0xe5, 0x5f, 0x30, 0xe7, 0x3, 0x43, 0x60, 0x40, 0xab, 0x61, 0xaa, 0x60, 0xe4, 
	0xfd, 0xaf, 0x5e, 0x2, 0x16, 0xb4, 0x8f, 0x27, 0x12, 0x1e, 0x30, 0x7b, 0xf0, 0x7a, 0xaf, 0x7d, 
	0x10, 0xaf, 0x27, 0x12, 0x16, 0xb4, 0x7d, 0x11, 0xaf, 0x27, 0x12, 0x11, 0xca, 0xee, 0xc4, 0xf8, 
	0x54, 0xf0, 0xc8, 0xef, 0xc4, 0x54, 0xf, 0x48, 0x54, 0x1f, 0xff, 0xc3, 0x94, 0x10, 0x40, 0x6, 
	0xef, 0xd3, 0x94, 0x14, 0x40, 0x12, 0xaf, 0x27, 0x12, 0x1e, 0x39, 0x75, 0x2c, 0x3, 0x12, 0x19, 
	0x70, 0x7d, 0xc, 0xaf, 0x27, 0x12, 0x18, 0x2f, 0xaf, 0x27, 0x2, 0x1e, 0x4a, 0x75, 0x27, 0x1, 
	0x7b, 0x6, 0xe4, 0xfd, 0x7f, 0x40, 0x12, 0x8, 0x5e, 0x7f, 0x80, 0x7e, 0x1, 0x7d, 00, 0x7c, 
	00, 0x12, 0x8, 0x65, 0x7b, 0x7, 0xe4, 0xfd, 0x7f, 0x40, 0x12, 0x8, 0x5e, 0xe4, 0x25, 0x27, 
	0xff, 0xe4, 0x33, 0xfe, 0xe4, 0x33, 0xfd, 0xe4, 0x34, 0xc2, 0xfc, 0x12, 0x8, 0x65, 0x7f, 0x1, 
	0x7e, 0x26, 0x12, 0x15, 0xe9, 0x5, 0x27, 0xe5, 0x27, 0xd3, 0x94, 0xf, 0x40, 0xc2, 0x22, 0x75, 
	0x29, 0x41, 0x75, 0x27, 00, 0x75, 0x28, 0x2, 0x12, 0x13, 0xad, 0x12, 0x17, 0x3f, 0x75, 0x29, 
	0x5e, 0x75, 0x27, 00, 0x75, 0x28, 0x1, 0x12, 0x13, 0xad, 0x2, 0x17, 0x3f, 0x7b, 0xe0, 0xe4, 
	0xfd, 0x7f, 0x40, 0x12, 0x8, 0x5e, 0xae, 0x27, 0xaf, 0x28, 0xe4, 0xfc, 0xfd, 0x12, 0x8, 0x65, 
	0xaf, 0x29, 0xe4, 0xfc, 0xfd, 0xfe, 0x78, 0x2, 0x12, 0x4, 0x25, 0xef, 0x44, 0x2, 0xff, 0x22, 
	0x8b, 0x4d, 0x8a, 0x4e, 0x89, 0x4f, 0xe4, 0xf5, 0x5a, 0xf5, 0x5b, 0x12, 0x14, 0x4, 0xe5, 0x50, 
	0x54, 0x3f, 0xf5, 0x54, 0x85, 0x51, 0x55, 0xe5, 0x4e, 0x45, 0x4f, 0x45, 0x4d, 0x60, 0x10, 0xe5, 
	0x53, 0x54, 0x7, 0xab, 0x4d, 0xaa, 0x4e, 0xa9, 0x4f, 0x12, 0x3, 0x93, 0x12, 0x14, 0x4, 0xae, 
	0x54, 0xaf, 0x55, 0x22, 0x12, 0x1b, 0xe8, 0x8f, 0x53, 0x8e, 0x52, 0x8d, 0x51, 0x8c, 0x50, 0x22, 
	0xc0, 0xe0, 0xc0, 0xd0, 0x75, 0xd0, 0x10, 0x30, 0x98, 0x29, 0xaf, 0x99, 0xc2, 0x98, 0xe5, 0xe, 
	0x24, 0x1, 0xfd, 0xe4, 0x33, 0xfc, 0xe5, 0xf, 0x7a, 00, 0xb5, 0x15, 0x8, 0xea, 0xb5, 0x14, 
	0x4, 0xd2, 0x5, 0x80, 0xe, 0x74, 0xaa, 0x25, 0xe, 0xf8, 0xa6, 0x17, 0xe5, 0xe, 0x4, 0x54, 
	0x1f, 0xf5, 0xe, 0x30, 0x99, 0x4, 0xc2, 0x99, 0xd2, 0x6, 0xd0, 0xd0, 0xd0, 0xe0, 0x32, 0x8b, 
	0x23, 0x8a, 0x24, 0x89, 0x25, 0xab, 0x26, 0xaa, 0x27, 0xa9, 0x28, 0x12, 0x3, 0x7a, 0xf5, 0x29, 
	0x70, 0x2, 0xff, 0x22, 0x74, 0x1, 0x25, 0x28, 0xf5, 0x28, 0xe4, 0x35, 0x27, 0xf5, 0x27, 0xab, 
	0x23, 0x5, 0x25, 0xe5, 0x25, 0xaa, 0x24, 0x70, 0x2, 0x5, 0x24, 0x14, 0xf9, 0x12, 0x3, 0x7a, 
	0xff, 0x12, 0x1b, 0x31, 0xef, 0x65, 0x29, 0x60, 0xcc, 0x7f, 0x1, 0x22, 0x8f, 0x59, 0x8e, 0x58, 
	0x8d, 0x57, 0x8c, 0x56, 0x20, 0x8, 0x30, 0x7b, 0xc0, 0xad, 0x5c, 0x7f, 0x20, 0x12, 0x8, 0x5e, 
	0xaf, 0x59, 0xae, 0x58, 0xad, 0x57, 0xac, 0x56, 0x12, 0x8, 0x65, 0xe4, 0x78, 0xd1, 0xf6, 0x7b, 
	0xc8, 0xad, 0x5c, 0x7f, 0x20, 0x12, 0x8, 0x4a, 0xee, 0x30, 0xe0, 0xb, 0x78, 0xd1, 0x6, 0xe6, 
	0xd3, 0x94, 0x64, 0x40, 0xea, 0xd2, 0x8, 0x22, 0x12, 0x18, 0x9, 0x8f, 0x2b, 0xe4, 0xf5, 0x2a, 
	0xaf, 0x2a, 0x12, 0x8, 0x9c, 0xef, 0xf4, 0xff, 0xef, 0x55, 0x2b, 0xff, 0xe4, 0x8f, 0x39, 0xf5, 
	0x38, 0xf5, 0x37, 0xf5, 0x36, 0xe5, 0x2a, 0x24, 0x80, 0xfb, 0xe4, 0x75, 0x3d, 0x5f, 0xf5, 0x3c, 
	0xf5, 0x3b, 0xf5, 0x3a, 0xfd, 0x7f, 0x40, 0x12, 0xf, 0xdd, 0x5, 0x2a, 0xe5, 0x2a, 0xc3, 0x94, 
	0x7, 0x40, 0xcd, 0x22, 0x12, 0x1d, 0xad, 0x50, 0x35, 0x12, 0x1c, 0xd8, 0xef, 0x64, 0xa, 0x60, 
	0x2d, 0xbf, 0x8, 0xb, 0xe5, 0x68, 0xd3, 0x94, 00, 0x40, 0x23, 0x15, 0x68, 0x80, 0x1f, 0xe5, 
	0x68, 0xc3, 0x94, 0x1e, 0x50, 0xa, 0xae, 0x68, 0x5, 0x68, 0x74, 0x8c, 0x2e, 0xf8, 0xa6, 0x7, 
	0xbf, 0xd, 0xb, 0xe5, 0x68, 0xb4, 0x1e, 0x4, 0x78, 0xa9, 0x76, 0xd, 0xd3, 0x22, 0xc3, 0x22, 
	0x8f, 0x28, 0x8d, 0x29, 0xe5, 0x29, 0xf4, 0x60, 0x2e, 0x7b, 0x10, 0xad, 0x28, 0x7f, 0x20, 0x12, 
	0x8, 0x5e, 0x7f, 00, 0x7e, 0x6, 0x7d, 00, 0x7c, 00, 0x12, 0x8, 0x65, 0xad, 0x29, 0xaf, 
	0x28, 0x12, 0x5, 0x95, 0xe4, 0xfd, 0xaf, 0x28, 0x12, 00, 0x26, 0x7d, 0x1, 0xaf, 0x28, 0x12, 
	0x1d, 0x94, 0xaf, 0x28, 0x12, 0x7, 0x73, 0x2, 0x14, 0xc8, 0x8f, 0x25, 0x8d, 0x26, 0xe5, 0x25, 
	0x64, 0x6, 0x60, 0x2d, 0xe5, 0x26, 0x54, 0x3, 0xff, 0xbf, 0x2, 0x5, 0xaf, 0x25, 0x12, 0x13, 
	0x6, 0xe5, 0x26, 0x54, 0x3, 0x70, 0x1a, 0xaf, 0x25, 0x12, 0x1e, 0x39, 0x12, 0x19, 0x55, 0xaf, 
	0x25, 0x12, 0x18, 0x2f, 0x12, 0x19, 0x55, 0xaf, 0x25, 0x12, 0x18, 0x2f, 0xaf, 0x25, 0x12, 0x1e, 
	0x4a, 0x22, 0xaf, 0x5d, 0x15, 0x5d, 0xef, 0xd3, 0x94, 00, 0x40, 0xd, 0x7f, 0x55, 0x7e, 0x55, 
	0x7d, 0x55, 0x7c, 0x55, 0x12, 0x14, 0x8c, 0x80, 0xe9, 0x30, 0x8, 0x9, 0x75, 0x39, 0x1, 0x12, 
	0x15, 0xdc, 0xc2, 0x8, 0x22, 0xe4, 0xf5, 0x39, 0x12, 0x15, 0xdc, 0x22, 0x75, 0x3a, 0x1, 0x7b, 
	0xc4, 0xad, 0x5c, 0x7f, 0x20, 0x12, 0x16, 0x52, 0x22, 0x8e, 0x28, 0x8f, 0x29, 0x7b, 0xb0, 0xe4, 
	0xfd, 0x7f, 0x40, 0x12, 0x8, 0x5e, 0xae, 0x28, 0xaf, 0x29, 0xe4, 0xfc, 0xfd, 0x12, 0x8, 0x65, 
	0x7f, 0xa, 0x12, 0x1e, 0x6a, 0x7b, 0xb0, 0xe4, 0xfd, 0x7f, 0x40, 0x12, 0x8, 0x4a, 0xef, 0x54, 
	0x7, 0x60, 0xa, 0x12, 0x1e, 0xa2, 0x50, 0xed, 0x7f, 0x4, 0x12, 0x1e, 0xaa, 0x22, 0xe4, 0xf5, 
	0x22, 0xe5, 0x22, 0x24, 0x80, 0xfb, 0xe4, 0xfd, 0x7f, 0x40, 0x12, 0x8, 0x5e, 0x7f, 00, 0x12, 
	0x1e, 0x7f, 0x12, 0x8, 0x65, 0xe5, 0x22, 0xb4, 0x6, 0x5, 0x7d, 0x32, 0xff, 0x80, 0x4, 0x7d, 
	0xff, 0xaf, 0x22, 0x12, 0x1c, 0x6f, 0x5, 0x22, 0xe5, 0x22, 0xc3, 0x94, 0x7, 0x40, 0xd2, 0x2, 
	0x14, 0xc8, 0x8f, 0x36, 0x8d, 0x37, 0x8b, 0x38, 0x12, 0x8, 0x4a, 0x8f, 0x3e, 0x8e, 0x3d, 0x8d, 
	0x3c, 0x8c, 0x3b, 0x7b, 00, 0x7a, 00, 0x79, 0x3b, 0xad, 0x3a, 0xaf, 0x39, 0x12, 0x8, 0xcb, 
	0xab, 0x38, 0xad, 0x37, 0xaf, 0x36, 0x12, 0x8, 0x5e, 0xaf, 0x3e, 0xae, 0x3d, 0xad, 0x3c, 0xac, 
	0x3b, 0x2, 0x8, 0x65, 00, 00, 0x10, 0x10, 0x80, 00, 00, 00, 00, 0xe, 00, 0x15, 
	00, 0x10, 00, 0x15, 00, 0x12, 00, 0x15, 00, 0x14, 00, 0x15, 00, 00, 0x10, 0xc, 
	0x80, 00, 00, 0x1, 00, 0xa, 00, 0x18, 00, 0xb, 00, 0x18, 00, 0xc, 00, 0x18, 
	00, 0xd, 00, 0x18, 0x8f, 0x2e, 0x8d, 0x2f, 0x8a, 0x30, 0x8b, 0x31, 0x12, 0x1e, 0x8d, 0x8f, 
	0x32, 0xaf, 0x2e, 0x12, 0x1e, 0x94, 0x8f, 0x33, 0x7b, 0x1, 0xad, 0x32, 0x7f, 0x60, 0x12, 0x8, 
	0x5e, 0xab, 0x33, 0xad, 0x2f, 0xaf, 0x31, 0xae, 0x30, 0x12, 0x1d, 0xe3, 0x12, 0x8, 0x65, 0xaf, 
	0x32, 0x2, 0x1a, 0x1c, 0x8f, 0x24, 0x7b, 00, 0x7a, 0x80, 0xe4, 0xfd, 0x12, 0x16, 0xb4, 0x7f, 
	0x2, 0x12, 0x1c, 0x37, 0xe4, 0xf5, 0x25, 0xe4, 0xfd, 0xaf, 0x24, 0x12, 0x11, 0xca, 0xee, 0x30, 
	0xe7, 0x10, 0x5, 0x25, 0xe5, 0x25, 0xd3, 0x94, 0xc8, 0x50, 0x7, 0x7f, 0x1, 0x12, 0x1c, 0x37, 
	0x80, 0xe5, 0x22, 0xc1, 0x5, 0xc1, 0x4, 0x1, 0xe, 00, 0x1, 0xf, 00, 0xc1, 00, 0xc1, 
	0x1, 0x1, 0x1f, 00, 0xc1, 0x2, 0xc1, 0x3, 0x1, 0x1d, 0xa, 0x1, 0x1e, 0x64, 0xc1, 0x7, 
	0x1, 0x79, 00, 0xc1, 0x8, 0x1, 0x68, 00, 0x1, 0x75, 00, 0x1, 0x7b, 00, 00, 0x8f, 
	0x2d, 0x8e, 0x2c, 0x8d, 0x2b, 0x8c, 0x2a, 0x7b, 0xd0, 0xe4, 0xfd, 0x7f, 0x40, 0x12, 0x8, 0x5e, 
	0xaf, 0x2d, 0xae, 0x2c, 0xad, 0x2b, 0xac, 0x2a, 0x12, 0x8, 0x65, 0x7b, 0xd0, 0xe4, 0xfd, 0x7f, 
	0x40, 0x12, 0x8, 0x4a, 0xef, 0x54, 0x3, 0x70, 0xf2, 0x22, 0xe4, 0xfe, 0xee, 0x90, 00, 0x40, 
	0x93, 0xfc, 0xed, 0x2e, 0xf8, 0xa6, 0x4, 0xe, 0xbe, 0x6, 0xf1, 0xed, 0x24, 0x5, 0xf8, 0xe6, 
	0x2f, 0xf6, 0xc3, 0x9f, 0x50, 0xd, 0xed, 0x24, 0x4, 0xf8, 0x6, 0xe6, 0x70, 0x5, 0xed, 0x24, 
	0x3, 0xf8, 0x6, 0x22, 0x7b, 0x17, 0xe4, 0xfd, 0x7f, 0xe0, 0x12, 0x8, 0x5e, 0x7f, 0xff, 0x7e, 
	0xff, 0x7d, 0xff, 0x7c, 0xff, 0x12, 0x8, 0x65, 0x7b, 0x16, 0xe4, 0xfd, 0x7f, 0xe0, 0x12, 0x8, 
	0x5e, 0x7f, 0x90, 0x7e, 0x22, 0x7d, 00, 0x7c, 00, 0x2, 0x8, 0x65, 0x8f, 0x62, 0x8d, 0x63, 
	0x8b, 0x64, 0x12, 0x1e, 0x1, 0xaf, 0x63, 0x12, 0x1e, 0x1, 0x7f, 0x5, 0x12, 0x1d, 0xd9, 0xe5, 
	0x64, 0x44, 0x10, 0xfd, 0xaf, 0x62, 0x12, 0x12, 0xbf, 0xe5, 0x64, 0x44, 0x10, 0xfd, 0xaf, 0x63, 
	0x2, 0x12, 0xbf, 0x8f, 0x32, 0xed, 0x70, 0x2, 0x80, 0x1, 0xe4, 0x75, 0x39, 0x86, 0x75, 0x38, 
	0x2, 0xf5, 0x37, 0xf5, 0x36, 0x75, 0x3d, 0xff, 0x75, 0x3c, 0xf, 0xf5, 0x3b, 0xf5, 0x3a, 0x7b, 
	0x2, 0xad, 0x32, 0x7f, 0x20, 0x12, 0xf, 0xdd, 0x22, 0xe4, 0xf5, 0x2c, 0xf5, 0x2d, 0xaf, 0x2d, 
	0x12, 00, 0x3, 0xef, 0x60, 0xd, 0x7b, 00, 0x7a, 00, 0x79, 0x2c, 0x7d, 0x1, 0xaf, 0x2d, 
	0x12, 0x8, 0xa3, 0x5, 0x2d, 0xe5, 0x2d, 0xc3, 0x94, 0x7, 0x40, 0xe2, 0xaf, 0x2c, 0x22, 0x8f, 
	0x28, 0x8d, 0x29, 0x8a, 0x2a, 0x8b, 0x2b, 0x12, 0x11, 0xca, 0xe5, 0x2d, 0xf4, 0xfd, 0xe5, 0x2c, 
	0xf4, 0x5e, 0xfe, 0xef, 0x5d, 0x42, 0x2b, 0xee, 0x42, 0x2a, 0xab, 0x2b, 0xaa, 0x2a, 0xad, 0x29, 
	0xaf, 0x28, 0x2, 0x16, 0xb4, 0x8e, 0x6e, 0x8f, 0x6f, 0xe4, 0xff, 0x12, 0x1c, 0x93, 0x7f, 0x3, 
	0x12, 0x19, 0x32, 0xe5, 0x6e, 0xff, 0x12, 0x19, 0x32, 0xaf, 0x6f, 0x12, 0x19, 0x32, 0x12, 0x19, 
	0xfb, 0x8f, 0x70, 0x7f, 0x1, 0x12, 0x1c, 0x93, 0xaf, 0x70, 0x22, 0x12, 0x1b, 0x31, 0x74, 0x30, 
	0xd3, 0x9f, 0x50, 0xa, 0xef, 0x94, 0x39, 0x50, 0x5, 0xef, 0x24, 0xd0, 0xff, 0x22, 0x74, 0x41, 
	0xd3, 0x9f, 0x50, 0xa, 0xef, 0x94, 0x46, 0x50, 0x5, 0xef, 0x24, 0xc9, 0xff, 0x22, 0x7f, 0xff, 
	0x22, 0xc2, 0x8e, 0x75, 0x8b, 0x92, 0x75, 0x8d, 0xe6, 0xd2, 0x8e, 0xd2, 0x3, 0xd5, 0x1d, 0x15, 
	0x75, 0x1d, 0xa, 0xd2, 00, 0xd5, 0x1e, 0x5, 0x75, 0x1e, 0x64, 0xd2, 0x1, 0x30, 0x2, 0x5, 
	0xd5, 0x1f, 0x2, 0xc2, 0x2, 0x32, 0x8b, 0x24, 0x8a, 0x25, 0x89, 0x26, 0xab, 0x24, 0xaa, 0x25, 
	0xa9, 0x26, 0x12, 0x3, 0x7a, 0x60, 0x13, 0x5, 0x26, 0xe5, 0x26, 0x70, 0x2, 0x5, 0x25, 0x14, 
	0xf9, 0x12, 0x3, 0x7a, 0xff, 0x12, 0x1c, 0x5d, 0x80, 0xe2, 0x22, 0x78, 0x80, 0x12, 0x4, 0x38, 
	0xad, 0x7, 0x22, 0x78, 0x84, 0x12, 0x4, 0x38, 0xab, 0x7, 0x22, 0x78, 0x7c, 0x12, 0x4, 0x38, 
	0x78, 0x5, 0x2, 0x4, 0x25, 0xa9, 0x6a, 0x7a, 00, 0x7b, 00, 0x75, 0x26, 0xff, 0x22, 0x7b, 
	0xe0, 0xe4, 0xfd, 0x7f, 0x40, 0x12, 0x8, 0x5e, 0x7f, 00, 0x12, 0x1e, 0x17, 0x12, 0x8, 0x65, 
	0x12, 0x1e, 0x15, 0x12, 0x17, 0x3f, 0x12, 0x1e, 0x15, 0x12, 0x17, 0x3f, 0x12, 0x1e, 0x15, 0x2, 
	0x17, 0x3f, 0x8f, 0x29, 0xe4, 0xf5, 0x2a, 0xe5, 0x29, 0x54, 0x80, 0xff, 0x12, 0x1c, 0xa5, 0xe5, 
	0x29, 0x25, 0xe0, 0xf5, 0x29, 0x12, 0x1e, 0x5a, 0x12, 0x1e, 0x62, 0x5, 0x2a, 0xe5, 0x2a, 0xc3, 
	0x94, 0x8, 0x40, 0xe3, 0x22, 0x75, 0x2c, 0x60, 0x75, 0x2d, 00, 0x7b, 00, 0x7a, 0x20, 0x7d, 
	0x14, 0x22, 0x75, 0x2c, 0x2, 0x75, 0x2d, 00, 0x7b, 00, 0x7a, 0x2, 0x22, 0x75, 0x2c, 0x2, 
	0x75, 0x2d, 00, 0xe4, 0xfb, 0xfa, 0x22, 0x8f, 0x71, 0x7d, 0x80, 0x7c, 00, 0x12, 0xf, 0xa, 
	0xe4, 0xf5, 0x72, 0xe4, 0x25, 0x71, 0x12, 0x1a, 0x45, 0x12, 0x14, 0x8c, 0x5, 0x72, 0xe5, 0x72, 
	0xc3, 0x94, 0x20, 0x40, 0xee, 0x2, 0x15, 0xb2, 0x7b, 0x18, 0xe4, 0xfd, 0x7f, 0xe0, 0x12, 0x8, 
	0x4a, 0xec, 0x54, 0xf, 0xfc, 0x7b, 0xe9, 0x7a, 0x50, 0x79, 0x39, 0x78, 0x7, 0xc3, 0x12, 0x4, 
	0x1, 0x60, 0x3, 0x7f, 0x1, 0x22, 0x7f, 00, 0x22, 0x8f, 0x2c, 0x7f, 0x32, 0x12, 0x1e, 0x6a, 
	0x7b, 0xc, 0xe4, 0xfd, 0x7f, 0xa0, 0x12, 0x8, 0x4a, 0xe5, 0x2c, 0x5f, 0xb5, 0x2c, 0x3, 0x7f, 
	0x1, 0x22, 0x12, 0x1e, 0xa2, 0x50, 0xe9, 0x7f, 00, 0x22, 0xe4, 0xf5, 0x76, 0x7f, 0x3, 0x12, 
	0x1d, 0xd9, 0xe4, 0xf5, 0x22, 0xaf, 0x22, 0x12, 0x1d, 0x5d, 0x50, 0x5, 0xaf, 0x22, 0x12, 0x9, 
	0xd1, 0x5, 0x22, 0xe5, 0x22, 0xc3, 0x94, 0x7, 0x40, 0xeb, 0x22, 0xe4, 0xf5, 0x29, 0xe5, 0x2a, 
	0x25, 0xe0, 0xf5, 0x2a, 0x12, 0x1e, 0x5a, 0x12, 0x1c, 0xb7, 0xef, 0x42, 0x2a, 0x12, 0x1e, 0x62, 
	0x5, 0x29, 0xe5, 0x29, 0xc3, 0x94, 0x8, 0x40, 0xe5, 0xaf, 0x2a, 0x22, 0x8f, 0x36, 0x7f, 0x5, 
	0x12, 0x1e, 0x6a, 0xe4, 0xfb, 0xad, 0x36, 0x7f, 0x60, 0x12, 0x8, 0x4a, 0xef, 0x54, 0xb, 0x60, 
	0xa, 0x12, 0x1e, 0xa2, 0x50, 0xed, 0x7f, 0x2, 0x12, 0x1e, 0xaa, 0x22, 0x7f, 00, 0x7e, 00, 
	0x7d, 00, 0x7c, 00, 0x22, 0xff, 0xe4, 0x34, 0x56, 0xfe, 0xe4, 0x34, 0x34, 0xfd, 0xe4, 0x34, 
	0x12, 0xfc, 0x22, 0xff, 0x74, 0x3, 0xa8, 0x6, 0x8, 0x22, 0xad, 0x7, 0xe4, 0xf5, 0x39, 0xf5, 
	0x38, 0x75, 0x37, 0x1, 0x75, 0x36, 0x10, 0xf5, 0x3d, 0xf5, 0x3c, 0x75, 0x3b, 0x1, 0x75, 0x3a, 
	0x18, 0xfb, 0x7f, 0x20, 0x2, 0xf, 0xdd, 0x8f, 0x25, 0x12, 0x1e, 0x86, 0xe4, 0xfd, 0x12, 0x8, 
	0xa3, 0xaf, 0x25, 0x12, 0x1d, 0x5d, 0x50, 0x5, 0xaf, 0x25, 0x12, 0x1e, 0x1e, 0x7d, 0xff, 0xaf, 
	0x25, 0x2, 0x1c, 0x6f, 0xbf, 0x6, 0x3, 0x7f, 0x12, 0x22, 0x7d, 0x1c, 0x12, 0x11, 0xca, 0xef, 
	0x13, 0x13, 0x13, 0x54, 0x3, 0xfe, 0xef, 0x30, 0xe5, 0x3, 0x43, 0x6, 0x10, 0xaf, 0x6, 0x22, 
	0x8f, 0x2b, 0x12, 0x1d, 0x41, 0x7d, 0x1, 0xaf, 0x2b, 0x12, 0x1d, 0x87, 0xaf, 0x2b, 0x12, 0x8, 
	0x9c, 0x12, 0x19, 0xb9, 0xe4, 0xfd, 0xaf, 0x2b, 0x2, 0x1d, 0x87, 0xe4, 0xf5, 0x22, 0xaf, 0x22, 
	0x12, 0x1d, 0x5d, 0x50, 0x5, 0xaf, 0x22, 0x12, 0xd, 0x27, 0x5, 0x22, 0xe5, 0x22, 0xc3, 0x94, 
	0x7, 0x40, 0xeb, 0xc2, 0x9, 0x22, 0x8f, 0x2a, 0xe5, 0x2a, 0xc4, 0x54, 0xf, 0xff, 0x12, 0x1d, 
	0x32, 0x12, 0x1c, 0x5d, 0xe5, 0x2a, 0x54, 0xf, 0xff, 0x12, 0x1d, 0x32, 0x2, 0x1c, 0x5d, 0x8f, 
	0x27, 0x8e, 0x26, 0x8d, 0x25, 0x8c, 0x24, 0x78, 0x10, 0x12, 0x4, 0x12, 0x12, 0x1c, 0xe7, 0xe5, 
	0x27, 0xff, 0xe5, 0x26, 0xfe, 0x2, 0x1c, 0xe7, 0xe4, 0xf5, 0x22, 0xaf, 0x22, 0x12, 0x1d, 0xce, 
	0x74, 0xca, 0x25, 0x22, 0xf8, 0xa6, 0x7, 0x5, 0x22, 0xe5, 0x22, 0xc3, 0x94, 0x7, 0x40, 0xeb, 
	0x22, 0xd3, 0xef, 0x64, 0x80, 0xf8, 0x74, 0x61, 0x64, 0x80, 0x98, 0x50, 0xc, 0xef, 0x64, 0x80, 
	0x94, 0xfa, 0x50, 0x5, 0xef, 0x24, 0xe0, 0xff, 0x22, 0x22, 0x8f, 0x25, 0x12, 0x1a, 0x94, 0x8f, 
	0x26, 0x7d, 0x5, 0xaf, 0x25, 0x12, 0x11, 0xca, 0xee, 0x30, 0xe2, 0x3, 0x43, 0x26, 0x20, 0xaf, 
	0x26, 0x22, 0x75, 0xcb, 0xff, 0x75, 0xca, 0x2, 0x75, 0xcd, 0xff, 0x75, 0xcc, 0x2, 0x75, 0xc8, 
	0x34, 0x75, 0x98, 0x52, 0xd2, 0xbc, 0xd2, 0xac, 0x22, 0x12, 0x19, 0x98, 0xef, 0x60, 0x5, 0x7f, 
	0x5, 0x12, 0x1e, 0xaa, 0x12, 0x1b, 0xbe, 0xef, 0x60, 0x5, 0x7f, 0x6, 0x12, 0x1e, 0xaa, 0x22, 
	0x8f, 0x26, 0x12, 0x1e, 0x39, 0x75, 0x2c, 0x3, 0x12, 0x19, 0x65, 0x7d, 0xc, 0xaf, 0x26, 0x12, 
	0x18, 0x2f, 0xaf, 0x26, 0x2, 0x1e, 0x4a, 0xe4, 0xff, 0x12, 0x1c, 0x93, 0x7f, 0x5, 0x12, 0x19, 
	0x32, 0x12, 0x19, 0xfb, 0x8f, 0x28, 0x7f, 0x1, 0x12, 0x1c, 0x93, 0xaf, 0x28, 0x22, 0xe4, 0xf5, 
	0x23, 0xf5, 0x22, 0xaf, 0x22, 0x12, 0x1d, 0x5d, 0x5, 0x22, 0xe5, 0x22, 0xc3, 0x94, 0x7, 0x40, 
	0xf2, 0xaf, 0x23, 0x22, 0xe4, 0xf5, 0xff, 0xf5, 0xfe, 0xf5, 0xfd, 0x75, 0xfc, 0x1, 0x75, 0xf9, 
	0xe0, 0x75, 0xfb, 0x14, 0xd2, 0xf8, 0x80, 0xfe, 0xe5, 0x5a, 0xfd, 0xe5, 0x5b, 0xfb, 0x7f, 0x80, 
	0x12, 0x8, 0x4a, 0x5, 0x5b, 0xe5, 0x5b, 0x70, 0x2, 0x5, 0x5a, 0x22, 0xe4, 0xfb, 0x7d, 0x4, 
	0x7f, 0x80, 0x12, 0x8, 0x5e, 0x7f, 00, 0x7e, 00, 0x7d, 00, 0x7c, 00, 0x2, 0x8, 0x65, 
	0x8f, 0x7b, 0x7b, 0x35, 0xe4, 0xfd, 0x7f, 0xe0, 0x12, 0x8, 0x5e, 0xaf, 0x7b, 0xe4, 0xfc, 0xfd, 
	0xfe, 0x2, 0x8, 0x65, 0x53, 0x89, 0xf, 0x43, 0x89, 0x10, 0x75, 0x8b, 0x92, 0x75, 0x8d, 0xe6, 
	0xc2, 0xbb, 0xd2, 0xab, 0xd2, 0x8e, 0x22, 0xc2, 0x3, 0xae, 0x7, 0x1f, 0xee, 0xd3, 0x94, 00, 
	0x40, 0x7, 0x30, 0x3, 0xfd, 0xc2, 0x3, 0x80, 0xf0, 0x22, 0x7b, 0x14, 0xe4, 0xfd, 0x7f, 0xe0, 
	0x12, 0x8, 0x5e, 0x7f, 0x1, 0x12, 0x1e, 0x7f, 0x12, 0x8, 0x65, 0x80, 0xfe, 0xc2, 0x6, 0x8f, 
	0x99, 0x7f, 0x2, 0x12, 0x1e, 0x6a, 0x20, 0x6, 0x5, 0x12, 0x1e, 0xa2, 0x50, 0xf8, 0x22, 0x8f, 
	0x26, 0x8d, 0x27, 0xe5, 0x27, 0x7c, 00, 0x12, 0x1c, 0x81, 0xad, 0x27, 0xaf, 0x26, 0x2, 0x15, 
	0x40, 0x8f, 0x28, 0x8c, 0x29, 0x8d, 0x2a, 0x12, 0x1a, 0xb0, 0xad, 0x2a, 0xac, 0x29, 0xaf, 0x28, 
	0x2, 0x6, 0x8b, 0xef, 0x60, 0x6, 0xe5, 0x7b, 0x44, 0x2, 0x80, 0x4, 0xe5, 0x7b, 0x54, 0xfd, 
	0xff, 0x12, 0x1c, 0x10, 0x22, 0xef, 0x60, 0x6, 0xe5, 0x7b, 0x44, 0x1, 0x80, 0x4, 0xe5, 0x7b, 
	0x54, 0xfe, 0xff, 0x12, 0x1c, 0x10, 0x22, 0x7b, 0x35, 0xe4, 0xfd, 0x7f, 0xe0, 0x12, 0x8, 0x4a, 
	0xef, 0x54, 0x10, 0xc4, 0x54, 0xf, 0xff, 0x22, 0x5, 0x79, 0xe5, 0x79, 0xc3, 0x94, 0x96, 0x40, 
	0x6, 0xe4, 0xf5, 0x79, 0x12, 0x1e, 0x71, 0x22, 0x74, 0xaa, 0x25, 0xf, 0xf8, 0xe6, 0xff, 0xe5, 
	0xf, 0x4, 0x54, 0x1f, 0xf5, 0xf, 0x22, 0x8e, 0x28, 0x8f, 0x29, 0xe5, 0x28, 0xff, 0x12, 0x1a, 
	0xe6, 0xaf, 0x29, 0x2, 0x1a, 0xe6, 0x5, 0x75, 0xe5, 0x75, 0xc3, 0x94, 0xa, 0x40, 0x5, 0xe4, 
	0xf5, 0x75, 0xd2, 0x9, 0x22, 0x75, 0x2c, 0x12, 0x75, 0x2d, 00, 0x7b, 00, 0x7a, 0x12, 0xe4, 
	0xfd, 0x2, 0x18, 0x2f, 0x4c, 0x75, 0x74, 0x6f, 0x6e, 0x20, 0x32, 0x2e, 0x32, 0x39, 00, 00, 
	00, 00, 00, 0xe4, 0xff, 0x12, 0x1c, 0x93, 0x7f, 0x6, 0x12, 0x19, 0x32, 0x7f, 0x1, 0x2, 
	0x1c, 0x93, 0xef, 0xc3, 0x94, 0xa, 0xef, 0x50, 0x4, 0x24, 0x30, 0xff, 0x22, 0x24, 0x57, 0xff, 
	0x22, 0xad, 0x7, 0x75, 0x39, 0x10, 0xe4, 0xf5, 0x3a, 0xfb, 0x7f, 0x20, 0x2, 0x16, 0x52, 0x45, 
	0x52, 0x41, 0x53, 0x45, 00, 0x50, 0x52, 0x4f, 0x47, 0x52, 0x41, 0x4d, 00, 0xef, 0x90, 0x1d, 
	0x72, 0x93, 0xd3, 0x94, 0x1, 0x50, 0x2, 0x80, 0x1, 0xc3, 0x22, 00, 0x1, 0x2, 0x3, 0x4, 
	0xff, 0xff, 00, 00, 00, 00, 00, 0xff, 0xff, 0x7d, 0x1, 0x12, 0x11, 0xca, 0xef, 0x30, 
	0xe2, 0x3, 0xd3, 0x80, 0x1, 0xc3, 0x22, 0x8f, 0x39, 0x8d, 0x3a, 0x7b, 0xe, 0xe4, 0xfd, 0x7f, 
	0xa0, 0x2, 0x16, 0x52, 0x8f, 0x39, 0x8d, 0x3a, 0x7b, 0x10, 0xe4, 0xfd, 0x7f, 0x40, 0x2, 0x16, 
	0x52, 0x7b, 0xff, 0x7a, 0x1d, 0x79, 0x14, 0x12, 0x18, 0xc6, 0x2, 0x1d, 0xf7, 0xe5, 0xe, 0x65, 
	0xf, 0x60, 0x3, 0xd3, 0x80, 0x1, 0xc3, 0x22, 0xe5, 0xaa, 0x54, 0x3, 0x60, 0x3, 0xd3, 0x80, 
	0x1, 0xc3, 0x22, 0xa8, 0x6a, 0xe6, 0xb4, 0x20, 0x4, 0x5, 0x6a, 0x80, 0xf6, 0x22, 0x12, 0x1d, 
	0x5d, 0x50, 0x3, 0x7f, 0x1, 0x22, 0x7f, 00, 0x22, 0xc2, 0x2, 0x8f, 0x1f, 0xd2, 0x2, 0x20, 
	0x2, 0xfd, 0x22, 0xeb, 0x75, 0xf0, 0x20, 0xa4, 0x4d, 0xac, 0xf0, 0xfd, 0x22, 0x7f, 0x30, 0x12, 
	0x1c, 0x5d, 0x7f, 0x78, 0x2, 0x1c, 0x5d, 0x7f, 0xd, 0x12, 0x1c, 0x5d, 0x7f, 0xa, 0x2, 0x1c, 
	0x5d, 0x8f, 0x7a, 0x12, 0x16, 0xe4, 0xaf, 0x7a, 0x2, 0x9, 0xd1, 0xd3, 0xe5, 0x26, 0x94, 00, 
	0xe5, 0x25, 0x94, 00, 0x22, 0x7f, 0x2, 0x7e, 00, 0x7d, 00, 0x7c, 00, 0x22, 0xef, 0x64, 
	0x6, 0x60, 0x3, 0x12, 0x1b, 0x90, 0x22, 0x7b, 0x1, 0x7a, 00, 0x7d, 0x1f, 0x2, 0x16, 0xb4, 
	0x7b, 0xb5, 0x7a, 0x52, 0x7d, 0x1f, 0x2, 0x16, 0xb4, 0x7b, 0x30, 0x7a, 0x2a, 0x7d, 0x1f, 0x2, 
	0x16, 0xb4, 0x7f, 0x3, 0x12, 0x1e, 0x17, 0x2, 0x17, 0x3f, 0xe4, 0xfb, 0xfa, 0x7d, 0x1f, 0x2, 
	0x16, 0xb4, 0x7d, 0x32, 0x12, 0x17, 0x6a, 0xaf, 0x37, 0x22, 0xe5, 0x7b, 0x44, 0x4, 0xff, 0x2, 
	0x1c, 0x10, 0xe5, 0x7b, 0x54, 0xfb, 0xff, 0x2, 0x1c, 0x10, 0xc2, 0x2, 0x8f, 0x1f, 0xd2, 0x2, 
	0x22, 0x7f, 0x3, 0x7e, 00, 0x2, 0x15, 0xe9, 0x7f, 0x5, 0x7e, 00, 0x2, 0x15, 0xe9, 0x7e, 
	00, 0x7d, 00, 0x7c, 00, 0x22, 0x7b, 00, 0x7a, 00, 0x79, 0x76, 0x22, 0xef, 0x90, 0x1d, 
	0x72, 0x93, 0xff, 0x22, 0xef, 0x90, 0x1d, 0x6b, 0x93, 0xff, 0x22, 00, 0x1, 0x2, 0x3, 0x4, 
	0x6, 0x5, 0xa2, 0x2, 0xb3, 0x22, 0x12, 0x10, 0xaa, 0x22, 0xd2, 0x7, 0x22, 0x2, 0x14, 0xc8, 
	0x22, 0x22, 
};
#endif

//no vlan
static unsigned char lutonu [] = {
	0x2, 0xc, 0xc, 0x8f, 0x2e, 0x12, 0x1f, 0xc1, 0x12, 0xa, 0xad, 0x50, 0x3, 0x7f, 0x1, 0x22, 
	0xe5, 0x2e, 0xb4, 0x6, 0x3, 0x7f, 0x1, 0x22, 0x7f, 00, 0x22, 0x2, 0x19, 0xc4, 0x75, 0x22, 
	0x1, 0x80, 0xfe, 0x2, 0x15, 0x2c, 0xae, 0x5, 0xad, 0x7, 0x75, 0x3c, 0x6, 0xee, 0x70, 0x4, 
	0x7f, 0x1, 0x80, 0x2, 0x7f, 00, 0x8f, 0x3d, 0x7b, 0x19, 0x7f, 0x20, 0x2, 0x17, 0x9d, 0x22, 
	00, 0x1, 0xc1, 00, 00, 00, 0x8f, 0x24, 0x8d, 0x25, 0x8b, 0x26, 0xe5, 0x26, 0x7b, 0xff, 
	0x60, 0x6, 0x7a, 0x9, 0x79, 0xb4, 0x80, 0x4, 0x7a, 0x9, 0x79, 0x97, 0x12, 0x19, 0xe9, 0x7b, 
	0xff, 0x7a, 0x9, 0x79, 0xd1, 0x12, 0x19, 0xe9, 0xe5, 0x24, 0xc3, 0x94, 0x7, 0x40, 0x8, 0xe4, 
	0xf5, 0x24, 0x75, 0x27, 0x6, 0x80, 0x11, 0xe5, 0x26, 0x60, 0xa, 0x53, 0x24, 0xfe, 0xe5, 0x24, 
	0x4, 0xf5, 0x27, 0x80, 0x3, 0x85, 0x24, 0x27, 0xe5, 0x25, 0xb4, 0xff, 0x5, 0x75, 0x29, 0x7, 
	0x80, 0xf, 0xaf, 0x25, 0x74, 0x1, 0xa8, 0x7, 0x8, 0x80, 0x2, 0xc3, 0x33, 0xd8, 0xfc, 0xf5, 
	0x29, 0x7b, 0xd, 0x12, 0x3, 0x2b, 0x7b, 0xe, 0x12, 0x3, 0x2b, 0x12, 0x3, 0x66, 0x12, 0x1b, 
	0x5d, 0x12, 0xa, 0x31, 0xe5, 0x24, 0x90, 0x1f, 0xd6, 0x93, 0xf5, 0x30, 0xe5, 0x26, 0x60, 0x1e, 
	0xe5, 0x24, 0x24, 0x1, 0xff, 0xe4, 0x33, 0xfe, 0xc3, 0xef, 0x94, 0x7, 0xee, 0x64, 0x80, 0x94, 
	0x80, 0x50, 0x8, 0xe5, 0x24, 0xa3, 0x93, 0xf5, 0x31, 0x80, 0x3, 0x75, 0x31, 0x7, 0x7e, 00, 
	0x7f, 0x2, 0x7d, 0xff, 0x7b, 00, 0x7a, 00, 0x79, 0x2a, 0x12, 0x5, 0x2e, 0x75, 0x28, 0x2, 
	0xe5, 0x28, 0xf4, 0x70, 0x3, 0x2, 0x2, 0x50, 0x74, 0x1, 0x7e, 00, 0xa8, 0x28, 0x8, 0x80, 
	0x5, 0xc3, 0x33, 0xce, 0x33, 0xce, 0xd8, 0xf9, 0xff, 0xef, 0x55, 0x29, 0x70, 0x3, 0x2, 0x2, 
	0x4b, 0xe5, 0x28, 0xd3, 0x94, 0x2, 0x40, 0x3, 0x2, 0x1, 0xec, 0xaf, 0x30, 0x12, 0x1e, 0xa0, 
	0x40, 0x3, 0x2, 0x2, 0x4b, 0xaf, 0x31, 0x12, 0x1e, 0xa0, 0x50, 0x7, 0xe5, 0x31, 0xc3, 0x94, 
	0x7, 0x40, 0x7, 0xe5, 0x26, 0x60, 0x3, 0x2, 0x2, 0x4b, 0xe5, 0x30, 0x64, 0x6, 0x70, 0x1b, 
	0xe5, 0x28, 0x70, 0x9, 0x12, 0x3, 0x5d, 0x20, 0xe4, 0x3, 0x2, 0x2, 0x4b, 0xe5, 0x28, 0xb4, 
	0x1, 0x9, 0x12, 0x3, 0x5d, 0x20, 0xe6, 0x3, 0x2, 0x2, 0x4b, 0xe5, 0x26, 0x60, 0x67, 0xe5, 
	0x28, 0xc3, 0x94, 0x2, 0x50, 0x6f, 0xe5, 0x31, 0x64, 0x6, 0x70, 0x25, 0xe5, 0x28, 0x70, 0xe, 
	0x7d, 0x1, 0xaf, 0x31, 0x12, 0x13, 0x27, 0xee, 0x20, 0xe4, 0x3, 0x2, 0x2, 0x4b, 0xe5, 0x28, 
	0xb4, 0x1, 0xe, 0x7d, 0x1, 0xaf, 0x31, 0x12, 0x13, 0x27, 0xee, 0x20, 0xe6, 0x3, 0x2, 0x2, 
	0x4b, 0x12, 0x3, 0x70, 0xe4, 0xf5, 0x2e, 0xf5, 0x2f, 0xaf, 0x30, 0x12, 0x1e, 0xbc, 0x50, 0x7, 
	0xaf, 0x31, 0x12, 0x1e, 0xbc, 0x40, 0x2e, 0xe5, 0x2f, 0xc3, 0x94, 0x3, 0x50, 0x27, 0x7f, 0xa, 
	0x12, 0x1d, 0x6c, 0x5, 0x2e, 0xe5, 0x2e, 0xc3, 0x94, 0xfa, 0x40, 0xdd, 0x12, 0x3, 0x70, 0x5, 
	0x2f, 0xe4, 0xf5, 0x2e, 0x80, 0xd3, 0xe5, 0x28, 0x44, 0x90, 0xfd, 0xaf, 0x30, 0x12, 0x14, 0x1c, 
	0x7f, 0xa, 0x12, 0x1f, 0x1b, 0xad, 0x28, 0xaf, 0x30, 0x12, 0x11, 0xa2, 0xe5, 0x26, 0x60, 0x7, 
	0xad, 0x28, 0xaf, 0x31, 0x12, 0x11, 0xa2, 0x7f, 0x5, 0x12, 0x1f, 0x1b, 0xe5, 0x26, 0x60, 0x41, 
	0xad, 0x31, 0x12, 0x3, 0x4e, 0x80, 0x2, 0xc3, 0x33, 0xd8, 0xfc, 0x12, 0x1b, 0x74, 0x80, 0x2, 
	0xc3, 0x33, 0xd8, 0xfc, 0xf4, 0x55, 0x2a, 0x4f, 0xf5, 0x2a, 0xad, 0x30, 0xaf, 0x31, 0x12, 0xf, 
	0xfa, 0xe5, 0x28, 0x25, 0xe0, 0xfe, 0xef, 0xa8, 0x6, 0x8, 0x80, 0x2, 0xc3, 0x33, 0xd8, 0xfc, 
	0x12, 0x1b, 0x74, 0x80, 0x2, 0xc3, 0x33, 0xd8, 0xfc, 0xf4, 0x55, 0x2b, 0x4f, 0xf5, 0x2b, 0x80, 
	0x1a, 0xad, 0x30, 0x12, 0x3, 0x4e, 0x80, 0x2, 0xc3, 0x33, 0xd8, 0xfc, 0x12, 0x1b, 0x74, 0x80, 
	0x2, 0xc3, 0x33, 0xd8, 0xfc, 0xf4, 0x55, 0x2a, 0x4f, 0xf5, 0x2a, 0x15, 0x28, 0x2, 00, 0xf0, 
	0xe4, 0xf5, 0x2d, 0xe5, 0x26, 0x70, 0x7, 0xe5, 0x2d, 0xc3, 0x94, 0x1, 0x40, 0x11, 0xe5, 0x26, 
	0x70, 0x3, 0x2, 0x2, 0xfd, 0xe5, 0x2d, 0xc3, 0x94, 0x2, 0x40, 0x3, 0x2, 0x2, 0xfd, 0x12, 
	0x1f, 0x39, 0x7f, 0x4, 0x12, 0x1d, 0x31, 0xaf, 0x24, 0x12, 0x1a, 0xda, 0x7f, 0x4, 0x12, 0x1d, 
	0x31, 0xe4, 0xf5, 0x28, 0xe5, 0x28, 0x25, 0xe0, 0xff, 0x74, 0x2a, 0x25, 0x2d, 0xf8, 0xe6, 0xfe, 
	0xa8, 0x7, 0x8, 0x80, 0x2, 0xc3, 0x13, 0xd8, 0xfc, 0x54, 0x3, 0xf5, 0x2c, 0xb4, 0x3, 0x7, 
	0x7f, 0x8, 0x12, 0x1d, 0x31, 0x80, 0x24, 0xe5, 0x2c, 0x7b, 0xff, 0x70, 0x9, 0x7a, 0x9, 0x79, 
	0xf3, 0x12, 0x19, 0xe9, 0x80, 0x15, 0x7a, 0x9, 0x79, 0xfc, 0x12, 0x19, 0xe9, 0xaf, 0x2c, 0xe4, 
	0xfc, 0xfd, 0xfe, 0x12, 0xf, 0x15, 0x7f, 0x20, 0x12, 0x1f, 0xe8, 0x5, 0x28, 0xe5, 0x28, 0xc3, 
	0x94, 0x3, 0x40, 0xb0, 0xe5, 0x2d, 0x70, 0xb, 0xaf, 0x30, 0x12, 0x1e, 0xa0, 0x50, 0x10, 0xaf, 
	0x30, 0x80, 0x9, 0xaf, 0x31, 0x12, 0x1e, 0xa0, 0x50, 0x5, 0xaf, 0x31, 0x12, 0x1f, 0x4d, 0x5, 
	0x24, 0xe5, 0x24, 0xd3, 0x95, 0x27, 0x50, 0x5, 0x5, 0x2d, 0x2, 0x2, 0x53, 0xe5, 0x24, 0xd3, 
	0x95, 0x27, 0x50, 0x3, 0x2, 00, 0xb4, 0x7b, 0xff, 0x7a, 0xa, 0x79, 0x3, 0x12, 0x19, 0xe9, 
	0x7b, 0xd, 0x12, 0x3, 0x3e, 0x12, 0xa, 0x31, 0x7b, 0xe, 0x12, 0x3, 0x3e, 0x12, 0xa, 0x31, 
	0x12, 0x3, 0x66, 0x7f, 0x5f, 0x12, 0x1b, 0x5f, 0x2, 0xa, 0x31, 0xe4, 0xfd, 0x7f, 0x40, 0x12, 
	0xa, 0x2a, 0x7f, 00, 0x7e, 00, 0x7d, 00, 0x7c, 00, 0x12, 0xa, 0x31, 0x22, 0xe4, 0xfd, 
	0x7f, 0x40, 0x12, 0xa, 0x2a, 0x7f, 0x5f, 0x7e, 00, 0x7d, 00, 0x7c, 00, 0x22, 0xaf, 0x30, 
	0x12, 0xf, 0xfa, 0xe5, 0x28, 0x25, 0xe0, 0xfe, 0xef, 0xa8, 0x6, 0x8, 0x22, 0x7d, 0x1, 0xaf, 
	0x30, 0x12, 0x13, 0x27, 0xee, 0x22, 0x7b, 0xf, 0xe4, 0xfd, 0x7f, 0x40, 0x12, 0xa, 0x2a, 0x22, 
	0xab, 0x28, 0xad, 0x31, 0xaf, 0x30, 0x12, 0x18, 0xdf, 0x22, 0xbb, 0x1, 0x6, 0x89, 0x82, 0x8a, 
	0x83, 0xe0, 0x22, 0x50, 0x2, 0xe7, 0x22, 0xbb, 0xfe, 0x2, 0xe3, 0x22, 0x89, 0x82, 0x8a, 0x83, 
	0xe4, 0x93, 0x22, 0xbb, 0x1, 0x6, 0x89, 0x82, 0x8a, 0x83, 0xf0, 0x22, 0x50, 0x2, 0xf7, 0x22, 
	0xbb, 0xfe, 0x1, 0xf3, 0x22, 0xe8, 0x8f, 0xf0, 0xa4, 0xcc, 0x8b, 0xf0, 0xa4, 0x2c, 0xfc, 0xe9, 
	0x8e, 0xf0, 0xa4, 0x2c, 0xfc, 0x8a, 0xf0, 0xed, 0xa4, 0x2c, 0xfc, 0xea, 0x8e, 0xf0, 0xa4, 0xcd, 
	0xa8, 0xf0, 0x8b, 0xf0, 0xa4, 0x2d, 0xcc, 0x38, 0x25, 0xf0, 0xfd, 0xe9, 0x8f, 0xf0, 0xa4, 0x2c, 
	0xcd, 0x35, 0xf0, 0xfc, 0xeb, 0x8e, 0xf0, 0xa4, 0xfe, 0xa9, 0xf0, 0xeb, 0x8f, 0xf0, 0xa4, 0xcf, 
	0xc5, 0xf0, 0x2e, 0xcd, 0x39, 0xfe, 0xe4, 0x3c, 0xfc, 0xea, 0xa4, 0x2d, 0xce, 0x35, 0xf0, 0xfd, 
	0xe4, 0x3c, 0xfc, 0x22, 0x75, 0xf0, 0x8, 0x75, 0x82, 00, 0xef, 0x2f, 0xff, 0xee, 0x33, 0xfe, 
	0xcd, 0x33, 0xcd, 0xcc, 0x33, 0xcc, 0xc5, 0x82, 0x33, 0xc5, 0x82, 0x9b, 0xed, 0x9a, 0xec, 0x99, 
	0xe5, 0x82, 0x98, 0x40, 0xc, 0xf5, 0x82, 0xee, 0x9b, 0xfe, 0xed, 0x9a, 0xfd, 0xec, 0x99, 0xfc, 
	0xf, 0xd5, 0xf0, 0xd6, 0xe4, 0xce, 0xfb, 0xe4, 0xcd, 0xfa, 0xe4, 0xcc, 0xf9, 0xa8, 0x82, 0x22, 
	0xb8, 00, 0xc1, 0xb9, 00, 0x59, 0xba, 00, 0x2d, 0xec, 0x8b, 0xf0, 0x84, 0xcf, 0xce, 0xcd, 
	0xfc, 0xe5, 0xf0, 0xcb, 0xf9, 0x78, 0x18, 0xef, 0x2f, 0xff, 0xee, 0x33, 0xfe, 0xed, 0x33, 0xfd, 
	0xec, 0x33, 0xfc, 0xeb, 0x33, 0xfb, 0x10, 0xd7, 0x3, 0x99, 0x40, 0x4, 0xeb, 0x99, 0xfb, 0xf, 
	0xd8, 0xe5, 0xe4, 0xf9, 0xfa, 0x22, 0x78, 0x18, 0xef, 0x2f, 0xff, 0xee, 0x33, 0xfe, 0xed, 0x33, 
	0xfd, 0xec, 0x33, 0xfc, 0xc9, 0x33, 0xc9, 0x10, 0xd7, 0x5, 0x9b, 0xe9, 0x9a, 0x40, 0x7, 0xec, 
	0x9b, 0xfc, 0xe9, 0x9a, 0xf9, 0xf, 0xd8, 0xe0, 0xe4, 0xc9, 0xfa, 0xe4, 0xcc, 0xfb, 0x22, 0x75, 
	0xf0, 0x10, 0xef, 0x2f, 0xff, 0xee, 0x33, 0xfe, 0xed, 0x33, 0xfd, 0xcc, 0x33, 0xcc, 0xc8, 0x33, 
	0xc8, 0x10, 0xd7, 0x7, 0x9b, 0xec, 0x9a, 0xe8, 0x99, 0x40, 0xa, 0xed, 0x9b, 0xfd, 0xec, 0x9a, 
	0xfc, 0xe8, 0x99, 0xf8, 0xf, 0xd5, 0xf0, 0xda, 0xe4, 0xcd, 0xfb, 0xe4, 0xcc, 0xfa, 0xe4, 0xc8, 
	0xf9, 0x22, 0xcf, 0xf4, 0xcf, 0xce, 0xf4, 0xce, 0xcd, 0xf4, 0xcd, 0xcc, 0xf4, 0xcc, 0x22, 0xeb, 
	0x9f, 0xf5, 0xf0, 0xea, 0x9e, 0x42, 0xf0, 0xe9, 0x9d, 0x42, 0xf0, 0xe8, 0x9c, 0x45, 0xf0, 0x22, 
	0xe8, 0x60, 0xf, 0xec, 0xc3, 0x13, 0xfc, 0xed, 0x13, 0xfd, 0xee, 0x13, 0xfe, 0xef, 0x13, 0xff, 
	0xd8, 0xf1, 0x22, 0xe8, 0x60, 0xf, 0xef, 0xc3, 0x33, 0xff, 0xee, 0x33, 0xfe, 0xed, 0x33, 0xfd, 
	0xec, 0x33, 0xfc, 0xd8, 0xf1, 0x22, 0xe6, 0xfc, 0x8, 0xe6, 0xfd, 0x8, 0xe6, 0xfe, 0x8, 0xe6, 
	0xff, 0x22, 0xe4, 0x93, 0xfc, 0x74, 0x1, 0x93, 0xfd, 0x74, 0x2, 0x93, 0xfe, 0x74, 0x3, 0x93, 
	0xff, 0x22, 0xec, 0xf6, 0x8, 0xed, 0xf6, 0x8, 0xee, 0xf6, 0x8, 0xef, 0xf6, 0x22, 0xef, 0x4e, 
	0x60, 0x12, 0xef, 0x60, 0x1, 0xe, 0xed, 0xbb, 0x1, 0xb, 0x89, 0x82, 0x8a, 0x83, 0xf0, 0xa3, 
	0xdf, 0xfc, 0xde, 0xfa, 0x22, 0x89, 0xf0, 0x50, 0x7, 0xf7, 0x9, 0xdf, 0xfc, 0xa9, 0xf0, 0x22, 
	0xbb, 0xfe, 0xfc, 0xf3, 0x9, 0xdf, 0xfc, 0xa9, 0xf0, 0x22, 0xa8, 0x1a, 0xe6, 0xf5, 0x24, 0x64, 
	0x20, 0x60, 0x9, 0xe5, 0x24, 0x64, 0xd, 0x60, 0x3, 0x7f, 0x1, 0x22, 0xe4, 0xf5, 0x19, 0xfe, 
	0x7f, 0x10, 0x7d, 0xff, 0xfb, 0x7a, 00, 0x79, 0x52, 0x12, 0x5, 0x2e, 0x12, 0x1f, 0x5, 0x75, 
	0x25, 0xa, 0xe4, 0xf5, 0x27, 0xa8, 0x1a, 0xe6, 0x64, 0x30, 0x70, 0x1b, 0xe8, 0x4, 0xf8, 0xe6, 
	0xff, 0x12, 0x1c, 0x52, 0xbf, 0x58, 0x10, 0x75, 0x25, 0x10, 0x5, 0x1a, 0x5, 0x1a, 0xa8, 0x1a, 
	0xe6, 0xb4, 0x20, 0x3, 0x7f, 0x1, 0x22, 0xa8, 0x1a, 0xe6, 0xf5, 0x24, 0x64, 0x20, 0x60, 0x6, 
	0xe5, 0x24, 0x64, 0xd, 0x70, 0x7c, 0xe5, 0x27, 0xd3, 0x94, 00, 0x40, 0x6a, 0xe4, 0xf5, 0x35, 
	0xf5, 0x34, 0xf5, 0x33, 0xf5, 0x32, 0xf5, 0x26, 0xe5, 0x26, 0xc3, 0x95, 0x27, 0x50, 0x3a, 0xaf, 
	0x25, 0xe4, 0xfc, 0xfd, 0xfe, 0xab, 0x35, 0xaa, 0x34, 0xa9, 0x33, 0xa8, 0x32, 0x12, 0x3, 0xa5, 
	0xc0, 0x4, 0xa9, 0x5, 0xaa, 0x6, 0xab, 0x7, 0x74, 0x28, 0x25, 0x26, 0xf8, 0xe6, 0xff, 0xe4, 
	0xfc, 0xfd, 0xfe, 0xd0, 00, 0xeb, 0x2f, 0xf5, 0x35, 0xee, 0x3a, 0xf5, 0x34, 0xed, 0x39, 0xf5, 
	0x33, 0xec, 0x38, 0xf5, 0x32, 0x5, 0x26, 0x80, 0xbf, 0xe5, 0x19, 0xc3, 0x94, 0x4, 0x50, 0x17, 
	0xaf, 0x35, 0xae, 0x34, 0xad, 0x33, 0xac, 0x32, 0xab, 0x19, 0x5, 0x19, 0xeb, 0x25, 0xe0, 0x25, 
	0xe0, 0x24, 0x52, 0xf8, 0x12, 0x5, 0x22, 0xe5, 0x24, 0x64, 0xd, 0x60, 0x3, 0x2, 0x5, 0x7c, 
	0xff, 0x22, 0xaf, 0x24, 0x12, 0x19, 0x9e, 0x8f, 0x24, 0xe5, 0x24, 0xf4, 0x60, 0x1d, 0xe5, 0x27, 
	0xc3, 0x94, 0xa, 0x50, 0x19, 0xaf, 0x27, 0x5, 0x27, 0x74, 0x28, 0x2f, 0xf8, 0xa6, 0x24, 0xe5, 
	0x24, 0xd3, 0x94, 0x9, 0x40, 0x8, 0x75, 0x25, 0x10, 0x80, 0x3, 0x7f, 0x1, 0x22, 0x5, 0x1a, 
	0x2, 0x5, 0xa7, 0xa8, 0x1a, 0xe6, 0xf5, 0x23, 0x5, 0x1a, 0x12, 0x5, 0x5a, 0xef, 0x60, 0x3, 
	0x7f, 0x1, 0x22, 0xaf, 0x23, 0x12, 0x1c, 0x52, 0xef, 0x24, 0xb7, 0x60, 0x64, 0x24, 0xfa, 0x60, 
	0x7a, 0x24, 0xfd, 0x60, 0x17, 0x24, 0xfe, 0x70, 0x3, 0x2, 0x7, 0x25, 0x24, 0xfd, 0x60, 0x28, 
	0x4, 0x60, 0x3, 0x2, 0x7, 0x60, 0x12, 0xf, 0x8b, 0x2, 0x7, 0x63, 0x12, 0x1f, 0x2f, 0x12, 
	0x1a, 0x1e, 0xc0, 0x7, 0x12, 0x1a, 0xe, 0xc0, 0x5, 0x12, 0x1a, 0x16, 0xd0, 0x5, 0xd0, 0x7, 
	0x12, 0xa, 0x16, 0x12, 0x1c, 0x20, 0x80, 0x3e, 0xe5, 0x19, 0xc3, 0x94, 0x4, 0x50, 0x3, 0x2, 
	0x7, 0x63, 0x12, 0x1a, 0x1e, 0xc0, 0x7, 0x12, 0x1a, 0xe, 0xc0, 0x5, 0x12, 0x1a, 0x16, 0xd0, 
	0x5, 0xd0, 0x7, 0x12, 0xa, 0x2a, 0x78, 0x5e, 0x12, 0x5, 0x6, 0x12, 0xa, 0x31, 0x2, 0x7, 
	0x63, 0x12, 0x1f, 0x2f, 0x78, 0x52, 0x12, 0x5, 0x6, 0xc0, 0x7, 0x12, 0x1a, 0xe, 0xd0, 0x7, 
	0x12, 0x13, 0x27, 0x12, 0x1e, 0x1c, 0x12, 0x1f, 0x39, 0x80, 0x68, 0xe5, 0x19, 0xc3, 0x94, 0x3, 
	0x40, 0x61, 0x78, 0x52, 0x12, 0x5, 0x6, 0xc0, 0x7, 0x12, 0x1a, 0xe, 0xc0, 0x5, 0x12, 0x1a, 
	0x16, 0xaa, 0x6, 0x78, 0x5e, 0x12, 0x5, 0x6, 0x8e, 0x3d, 0x8f, 0x3e, 0xd0, 0x5, 0xd0, 0x7, 
	0x12, 0x19, 0x52, 0x80, 0x3e, 0x78, 0x52, 0x12, 0x5, 0x6, 0xef, 0x24, 0xfe, 0x60, 0x17, 0x4, 
	0x70, 0x2b, 0x78, 0x56, 0x12, 0x5, 0x6, 0xc0, 0x7, 0x78, 0x5a, 0x12, 0x5, 0x6, 0xad, 0x7, 
	0xd0, 0x7, 0xe4, 0xfb, 0x80, 0x12, 0x78, 0x56, 0x12, 0x5, 0x6, 0xc0, 0x7, 0x78, 0x5a, 0x12, 
	0x5, 0x6, 0xad, 0x7, 0xd0, 0x7, 0x7b, 0x1, 0x12, 00, 0x46, 0x80, 0x6, 0x7f, 0x1, 0x22, 
	0x7f, 0x1, 0x22, 0x7f, 00, 0x22, 0x8f, 0x2a, 0x7f, 0x1, 0xed, 0x54, 0x30, 0xfe, 0xbe, 0x30, 
	0x4, 0x7e, 0x1, 0x80, 0x2, 0x7e, 00, 0xed, 0x54, 0x3, 0xfc, 0xbc, 0x2, 0x4, 0x7c, 0x1, 
	0x80, 0x2, 0x7c, 00, 0x8c, 0x2c, 0xed, 0x20, 0xe4, 0x7, 0xef, 0x60, 0x4, 0x7d, 0x1, 0x80, 
	0x2, 0x7d, 00, 0xef, 0x60, 0x3, 0xee, 0x70, 0x3, 0xed, 0x60, 0x4, 0x7f, 0x1, 0x80, 0x2, 
	0x7f, 00, 0x8f, 0x2b, 0x7d, 0x30, 0xaf, 0x2a, 0x12, 0x18, 0xb5, 0x7b, 0x8, 0xad, 0x2a, 0x7f, 
	0x20, 0x12, 0xa, 0x2a, 0xe5, 0x31, 0xfe, 0xe5, 0x32, 0xff, 0xad, 0x30, 0x7c, 00, 0x12, 0xa, 
	0x47, 0x12, 0xa, 0x31, 0x7b, 0xc, 0xad, 0x2a, 0x7f, 0x20, 0x12, 0xa, 0x2a, 0xe5, 0x34, 0xfe, 
	0xe5, 0x35, 0xff, 0xad, 0x33, 0x7c, 00, 0x12, 0xa, 0x47, 0x12, 0xa, 0x31, 0xe5, 0x2b, 0x60, 
	0x1e, 0x75, 0x2d, 0x17, 0x75, 0x2e, 0xe7, 0x12, 0x8, 0x52, 0x7f, 0xff, 0x7e, 00, 0x7d, 0x3, 
	0x7c, 00, 0x12, 0xa, 0x31, 0x85, 0x2a, 0x3c, 0xe4, 0xf5, 0x3d, 0x7b, 0x12, 0x80, 0x1d, 0x75, 
	0x2d, 0x17, 0x75, 0x2e, 0xcf, 0x12, 0x8, 0x52, 0x7f, 00, 0x7e, 00, 0x7d, 00, 0x7c, 00, 
	0x12, 0xa, 0x31, 0x85, 0x2a, 0x3c, 0x75, 0x3d, 0x1, 0x7b, 0x12, 0xe4, 0xfd, 0x7f, 0xa0, 0x12, 
	0x17, 0x9d, 0xe4, 0xf5, 0x2f, 0xe5, 0x2f, 0x24, 0xde, 0xfb, 0xad, 0x2a, 0x7f, 0x20, 0x12, 0xa, 
	0x2a, 0x85, 0x2e, 0x82, 0x85, 0x2d, 0x83, 0x12, 0x5, 0x12, 0x12, 0xa, 0x31, 0x74, 0x4, 0x25, 
	0x2e, 0xf5, 0x2e, 0xe4, 0x35, 0x2d, 0xf5, 0x2d, 0x5, 0x2f, 0xe5, 0x2f, 0xc3, 0x94, 0x6, 0x40, 
	0xd4, 0x22, 0x7b, 0x4, 0xad, 0x2a, 0x7f, 0x20, 0x12, 0xa, 0x2a, 0x22, 0x8f, 0x2b, 0x8c, 0x2c, 
	0x8d, 0x2d, 0xe5, 0x2d, 0xf4, 0x45, 0x2c, 0x70, 0xa, 0xf5, 0x31, 0xf5, 0x30, 0xf5, 0x2f, 0xf5, 
	0x2e, 0x80, 0x12, 0xe5, 0x2d, 0x54, 0x13, 0xfd, 0xaf, 0x2b, 0x12, 0x12, 0x6c, 0x8f, 0x31, 0x8e, 
	0x30, 0x8d, 0x2f, 0x8c, 0x2e, 0xe4, 0xf5, 0x39, 0xf5, 0x38, 0xf5, 0x37, 0x75, 0x36, 0x80, 0xf5, 
	0x3d, 0xf5, 0x3c, 0xf5, 0x3b, 0x75, 0x3a, 0x80, 0x7b, 0xdf, 0xad, 0x2b, 0x7f, 0x20, 0x12, 0x11, 
	0x3a, 0x12, 0x9, 0x3a, 0xe5, 0x31, 0x54, 0xf8, 0xff, 0xe5, 0x30, 0xfe, 0xe5, 0x2f, 0xfd, 0xe5, 
	0x2e, 0xfc, 0xef, 0x44, 0x31, 0xff, 0xec, 0x44, 0x20, 0xfc, 0x12, 0xa, 0x31, 0xe5, 0x2d, 0xf4, 
	0x45, 0x2c, 0x70, 0x23, 0xfd, 0xaf, 0x2b, 0x12, 0x1e, 0xd7, 0x12, 0x9, 0x3a, 0x7f, 0x30, 0x7e, 
	00, 0x7d, 00, 0x7c, 0x20, 0x12, 0xa, 0x31, 0x85, 0x2b, 0x3c, 0x75, 0x3d, 0x1, 0x7b, 0x12, 
	0xe4, 0xfd, 0x7f, 0xa0, 0x2, 0x17, 0x9d, 0xaf, 0x2b, 0x12, 0x1f, 0x8d, 0xe4, 0xfc, 0xfd, 0xfe, 
	0x78, 0x13, 0x12, 0x4, 0xf3, 0xec, 0x44, 0x8, 0xfc, 0xe5, 0x31, 0x4f, 0xf5, 0x31, 0xe5, 0x30, 
	0x4e, 0xf5, 0x30, 0xe5, 0x2f, 0x4d, 0xf5, 0x2f, 0xe5, 0x2e, 0x4c, 0xf5, 0x2e, 0xe4, 0xfb, 0xad, 
	0x2b, 0x7f, 0x20, 0x12, 0xa, 0x2a, 0xaf, 0x31, 0xae, 0x30, 0xad, 0x2f, 0xac, 0x2e, 0x12, 0xa, 
	0x31, 0xaf, 0x2b, 0x12, 0x1b, 0x7b, 0xe5, 0x2d, 0x54, 0x13, 0xfd, 0xaf, 0x2b, 0x12, 0x19, 0x6, 
	0xe5, 0x2b, 0xb4, 0x6, 0x4, 0xff, 0x12, 0x13, 0xd0, 0x22, 0xe4, 0xfb, 0xad, 0x2b, 0x7f, 0x20, 
	0x12, 0xa, 0x2a, 0x22, 0x49, 0x6e, 0x76, 0x61, 0x6c, 0x69, 0x64, 0x20, 0x63, 0x6f, 0x6d, 0x6d, 
	0x61, 0x6e, 0x64, 0xd, 0xa, 00, 0x43, 0x68, 0x69, 0x70, 0x20, 0x69, 0x64, 0x3a, 0x20, 00, 
	0x50, 0x6f, 0x72, 0x74, 0x20, 0x53, 0x4d, 0x41, 0x43, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 
	0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x4d, 0x49, 0x49, 0x4d, 0x20, 0x50, 0x48, 0x59, 0xd, 
	0xa, 00, 0x50, 0x6f, 0x72, 0x74, 0x20, 00, 0x2c, 0x20, 0x4d, 0x49, 0x49, 0x4d, 0x20, 00, 
	0x2c, 0x20, 0x50, 0x48, 0x59, 0x20, 00, 0x49, 0x6e, 0x74, 0x65, 0x72, 0x6e, 0x61, 0x6c, 0x20, 
	0x6c, 0x6f, 0x6f, 0x70, 0x62, 0x61, 0x63, 0x6b, 0x20, 0x6f, 0x6e, 0x20, 0x50, 0x48, 0x59, 0x73, 
	0x3a, 0xd, 0xa, 00, 0x45, 0x78, 0x74, 0x65, 0x72, 0x6e, 0x61, 0x6c, 0x20, 0x6c, 0x6f, 0x6f, 
	0x70, 0x62, 0x61, 0x63, 0x6b, 0x20, 0x6f, 0x6e, 0x20, 0x50, 0x48, 0x59, 0x73, 0x3a, 0xd, 0xa, 
	00, 0xd, 0xa, 0x20, 0x20, 0x20, 0x20, 0x50, 0x6f, 0x72, 0x74, 0x20, 0x20, 0x40, 0x31, 0x30, 
	0x20, 0x20, 0x20, 0x20, 0x20, 0x40, 0x31, 0x30, 0x30, 0x20, 0x20, 0x20, 0x20, 0x40, 0x31, 0x30, 
	0x30, 0x30, 00, 0x4f, 0x6b, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 00, 0x45, 0x72, 0x72, 0x6f, 
	0x72, 0x20, 00, 0xd, 0xa, 0x54, 0x65, 0x73, 0x74, 0x20, 0x63, 0x6f, 0x6d, 0x70, 0x6c, 0x65, 
	0x74, 0x65, 0x64, 0xd, 0xa, 00, 0xef, 0x2d, 0xf5, 0xf9, 0x8b, 0xfa, 0x30, 0xf8, 0xfd, 0xd2, 
	0xf8, 0xac, 0xff, 0xad, 0xfe, 0xae, 0xfd, 0xaf, 0xfc, 0x22, 0xef, 0x2d, 0xf5, 0x4b, 0x8b, 0x4c, 
	0x22, 0x8c, 0xff, 0x8d, 0xfe, 0x8e, 0xfd, 0x8f, 0xfc, 0xe5, 0x4b, 0xf5, 0xf9, 0xe5, 0x4c, 0xf5, 
	0xfb, 0x30, 0xf8, 0xfd, 0xd2, 0xf8, 0x22, 0x22, 0xec, 0xfe, 0xed, 0xff, 0x22, 0x1, 0x2, 0x4, 
	0x8, 0x10, 0x20, 0x40, 0x80, 0xe4, 0xfe, 0xcf, 0x90, 0xa, 0x4d, 0xb4, 0x8, 00, 0x50, 0x3, 
	0x93, 0xff, 0x22, 0x94, 0x8, 0x93, 0xfe, 0x22, 0xef, 0x90, 0xa, 0x4d, 0x93, 0xff, 0x22, 0x90, 
	0xa, 0x4d, 0xef, 0x93, 0xff, 0xf4, 0xfe, 0xbb, 0x1, 0xe, 0x89, 0x82, 0x8a, 0x83, 0xe0, 0x5e, 
	0xbd, 00, 0x2, 0x80, 0x1, 0x4f, 0xf0, 0x22, 0xbb, 00, 0xb, 0xe9, 0xf8, 0xe6, 0x5e, 0xbd, 
	00, 0x2, 0x80, 0x1, 0x4f, 0xf6, 0x22, 0x74, 0x1f, 0xc3, 0x9f, 0x13, 0x13, 0x13, 0x54, 0x3, 
	0x29, 0xf9, 0xe4, 0x3a, 0xfa, 0x74, 0x7, 0x5f, 0xff, 0x12, 0xa, 0x6f, 0x22, 0x90, 0xa, 0x4d, 
	0xef, 0x93, 0xff, 0xbb, 0x1, 0x7, 0x89, 0x82, 0x8a, 0x83, 0xe0, 0x80, 0x6, 0xbb, 00, 0x8, 
	0xe9, 0xf8, 0xe6, 0x5f, 0x60, 0x2, 0xd3, 0x22, 0xc3, 0x22, 0x74, 0x1f, 0xc3, 0x9f, 0x13, 0x13, 
	0x13, 0x54, 0x3, 0x29, 0xf9, 0xe4, 0x3a, 0xfa, 0x74, 0x7, 0x5f, 0xff, 0x12, 0xa, 0xad, 0x22, 
	0x8f, 0x36, 0xe5, 0x36, 0x64, 0x6, 0x70, 0x54, 0x7b, 0x22, 0x7a, 0x1a, 0x7d, 0x17, 0x12, 0x17, 
	0xff, 0xaf, 0x36, 0x12, 0x18, 0x2f, 0x12, 0xb, 0x6d, 0x7a, 0x2, 0x12, 0xb, 0x63, 0x7b, 0xb5, 
	0x7a, 0x52, 0x7d, 0x1f, 0xaf, 0x36, 0x12, 0x17, 0xff, 0x7b, 0xf, 0x7a, 00, 0x7d, 0x2, 0xaf, 
	0x36, 0x12, 0x17, 0xff, 0x7b, 0x2a, 0x7a, 0x47, 0x7d, 0x1, 0xaf, 0x36, 0x12, 0x17, 0xff, 0x7b, 
	0xa4, 0x7a, 0x8f, 0xe4, 0xfd, 0xaf, 0x36, 0x12, 0x17, 0xff, 0x12, 0xb, 0x6d, 0x7a, 00, 0x12, 
	0xb, 0x63, 0xe4, 0xfb, 0xfa, 0x7d, 0x1f, 0xaf, 0x36, 0x2, 0x17, 0xff, 0xaf, 0x36, 0x12, 0xe, 
	0x9e, 0xaf, 0x36, 0x12, 0x1c, 0xb1, 0x7b, 0x64, 0x7a, 00, 0x7d, 0x1b, 0xaf, 0x36, 0x12, 0x17, 
	0xff, 0x75, 0x3d, 00, 0x75, 0x3e, 0xf, 0x7b, 0x3, 0x7a, 00, 0x7d, 0x1e, 0xaf, 0x36, 0x12, 
	0x19, 0x52, 0x22, 0x7b, 0x12, 0x7d, 0x8, 0xaf, 0x36, 0x12, 0x17, 0xff, 0x22, 0x7b, 0x30, 0x7a, 
	0x2a, 0x7d, 0x1f, 0xaf, 0x36, 0x12, 0x17, 0xff, 0x22, 0x8f, 0x38, 0x12, 0x1f, 0x7c, 0x12, 0x1a, 
	0x62, 0x12, 0xb, 0xf8, 0x12, 0x1f, 0x73, 0x7a, 0xaf, 0x12, 0xb, 0xda, 0x75, 0x3d, 00, 0x75, 
	0x3e, 0xc, 0x7d, 0x11, 0x12, 0xb, 0xee, 0x7a, 0x8f, 0x12, 0xb, 0xda, 0x7a, 0xaf, 0x12, 0xb, 
	0xe4, 0x75, 0x3d, 00, 0x75, 0x3e, 0xc, 0x7d, 0x12, 0x12, 0xb, 0xee, 0x7a, 0x8f, 0x12, 0xb, 
	0xe4, 0x7a, 0xaf, 0x12, 0xc, 0x2, 0x75, 0x3d, 0x1, 0x75, 0x3e, 0x80, 0x7b, 00, 0x7a, 0x1, 
	0x7d, 0x11, 0xaf, 0x38, 0x12, 0x19, 0x52, 0x7a, 0x8f, 0x12, 0xc, 0x2, 0xaf, 0x38, 0x12, 0x1f, 
	0x7c, 0x12, 0x1a, 0x6d, 0x12, 0xb, 0xf8, 0x2, 0x1f, 0x85, 0x7b, 0x8a, 0x7d, 0x10, 0xaf, 0x38, 
	0x12, 0x17, 0xff, 0x22, 0x7b, 0x86, 0x7d, 0x10, 0xaf, 0x38, 0x12, 0x17, 0xff, 0x22, 0x7b, 0x8, 
	0x7a, 00, 0xaf, 0x38, 0x12, 0x19, 0x52, 0x22, 0x7d, 0x8, 0xaf, 0x38, 0x12, 0x19, 0x52, 0xaf, 
	0x38, 0x22, 0x7b, 0x82, 0x7d, 0x10, 0xaf, 0x38, 0x12, 0x17, 0xff, 0x22, 0x78, 0xfe, 0xe4, 0xf6, 
	0xd8, 0xfd, 0x75, 0x4f, 00, 0x75, 0x50, 00, 0x75, 0x81, 0xa7, 0x2, 0xc, 0x59, 0x2, 0xe, 
	0x23, 0xe4, 0x93, 0xa3, 0xf8, 0xe4, 0x93, 0xa3, 0x40, 0x3, 0xf6, 0x80, 0x1, 0xf2, 0x8, 0xdf, 
	0xf4, 0x80, 0x29, 0xe4, 0x93, 0xa3, 0xf8, 0x54, 0x7, 0x24, 0xc, 0xc8, 0xc3, 0x33, 0xc4, 0x54, 
	0xf, 0x44, 0x20, 0xc8, 0x83, 0x40, 0x4, 0xf4, 0x56, 0x80, 0x1, 0x46, 0xf6, 0xdf, 0xe4, 0x80, 
	0xb, 0x1, 0x2, 0x4, 0x8, 0x10, 0x20, 0x40, 0x80, 0x90, 0x18, 0x5e, 0xe4, 0x7e, 0x1, 0x93, 
	0x60, 0xbc, 0xa3, 0xff, 0x54, 0x3f, 0x30, 0xe5, 0x9, 0x54, 0x1f, 0xfe, 0xe4, 0x93, 0xa3, 0x60, 
	0x1, 0xe, 0xcf, 0x54, 0xc0, 0x25, 0xe0, 0x60, 0xa8, 0x40, 0xb8, 0xe4, 0x93, 0xa3, 0xfa, 0xe4, 
	0x93, 0xa3, 0xf8, 0xe4, 0x93, 0xa3, 0xc8, 0xc5, 0x82, 0xc8, 0xca, 0xc5, 0x83, 0xca, 0xf0, 0xa3, 
	0xc8, 0xc5, 0x82, 0xc8, 0xca, 0xc5, 0x83, 0xca, 0xdf, 0xe9, 0xde, 0xe7, 0x80, 0xbe, 0x12, 0x16, 
	0x20, 0x40, 0x3, 0x2, 0xd, 0x23, 0x75, 0x1a, 0x62, 0x12, 0x1f, 0x5, 0x12, 0x1a, 0x28, 0x75, 
	0x27, 0x1e, 0x75, 0x28, 0x92, 0x12, 0x15, 0x6b, 0xef, 0x70, 0x4, 0xf5, 0x22, 0x80, 0x4a, 0x12, 
	0x1a, 0x28, 0x75, 0x27, 0x1e, 0x75, 0x28, 0x98, 0x12, 0x15, 0x6b, 0xef, 0x70, 0x36, 0x74, 0x7, 
	0x25, 0x1a, 0xf5, 0x1a, 0x12, 0x5, 0x5a, 0xef, 0x60, 0x2, 0x80, 0x6, 0xe5, 0x19, 0x64, 0x2, 
	0x60, 0x5, 0x75, 0x22, 0x1, 0x80, 0x18, 0x78, 0x52, 0x12, 0x5, 0x6, 0xc0, 0x6, 0xc0, 0x7, 
	0x12, 0x1a, 0xe, 0xac, 0x6, 0xd0, 0x7, 0xd0, 0x6, 0x12, 0x10, 0xd1, 0x12, 0x1d, 0x7f, 0xe4, 
	0xf5, 0x22, 0x80, 0x5, 0x12, 0x6, 0x63, 0x8f, 0x22, 0xe5, 0x22, 0x60, 0x13, 0x12, 0x1e, 0xe4, 
	0x50, 0x5, 0x12, 0x1e, 0xd, 0x80, 0xf6, 0x7b, 0xff, 0x7a, 0x9, 0x79, 0x44, 0x12, 0x19, 0xe9, 
	0xe4, 0xf5, 0x18, 0x22, 0x8f, 0x23, 0x74, 0xa0, 0x2f, 0xf8, 0xe6, 0x24, 0xfe, 0x60, 0x2a, 0x14, 
	0x60, 0x5c, 0x24, 0x2, 0x70, 0x6e, 0x7b, 0xe1, 0x7a, 0x5, 0x7d, 0x4, 0xaf, 0x23, 0x12, 0x17, 
	0xff, 0x7b, 00, 0x7a, 0x6, 0x7d, 0x9, 0xaf, 0x23, 0x12, 0x17, 0xff, 0xaf, 0x23, 0x12, 0x1e, 
	0x3a, 0x74, 0xa0, 0x25, 0x23, 0xf8, 0x76, 0x2, 0x22, 0x30, 0x9, 0x48, 0xaf, 0x23, 0x12, 0x1e, 
	0xbc, 0x50, 0x41, 0x12, 0x1f, 0xc1, 0x7d, 0x1, 0xaf, 0x23, 0x12, 0xa, 0x6f, 0xaf, 0x23, 0x12, 
	0x1c, 0x6b, 0x8f, 0x24, 0xad, 0x24, 0xaf, 0x23, 0x12, 0x1d, 0xa4, 0xad, 0x24, 0xaf, 0x23, 0x12, 
	0x16, 0x5c, 0x74, 0xa0, 0x25, 0x23, 0xf8, 0x76, 0x3, 0xaf, 0x23, 0x2, 0x1f, 0xeb, 0x30, 0x9, 
	0x13, 0xaf, 0x23, 0x12, 0x1e, 0xbc, 0x40, 0xc, 0x74, 0xa0, 0x25, 0x23, 0xf8, 0x76, 0x2, 0xaf, 
	0x23, 0x12, 0x1b, 0x98, 0x22, 0x12, 00, 0x3f, 0xe4, 0x75, 0x26, 0xc0, 0xf5, 0x25, 0xf5, 0x24, 
	0xf5, 0x23, 0xf5, 0x22, 0x7b, 00, 0x7a, 00, 0x79, 0x23, 0xaf, 0x22, 0x12, 0xa, 0xca, 0x40, 
	0x21, 0xe4, 0xfb, 0x7d, 0x2, 0x7f, 0x60, 0x12, 0xa, 0x2a, 0xe4, 0x25, 0x22, 0xff, 0xe4, 0x34, 
	0x4, 0xfe, 0xe4, 0x34, 0x1, 0xfd, 0xe4, 0x34, 0x1, 0xfc, 0x12, 0xa, 0x31, 0x7f, 0x2, 0x12, 
	0x1d, 0x6c, 0x5, 0x22, 0xe5, 0x22, 0xd3, 0x94, 0xf, 0x40, 0xc9, 0x7f, 0x3, 0x12, 0x1f, 0x1b, 
	0x12, 0x1f, 0xb3, 0x12, 0x1e, 0xfa, 0x7f, 0x4, 0x12, 0x1f, 0x1b, 0x75, 0x3c, 0x1, 0x75, 0x3d, 
	0x1, 0x7b, 0x14, 0xe4, 0xfd, 0x7f, 0xe0, 0x12, 0x17, 0x9d, 0x7f, 0x14, 0x12, 0x1d, 0x6c, 0x7b, 
	0x4, 0xe4, 0xfd, 0x7f, 0x40, 0x12, 0xa, 0x2a, 0x7f, 0x5f, 0x12, 0x1f, 0xba, 0x12, 0xa, 0x31, 
	0x2, 0x14, 0xaa, 0x12, 0x1f, 0xee, 0xd2, 0xa1, 0xc2, 0x91, 0xe4, 0xf5, 0x39, 0x75, 0x38, 0x1, 
	0xf5, 0x37, 0xf5, 0x36, 0xf5, 0x3d, 0x75, 0x3c, 0x1f, 0xf5, 0x3b, 0xf5, 0x3a, 0x7b, 0x10, 0xfd, 
	0x7f, 0xe0, 0x12, 0x11, 0x3a, 0x12, 0x1d, 0x59, 0x12, 0x1c, 0x83, 0xd2, 0xaf, 0x7f, 0x14, 0x12, 
	0x1d, 0x6c, 0x12, 0xd, 0xa5, 0x12, 0x1f, 0xef, 0x12, 0x1a, 0x98, 0xef, 0x70, 0x3, 0x30, 0x7, 
	0x3, 0x12, 00, 0x1e, 0x12, 0x1a, 0xfb, 0x12, 0x1c, 0xdf, 0xef, 0x60, 0x3, 0x12, 00, 0x1e, 
	0x12, 0x17, 0x35, 0x12, 0x1c, 0x39, 0x30, 0x7, 0x5, 0xc2, 0x7, 0x12, 0x1c, 0xf5, 0x12, 0xc, 
	0x9e, 0x30, 00, 0x8, 0xc2, 00, 0x12, 0x1e, 0x2b, 0x12, 0x1b, 0xec, 0x30, 0x1, 0xe7, 0xc2, 
	0x1, 0x12, 0x1d, 0xfd, 0x12, 0x1c, 0x9a, 0xa2, 0x81, 0xb3, 0x92, 0x91, 0x80, 0xd8, 0x8f, 0x37, 
	0x12, 0xb, 0x79, 0xaf, 0x37, 0x12, 0x12, 0xcb, 0x75, 0x3d, 00, 0x75, 0x3e, 0x40, 0x7b, 0x40, 
	0x7d, 0x12, 0x12, 0xf, 0xb, 0x12, 0x1f, 0x7c, 0x75, 0x3d, 0xf, 0x75, 0x3e, 0xc0, 0x7b, 0x40, 
	0x7a, 0x2, 0x7d, 0x16, 0xaf, 0x37, 0x12, 0x19, 0x52, 0x75, 0x3d, 0x60, 0x75, 0x3e, 00, 0x7b, 
	00, 0x7a, 0x40, 0x7d, 0x14, 0xaf, 0x37, 0x12, 0x19, 0x52, 0x75, 0x3d, 00, 0x75, 0x3e, 0x38, 
	0x7b, 0x30, 0x7d, 0x18, 0x12, 0xf, 0xb, 0x12, 0x1f, 0x6a, 0x75, 0x3d, 0xe0, 0x12, 0x1a, 0x58, 
	0xaf, 0x37, 0x12, 0x19, 0x52, 0xaf, 0x37, 0x12, 0x1f, 0x85, 0x75, 0x3d, 0xe0, 0x75, 0x3e, 0xe, 
	0x7b, 0x2, 0x7a, 0x80, 0x7d, 0x18, 0xaf, 0x37, 0x2, 0x19, 0x52, 0x7a, 00, 0xaf, 0x37, 0x12, 
	0x19, 0x52, 0xaf, 0x37, 0x22, 0x8f, 0x35, 0x8e, 0x34, 0x8d, 0x33, 0x8c, 0x32, 0xe4, 0xf5, 0x40, 
	0xe4, 0xff, 0xfe, 0xfd, 0xfc, 0xab, 0x35, 0xaa, 0x34, 0xa9, 0x33, 0xa8, 0x32, 0xd3, 0x12, 0x4, 
	0xcf, 0x40, 0x37, 0xe4, 0x7b, 0xa, 0xfa, 0xf9, 0xf8, 0xaf, 0x35, 0xae, 0x34, 0xad, 0x33, 0xac, 
	0x32, 0x12, 0x4, 0x30, 0xaf, 0x3, 0x74, 0x36, 0x25, 0x40, 0xf8, 0xa6, 0x7, 0xaf, 0x35, 0xae, 
	0x34, 0xad, 0x33, 0xac, 0x32, 0xe4, 0x7b, 0xa, 0xfa, 0xf9, 0xf8, 0x12, 0x4, 0x30, 0x8f, 0x35, 
	0x8e, 0x34, 0x8d, 0x33, 0x8c, 0x32, 0x5, 0x40, 0x80, 0xb6, 0xe5, 0x40, 0x70, 0x5, 0x7f, 0x30, 
	0x2, 0x1d, 0x92, 0x15, 0x40, 0xe5, 0x40, 0xf4, 0x60, 0x10, 0x74, 0x36, 0x25, 0x40, 0xf8, 0xe6, 
	0x24, 0x30, 0xff, 0x12, 0x1d, 0x92, 0x15, 0x40, 0x80, 0xeb, 0x22, 0x7b, 0xff, 0x7a, 0x1e, 0x79, 
	0x49, 0x12, 0x19, 0xe9, 0x12, 0x1f, 0x39, 0x7b, 0xff, 0x7a, 0x9, 0x79, 0x56, 0x12, 0x19, 0xe9, 
	0x7b, 0x18, 0xe4, 0xfd, 0x7f, 0xe0, 0x12, 0xa, 0x16, 0x12, 0x1c, 0x20, 0x12, 0x1f, 0x39, 0x7b, 
	0xff, 0x7a, 0x9, 0x79, 0x60, 0x12, 0x19, 0xe9, 0xe4, 0xf5, 0x24, 0xaf, 0x24, 0x12, 0x1a, 0xda, 
	0x7f, 0x3, 0x12, 0x1d, 0x31, 0xaf, 0x24, 0x12, 0x1e, 0x84, 0x7f, 0x2, 0x12, 0x1d, 0x31, 0x7f, 
	0x1, 0x12, 0x1d, 0x31, 0xaf, 0x24, 0x12, 0x1f, 0xc8, 0xe4, 0xfc, 0xfd, 0xfe, 0x12, 0xf, 0x15, 
	0x7f, 0x3, 0x12, 0x1d, 0x31, 0xaf, 0x24, 0x12, 0x1f, 0xcf, 0x12, 0x1a, 0xda, 0x12, 0x1f, 0x39, 
	0x5, 0x24, 0xe5, 0x24, 0xc3, 0x94, 0x7, 0x40, 0xc2, 0x22, 0x8f, 0x32, 0x8d, 0x33, 0x7b, 0x24, 
	0x7f, 0x20, 0x12, 0xa, 0x2a, 0x7f, 0x6, 0x12, 0x1b, 0x5f, 0x12, 0xa, 0x31, 0xe5, 0x32, 0x24, 
	0x80, 0xfb, 0x75, 0x3c, 0x1b, 0x75, 0x3d, 0x1, 0xe4, 0xfd, 0x7f, 0x40, 0x12, 0x17, 0x9d, 0x7f, 
	0x2, 0x12, 0x1f, 0x1b, 0xaf, 0x33, 0x12, 0x1a, 0x77, 0x7f, 0xa, 0x12, 0x1f, 0xa5, 0x12, 0x1e, 
	0xef, 0x40, 0x5, 0x12, 0x1f, 0xdd, 0x50, 0xf6, 0x12, 0x1f, 0xdd, 0x40, 0x13, 0xad, 0x33, 0xaf, 
	0x32, 0x12, 0x13, 0x80, 0xef, 0x70, 0x4, 0xf5, 0x34, 0x80, 0x8, 0x75, 0x34, 0x1, 0x80, 0x3, 
	0x75, 0x34, 0x2, 0xe5, 0x32, 0x24, 0x80, 0xfb, 0x75, 0x3c, 0x1b, 0xe4, 0xf5, 0x3d, 0xfd, 0x7f, 
	0x40, 0x12, 0x17, 0x9d, 0xaf, 0x34, 0x22, 0x8f, 0xa, 0xc3, 0xed, 0x94, 0x40, 0xec, 0x94, 00, 
	0x50, 0x16, 0xc3, 0x74, 0x40, 0x9d, 0xff, 0xe4, 0x9c, 0xfe, 0xef, 0x78, 0x2, 0xce, 0xc3, 0x13, 
	0xce, 0x13, 0xd8, 0xf9, 0xf5, 0xb, 0x80, 0x1c, 0xed, 0x24, 0x3, 0xff, 0xe4, 0x3c, 0xfe, 0xef, 
	0x78, 0x2, 0xce, 0xc3, 0x13, 0xce, 0x13, 0xd8, 0xf9, 0x30, 0xe0, 0x5, 0x75, 0xb, 0x1, 0x80, 
	0x3, 0xe4, 0xf5, 0xb, 0xc3, 0xed, 0x94, 0x40, 0xec, 0x94, 00, 0x40, 0xe, 0xae, 0x4, 0xaf, 
	0x5, 0xe4, 0xfc, 0xfd, 0x78, 0x10, 0x12, 0x4, 0xf3, 0x80, 0x8, 0x7f, 00, 0x7e, 00, 0x7d, 
	0x40, 0x7c, 00, 0x12, 0x15, 0xa8, 0x7f, 0x20, 0x7e, 0x5, 0x7d, 00, 0x7c, 00, 0x2, 0x15, 
	0xa8, 0x8e, 0x23, 0x8f, 0x24, 0x8c, 0x25, 0x8d, 0x26, 0x7f, 0xa, 0x12, 0x1d, 0x45, 0x12, 0x1f, 
	0x57, 0x40, 0x52, 0x12, 0x1e, 0x58, 0xe4, 0xff, 0x12, 0x1d, 0xc8, 0x7f, 0x2, 0x12, 0x1a, 0x32, 
	0xe5, 0x23, 0xff, 0x12, 0x1a, 0x32, 0xaf, 0x24, 0x12, 0x1a, 0x32, 0xe4, 0xf5, 0x27, 0x12, 0x1f, 
	0x57, 0x40, 0x24, 0x12, 0x1e, 0xe4, 0x50, 0xfb, 0x12, 0x1e, 0xd, 0x12, 0x1a, 0x32, 0x5, 0x24, 
	0xe5, 0x24, 0x70, 0x2, 0x5, 0x23, 0xe5, 0x26, 0x15, 0x26, 0x70, 0x2, 0x15, 0x25, 0x5, 0x27, 
	0xe5, 0x27, 0xc3, 0x94, 0x20, 0x40, 0xd7, 0x7f, 0x1, 0x12, 0x1d, 0xc8, 0x12, 0x1c, 0xc8, 0xef, 
	0x30, 0xe0, 0xab, 0x80, 0xf7, 0xe4, 0xff, 0x2, 0x1d, 0x45, 0x8f, 0x33, 0x8d, 0x34, 0x8b, 0x35, 
	0xe5, 0x39, 0x55, 0x3d, 0xf5, 0x39, 0xe5, 0x38, 0x55, 0x3c, 0xf5, 0x38, 0xe5, 0x37, 0x55, 0x3b, 
	0xf5, 0x37, 0xe5, 0x36, 0x55, 0x3a, 0xf5, 0x36, 0x12, 0xa, 0x16, 0xa8, 0x4, 0xa9, 0x5, 0xaa, 
	0x6, 0xab, 0x7, 0xaf, 0x3d, 0xae, 0x3c, 0xad, 0x3b, 0xac, 0x3a, 0x12, 0x4, 0xc2, 0xeb, 0x5f, 
	0xff, 0xea, 0x5e, 0xfe, 0xe9, 0x5d, 0xfd, 0xe8, 0x5c, 0xfc, 0xe5, 0x39, 0x4f, 0xf5, 0x39, 0xe5, 
	0x38, 0x4e, 0xf5, 0x38, 0xe5, 0x37, 0x4d, 0xf5, 0x37, 0xe5, 0x36, 0x4c, 0xf5, 0x36, 0xab, 0x35, 
	0xad, 0x34, 0xaf, 0x33, 0x12, 0xa, 0x2a, 0xaf, 0x39, 0xae, 0x38, 0xad, 0x37, 0xac, 0x36, 0x2, 
	0xa, 0x31, 0x8f, 0x32, 0x8d, 0x33, 0x7b, 0x19, 0xad, 0x32, 0x7f, 0x20, 0x12, 0xa, 0x2a, 0x12, 
	0x1b, 0x5d, 0x12, 0xa, 0x31, 0xe5, 0x33, 0x44, 0x10, 0xfd, 0xaf, 0x32, 0x12, 0x12, 0x6c, 0xef, 
	0xf5, 0x37, 0xee, 0xf5, 0x36, 0xed, 0x44, 0x1, 0xf5, 0x35, 0xec, 0x44, 0x10, 0xf5, 0x34, 0x12, 
	0x11, 0xfd, 0xe5, 0x37, 0x44, 0x30, 0xff, 0xe5, 0x36, 0xfe, 0xe5, 0x35, 0xfd, 0xe5, 0x34, 0x44, 
	0x20, 0xfc, 0x12, 0xa, 0x31, 0x12, 0x11, 0xfd, 0xaf, 0x37, 0xae, 0x36, 0xad, 0x35, 0xac, 0x34, 
	0x12, 0xa, 0x31, 0xe5, 0x32, 0xb4, 0x6, 0x4, 0xff, 0x12, 0x13, 0xd0, 0x22, 0xe4, 0xfb, 0xad, 
	0x32, 0x7f, 0x20, 0x12, 0xa, 0x2a, 0x22, 0x7f, 0xa, 0x12, 0x1d, 0x45, 0xe4, 0xff, 0xfe, 0x12, 
	0x19, 0x78, 0xef, 0x7f, 00, 0xfe, 0xc0, 0x7, 0xc0, 0x6, 0xf, 0x7e, 00, 0x12, 0x19, 0x78, 
	0xef, 0xfd, 0xd0, 0xe0, 0xf5, 0x1b, 0xd0, 0xe0, 0x4d, 0xf5, 0x1c, 0x24, 0x1, 0xff, 0xe4, 0x35, 
	0x1b, 0xfe, 0x12, 0x19, 0x78, 0x8f, 0x1d, 0xe4, 0xff, 0x12, 0x1d, 0x45, 0xe5, 0x1c, 0x15, 0x1c, 
	0x70, 0x2, 0x15, 0x1b, 0xe4, 0xff, 0xfe, 0xc3, 0xef, 0x95, 0x1c, 0xee, 0x95, 0x1b, 0x50, 0x11, 
	0x8f, 0x82, 0x8e, 0x83, 0xe4, 0x93, 0x25, 0x1d, 0xf5, 0x1d, 0xf, 0xbf, 00, 0x1, 0xe, 0x80, 
	0xe6, 0xe5, 0x1d, 0x70, 0x4, 0x7f, 0x1, 0x80, 0x2, 0x7f, 00, 0x22, 0xef, 0x64, 0x6, 0xef, 
	0x64, 0x6, 0x70, 0x24, 0xbd, 0x10, 0x8, 0x7f, 0x43, 0x7e, 0x4, 0x7d, 0x4, 0xfc, 0x22, 0xed, 
	0x70, 0x4, 0x7f, 0x43, 0x80, 0x1e, 0xbd, 0x11, 0x4, 0x7f, 0x42, 0x80, 0x22, 0xbd, 0x1, 0x4, 
	0x7f, 0x42, 0x80, 0x27, 0x7f, 0x81, 0x80, 0x2c, 0xbd, 0x10, 0x4, 0x7f, 0x44, 0x80, 0x10, 0xed, 
	0x70, 0x8, 0x7f, 0x44, 0x7e, 0x4, 0xfd, 0x7c, 0x80, 0x22, 0xbd, 0x11, 0x9, 0x7f, 0x44, 0x7e, 
	0x4, 0x7d, 0x4, 0x7c, 00, 0x22, 0xbd, 0x1, 0x9, 0x7f, 0x44, 0x7e, 0x4, 0x7d, 00, 0x7c, 
	0x80, 0x22, 0x7f, 0x84, 0x7e, 0x1, 0x7d, 0x6, 0x7c, 00, 0x22, 0x8f, 0x38, 0x12, 0x1f, 0x7c, 
	0x12, 0x1a, 0x62, 0x12, 0x13, 0x13, 0x12, 0x1f, 0x73, 0x7a, 0xb6, 0x12, 0x13, 0x1d, 0x75, 0x3d, 
	0xff, 0x75, 0x3e, 0x7, 0x7b, 0x3, 0x7a, 00, 0x7d, 0x12, 0xaf, 0x38, 0x12, 0x19, 0x52, 0x75, 
	0x3d, 00, 0x75, 0x3e, 0xff, 0x7b, 0xa2, 0x7a, 00, 0x7d, 0x11, 0xaf, 0x38, 0x12, 0x19, 0x52, 
	0x7a, 0x96, 0x12, 0x13, 0x1d, 0xaf, 0x38, 0x12, 0x1f, 0x7c, 0x12, 0x1a, 0x6d, 0x12, 0x13, 0x13, 
	0x2, 0x1f, 0x85, 0x7d, 0x8, 0xaf, 0x38, 0x12, 0x19, 0x52, 0xaf, 0x38, 0x22, 0x7b, 0x8a, 0x7d, 
	0x10, 0xaf, 0x38, 0x12, 0x17, 0xff, 0x22, 0x8f, 0x3f, 0x8d, 0x40, 0x12, 0x1f, 0xc8, 0x8f, 0x45, 
	0xaf, 0x3f, 0x12, 0x1f, 0xcf, 0x8f, 0x46, 0x7b, 0x1, 0xad, 0x45, 0x7f, 0x60, 0x12, 0xa, 0x2a, 
	0xe5, 0x46, 0x75, 0xf0, 0x20, 0xa4, 0xff, 0xe5, 0xf0, 0x44, 0x4, 0xfc, 0xef, 0x45, 0x40, 0xfd, 
	0xe4, 0xff, 0xfe, 0x12, 0xa, 0x47, 0x12, 0xa, 0x31, 0xaf, 0x45, 0x12, 0x1b, 0x3d, 0x7b, 0x2, 
	0xad, 0x45, 0x7f, 0x60, 0x12, 0xa, 0x16, 0x8f, 0x44, 0x8e, 0x43, 0x8d, 0x42, 0x8c, 0x41, 0xe5, 
	0x42, 0x54, 0x1, 0xfd, 0xed, 0x70, 0x5, 0xae, 0x43, 0xaf, 0x44, 0x22, 0xe4, 0xfe, 0xff, 0x22, 
	0x8f, 0x35, 0x8d, 0x36, 0xe4, 0xf5, 0x38, 0xfb, 0x7a, 00, 0x79, 0x39, 0x12, 0x14, 0xec, 0xef, 
	0x64, 0x80, 0x4e, 0x60, 0x3, 0x75, 0x38, 0x1, 0xe5, 0x39, 0x65, 0x35, 0x60, 0x3, 0x75, 0x38, 
	0x1, 0xe5, 0x38, 0x70, 0x25, 0xf5, 0x37, 0x12, 0x1d, 0x9, 0xa8, 0x4, 0xa9, 0x5, 0xaa, 0x6, 
	0xab, 0x7, 0xe4, 0x25, 0x36, 0x12, 0x1b, 0x66, 0xc3, 0x12, 0x4, 0xcf, 0x60, 0x3, 0x75, 0x38, 
	0x1, 0x5, 0x37, 0xe5, 0x37, 0xc3, 0x94, 0x1f, 0x40, 0xdd, 0x12, 0x1d, 0x1d, 0xaf, 0x38, 0x22, 
	0x8f, 0x38, 0x75, 0x3c, 0x5, 0x12, 0x14, 0xf, 0xe5, 0x38, 0x64, 0x6, 0x60, 0x7, 0xe5, 0x38, 
	0xb4, 0x6, 0x8, 0x80, 0x6, 0x75, 0x3c, 0x3, 0x12, 0x14, 0xf, 0xe5, 0x38, 0xb4, 0x6, 0xe, 
	0x75, 0x3c, 0x2, 0x75, 0x3d, 0x1, 0x7b, 0x19, 0xfd, 0x7f, 0x20, 0x12, 0x17, 0x9d, 0x7b, 0x5, 
	0xe4, 0xfd, 0x7f, 0xe0, 0x12, 0xa, 0x2a, 0x7f, 0x20, 0x12, 0x1f, 0xba, 0x2, 0xa, 0x31, 0x75, 
	0x3d, 0x1, 0x7b, 0x19, 0xad, 0x38, 0x7f, 0x20, 0x12, 0x17, 0x9d, 0x22, 0x8f, 0x35, 0x8d, 0x36, 
	0xe4, 0xfb, 0xfa, 0xfd, 0x12, 0x17, 0xff, 0xe5, 0x36, 0x54, 0x3, 0xff, 0x70, 0x6, 0xf5, 0x37, 
	0xf5, 0x38, 0x80, 0x14, 0xbf, 0x1, 0x8, 0x75, 0x37, 0x20, 0x75, 0x38, 00, 0x80, 0x9, 0xbf, 
	0x2, 0x6, 0x75, 0x37, 00, 0x75, 0x38, 0x40, 0xe5, 0x36, 0x30, 0xe4, 0x3, 0x43, 0x37, 0x1, 
	0xe5, 0x36, 0x30, 0xe7, 0x3, 0x43, 0x37, 0x40, 0xab, 0x38, 0xaa, 0x37, 0xe4, 0xfd, 0xaf, 0x35, 
	0x2, 0x17, 0xff, 0x8f, 0x27, 0x12, 0x1f, 0x73, 0x7b, 0xf0, 0x7a, 0xaf, 0x7d, 0x10, 0xaf, 0x27, 
	0x12, 0x17, 0xff, 0x7d, 0x11, 0xaf, 0x27, 0x12, 0x13, 0x27, 0xee, 0xc4, 0xf8, 0x54, 0xf0, 0xc8, 
	0xef, 0xc4, 0x54, 0xf, 0x48, 0x54, 0x1f, 0xff, 0xc3, 0x94, 0x10, 0x40, 0x6, 0xef, 0xd3, 0x94, 
	0x14, 0x40, 0x12, 0xaf, 0x27, 0x12, 0x1f, 0x7c, 0x75, 0x3d, 0x3, 0x12, 0x1a, 0x70, 0x7d, 0xc, 
	0xaf, 0x27, 0x12, 0x19, 0x52, 0xaf, 0x27, 0x2, 0x1f, 0x85, 0x75, 0x27, 0x1, 0x7b, 0x6, 0xe4, 
	0xfd, 0x7f, 0x40, 0x12, 0xa, 0x2a, 0x7f, 0x80, 0x7e, 0x1, 0x7d, 00, 0x7c, 00, 0x12, 0xa, 
	0x31, 0x7b, 0x7, 0xe4, 0xfd, 0x7f, 0x40, 0x12, 0xa, 0x2a, 0xe4, 0x25, 0x27, 0xff, 0xe4, 0x33, 
	0xfe, 0xe4, 0x33, 0xfd, 0xe4, 0x34, 0xc2, 0xfc, 0x12, 0xa, 0x31, 0x7f, 0x1, 0x7e, 0x26, 0x12, 
	0x16, 0xcb, 0x5, 0x27, 0xe5, 0x27, 0xd3, 0x94, 0xf, 0x40, 0xc2, 0x22, 0x8b, 0x3a, 0x8a, 0x3b, 
	0x89, 0x3c, 0xe4, 0xf5, 0x8, 0xf5, 0x9, 0x12, 0x15, 0x20, 0xe5, 0x3d, 0x54, 0x3f, 0xf5, 0x41, 
	0x85, 0x3e, 0x42, 0xe5, 0x3b, 0x45, 0x3c, 0x45, 0x3a, 0x60, 0x10, 0xe5, 0x40, 0x54, 0x7, 0xab, 
	0x3a, 0xaa, 0x3b, 0xa9, 0x3c, 0x12, 0x3, 0x93, 0x12, 0x15, 0x20, 0xae, 0x41, 0xaf, 0x42, 0x22, 
	0x12, 0x1d, 0x9, 0x8f, 0x40, 0x8e, 0x3f, 0x8d, 0x3e, 0x8c, 0x3d, 0x22, 0xc0, 0xe0, 0xc0, 0xd0, 
	0x75, 0xd0, 0x10, 0x30, 0x98, 0x29, 0xaf, 0x99, 0xc2, 0x98, 0xe5, 0x1e, 0x24, 0x1, 0xfd, 0xe4, 
	0x33, 0xfc, 0xe5, 0x1f, 0x7a, 00, 0xb5, 0x15, 0x8, 0xea, 0xb5, 0x14, 0x4, 0xd2, 0x5, 0x80, 
	0xe, 0x74, 0x80, 0x25, 0x1e, 0xf8, 0xa6, 0x17, 0xe5, 0x1e, 0x4, 0x54, 0x1f, 0xf5, 0x1e, 0x30, 
	0x99, 0x4, 0xc2, 0x99, 0xd2, 0x6, 0xd0, 0xd0, 0xd0, 0xe0, 0x32, 0x8b, 0x23, 0x8a, 0x24, 0x89, 
	0x25, 0xab, 0x26, 0xaa, 0x27, 0xa9, 0x28, 0x12, 0x3, 0x7a, 0xf5, 0x29, 0x70, 0x2, 0xff, 0x22, 
	0x74, 0x1, 0x25, 0x28, 0xf5, 0x28, 0xe4, 0x35, 0x27, 0xf5, 0x27, 0xab, 0x23, 0x5, 0x25, 0xe5, 
	0x25, 0xaa, 0x24, 0x70, 0x2, 0x5, 0x24, 0x14, 0xf9, 0x12, 0x3, 0x7a, 0xff, 0x12, 0x1c, 0x52, 
	0xef, 0x65, 0x29, 0x60, 0xcc, 0x7f, 0x1, 0x22, 0x8f, 0x3a, 0x8e, 0x39, 0x8d, 0x38, 0x8c, 0x37, 
	0x20, 0x8, 0x30, 0x7b, 0xc0, 0xad, 0xa, 0x7f, 0x20, 0x12, 0xa, 0x2a, 0xaf, 0x3a, 0xae, 0x39, 
	0xad, 0x38, 0xac, 0x37, 0x12, 0xa, 0x31, 0xe4, 0x78, 0xa7, 0xf6, 0x7b, 0xc8, 0xad, 0xa, 0x7f, 
	0x20, 0x12, 0xa, 0x16, 0xee, 0x30, 0xe0, 0xb, 0x78, 0xa7, 0x6, 0xe6, 0xd3, 0x94, 0x64, 0x40, 
	0xea, 0xd2, 0x8, 0x22, 0x12, 0x19, 0x2c, 0x8f, 0x2b, 0xe4, 0xf5, 0x2a, 0xaf, 0x2a, 0x12, 0xa, 
	0x68, 0xef, 0xf4, 0xff, 0xef, 0x55, 0x2b, 0xff, 0xe4, 0x8f, 0x39, 0xf5, 0x38, 0xf5, 0x37, 0xf5, 
	0x36, 0xe5, 0x2a, 0x24, 0x80, 0xfb, 0xe4, 0x75, 0x3d, 0x5f, 0xf5, 0x3c, 0xf5, 0x3b, 0xf5, 0x3a, 
	0xfd, 0x7f, 0x40, 0x12, 0x11, 0x3a, 0x5, 0x2a, 0xe5, 0x2a, 0xc3, 0x94, 0x7, 0x40, 0xcd, 0x22, 
	0x12, 0x1e, 0xe4, 0x50, 0x35, 0x12, 0x1e, 0xd, 0xef, 0x64, 0xa, 0x60, 0x2d, 0xbf, 0x8, 0xb, 
	0xe5, 0x18, 0xd3, 0x94, 00, 0x40, 0x23, 0x15, 0x18, 0x80, 0x1f, 0xe5, 0x18, 0xc3, 0x94, 0x1e, 
	0x50, 0xa, 0xae, 0x18, 0x5, 0x18, 0x74, 0x62, 0x2e, 0xf8, 0xa6, 0x7, 0xbf, 0xd, 0xb, 0xe5, 
	0x18, 0xb4, 0x1e, 0x4, 0x78, 0x7f, 0x76, 0xd, 0xd3, 0x22, 0xc3, 0x22, 0x8f, 0x25, 0x8d, 0x26, 
	0xe5, 0x25, 0x64, 0x6, 0x60, 0x2d, 0xe5, 0x26, 0x54, 0x3, 0xff, 0xbf, 0x2, 0x5, 0xaf, 0x25, 
	0x12, 0x14, 0x63, 0xe5, 0x26, 0x54, 0x3, 0x70, 0x1a, 0xaf, 0x25, 0x12, 0x1f, 0x7c, 0x12, 0x1a, 
	0x55, 0xaf, 0x25, 0x12, 0x19, 0x52, 0x12, 0x1a, 0x55, 0xaf, 0x25, 0x12, 0x19, 0x52, 0xaf, 0x25, 
	0x12, 0x1f, 0x85, 0x22, 0xaf, 0xb, 0x15, 0xb, 0xef, 0xd3, 0x94, 00, 0x40, 0xd, 0x7f, 0x55, 
	0x7e, 0x55, 0x7d, 0x55, 0x7c, 0x55, 0x12, 0x15, 0xa8, 0x80, 0xe9, 0x30, 0x8, 0x9, 0x75, 0x3c, 
	0x1, 0x12, 0x16, 0xbe, 0xc2, 0x8, 0x22, 0xe4, 0xf5, 0x3c, 0x12, 0x16, 0xbe, 0x22, 0x75, 0x3d, 
	0x1, 0x7b, 0xc4, 0xad, 0xa, 0x7f, 0x20, 0x12, 0x17, 0x9d, 0x22, 0x8e, 0x28, 0x8f, 0x29, 0x7b, 
	0xb0, 0xe4, 0xfd, 0x7f, 0x40, 0x12, 0xa, 0x2a, 0xae, 0x28, 0xaf, 0x29, 0xe4, 0xfc, 0xfd, 0x12, 
	0xa, 0x31, 0x7f, 0xa, 0x12, 0x1f, 0xa5, 0x7b, 0xb0, 0xe4, 0xfd, 0x7f, 0x40, 0x12, 0xa, 0x16, 
	0xef, 0x54, 0x7, 0x60, 0xa, 0x12, 0x1f, 0xdd, 0x50, 0xed, 0x7f, 0x4, 0x12, 0x1f, 0xe5, 0x22, 
	0x8f, 0x28, 0x8d, 0x29, 0xe5, 0x29, 0xf4, 0x60, 0x29, 0x7b, 0x10, 0xad, 0x28, 0x7f, 0x20, 0x12, 
	0xa, 0x2a, 0x7f, 00, 0x7e, 0x6, 0x7d, 00, 0x7c, 00, 0x12, 0xa, 0x31, 0xad, 0x29, 0xaf, 
	0x28, 0x12, 0x7, 0x66, 0xe4, 0xfd, 0xaf, 0x28, 0x12, 00, 0x26, 0x7d, 0x1, 0xaf, 0x28, 0x12, 
	0x1e, 0xd7, 0x2, 0x15, 0xe4, 0xe4, 0xf5, 0x22, 0xe5, 0x22, 0x24, 0x80, 0xfb, 0xe4, 0xfd, 0x7f, 
	0x40, 0x12, 0xa, 0x2a, 0x7f, 00, 0x12, 0x1f, 0xba, 0x12, 0xa, 0x31, 0xe5, 0x22, 0xb4, 0x6, 
	0x5, 0x7d, 0x32, 0xff, 0x80, 0x4, 0x7d, 0xff, 0xaf, 0x22, 0x12, 0x1d, 0xa4, 0x5, 0x22, 0xe5, 
	0x22, 0xc3, 0x94, 0x7, 0x40, 0xd2, 0x2, 0x15, 0xe4, 0x8b, 0x2b, 0x8a, 0x2c, 0x89, 0x2d, 0xe4, 
	0xf5, 0x2e, 0xab, 0x2b, 0x5, 0x2d, 0xe5, 0x2d, 0xaa, 0x2c, 0x70, 0x2, 0x5, 0x2c, 0x14, 0xf9, 
	0x12, 0x3, 0x7a, 0xff, 0x12, 0x1c, 0x7, 0xe5, 0x2e, 0xc3, 0x94, 0x5, 0x50, 0x5, 0x7f, 0x2d, 
	0x12, 0x1f, 0xe8, 0x5, 0x2e, 0xe5, 0x2e, 0xc3, 0x94, 0x6, 0x40, 0xd6, 0x22, 0x8f, 0x39, 0x8d, 
	0x3a, 0x8b, 0x3b, 0x12, 0xa, 0x16, 0x8f, 0x41, 0x8e, 0x40, 0x8d, 0x3f, 0x8c, 0x3e, 0x7b, 00, 
	0x7a, 00, 0x79, 0x3e, 0xad, 0x3d, 0xaf, 0x3c, 0x12, 0xa, 0x97, 0xab, 0x3b, 0xad, 0x3a, 0xaf, 
	0x39, 0x12, 0xa, 0x2a, 0xaf, 0x41, 0xae, 0x40, 0xad, 0x3f, 0xac, 0x3e, 0x2, 0xa, 0x31, 00, 
	00, 0x10, 0x10, 0x80, 00, 00, 00, 00, 0xe, 00, 0x15, 00, 0x10, 00, 0x15, 00, 
	0x12, 00, 0x15, 00, 0x14, 00, 0x15, 00, 00, 0x10, 0xc, 0x80, 00, 00, 0x1, 00, 
	0xa, 00, 0x18, 00, 0xb, 00, 0x18, 00, 0xc, 00, 0x18, 00, 0xd, 00, 0x18, 0x8f, 
	0x3f, 0x8d, 0x40, 0x8a, 0x41, 0x8b, 0x42, 0x12, 0x1f, 0xc8, 0x8f, 0x43, 0xaf, 0x3f, 0x12, 0x1f, 
	0xcf, 0x8f, 0x44, 0x7b, 0x1, 0xad, 0x43, 0x7f, 0x60, 0x12, 0xa, 0x2a, 0xab, 0x44, 0xad, 0x40, 
	0xaf, 0x42, 0xae, 0x41, 0x12, 0x1f, 0x25, 0x12, 0xa, 0x31, 0xaf, 0x43, 0x2, 0x1b, 0x3d, 0x8f, 
	0x37, 0x7b, 00, 0x7a, 0x80, 0xe4, 0xfd, 0x12, 0x17, 0xff, 0x7f, 0x2, 0x12, 0x1d, 0x6c, 0xe4, 
	0xf5, 0x38, 0xe4, 0xfd, 0xaf, 0x37, 0x12, 0x13, 0x27, 0xee, 0x30, 0xe7, 0x10, 0x5, 0x38, 0xe5, 
	0x38, 0xd3, 0x94, 0xc8, 0x50, 0x7, 0x7f, 0x1, 0x12, 0x1d, 0x6c, 0x80, 0xe5, 0x22, 0xc1, 0x5, 
	0xc1, 0x4, 0x1, 0x1e, 00, 0x1, 0x1f, 00, 0xc1, 00, 0xc1, 0x1, 0x1, 0xe, 00, 0xc1, 
	0x2, 0xc1, 0x3, 0x1, 0xc, 0xa, 0x1, 0xd, 0x64, 0xc1, 0x7, 0x1, 0xf, 00, 0xc1, 0x8, 
	0x1, 0x18, 00, 0x1, 0x4d, 00, 0x1, 0x51, 00, 00, 0x8f, 0x2a, 0x8e, 0x29, 0x8d, 0x28, 
	0x8c, 0x27, 0x7b, 0xd0, 0xe4, 0xfd, 0x7f, 0x40, 0x12, 0xa, 0x2a, 0xaf, 0x2a, 0xae, 0x29, 0xad, 
	0x28, 0xac, 0x27, 0x12, 0xa, 0x31, 0x7b, 0xd0, 0xe4, 0xfd, 0x7f, 0x40, 0x12, 0xa, 0x16, 0xef, 
	0x54, 0x3, 0x70, 0xf2, 0x22, 0xe4, 0xfe, 0xee, 0x90, 00, 0x40, 0x93, 0xfc, 0xed, 0x2e, 0xf8, 
	0xa6, 0x4, 0xe, 0xbe, 0x6, 0xf1, 0xed, 0x24, 0x5, 0xf8, 0xe6, 0x2f, 0xf6, 0xc3, 0x9f, 0x50, 
	0xd, 0xed, 0x24, 0x4, 0xf8, 0x6, 0xe6, 0x70, 0x5, 0xed, 0x24, 0x3, 0xf8, 0x6, 0x22, 0x8f, 
	0x32, 0x8d, 0x33, 0x8b, 0x34, 0x12, 0x1f, 0x4d, 0xaf, 0x33, 0x12, 0x1f, 0x4d, 0x7f, 0x5, 0x12, 
	0x1f, 0x1b, 0xe5, 0x34, 0x44, 0x10, 0xfd, 0xaf, 0x32, 0x12, 0x14, 0x1c, 0xe5, 0x34, 0x44, 0x10, 
	0xfd, 0xaf, 0x33, 0x2, 0x14, 0x1c, 0x8f, 0x32, 0xed, 0x70, 0x2, 0x80, 0x1, 0xe4, 0x75, 0x39, 
	0x86, 0x75, 0x38, 0x2, 0xf5, 0x37, 0xf5, 0x36, 0x75, 0x3d, 0xff, 0x75, 0x3c, 0xf, 0xf5, 0x3b, 
	0xf5, 0x3a, 0x7b, 0x2, 0xad, 0x32, 0x7f, 0x20, 0x12, 0x11, 0x3a, 0x22, 0xe4, 0xf5, 0x2c, 0xf5, 
	0x2d, 0xaf, 0x2d, 0x12, 00, 0x3, 0xef, 0x60, 0xd, 0x7b, 00, 0x7a, 00, 0x79, 0x2c, 0x7d, 
	0x1, 0xaf, 0x2d, 0x12, 0xa, 0x6f, 0x5, 0x2d, 0xe5, 0x2d, 0xc3, 0x94, 0x7, 0x40, 0xe2, 0xaf, 
	0x2c, 0x22, 0x8f, 0x39, 0x8d, 0x3a, 0x8a, 0x3b, 0x8b, 0x3c, 0x12, 0x13, 0x27, 0xe5, 0x3e, 0xf4, 
	0xfd, 0xe5, 0x3d, 0xf4, 0x5e, 0xfe, 0xef, 0x5d, 0x42, 0x3c, 0xee, 0x42, 0x3b, 0xab, 0x3c, 0xaa, 
	0x3b, 0xad, 0x3a, 0xaf, 0x39, 0x2, 0x17, 0xff, 0x8e, 0x48, 0x8f, 0x49, 0xe4, 0xff, 0x12, 0x1d, 
	0xc8, 0x7f, 0x3, 0x12, 0x1a, 0x32, 0xe5, 0x48, 0xff, 0x12, 0x1a, 0x32, 0xaf, 0x49, 0x12, 0x1a, 
	0x32, 0x12, 0x1b, 0x1c, 0x8f, 0x4a, 0x7f, 0x1, 0x12, 0x1d, 0xc8, 0xaf, 0x4a, 0x22, 0x12, 0x1c, 
	0x52, 0x74, 0x30, 0xd3, 0x9f, 0x50, 0xa, 0xef, 0x94, 0x39, 0x50, 0x5, 0xef, 0x24, 0xd0, 0xff, 
	0x22, 0x74, 0x41, 0xd3, 0x9f, 0x50, 0xa, 0xef, 0x94, 0x46, 0x50, 0x5, 0xef, 0x24, 0xc9, 0xff, 
	0x22, 0x7f, 0xff, 0x22, 0xc2, 0x8e, 0x75, 0x8b, 0x92, 0x75, 0x8d, 0xe6, 0xd2, 0x8e, 0xd2, 0x3, 
	0xd5, 0xc, 0x15, 0x75, 0xc, 0xa, 0xd2, 00, 0xd5, 0xd, 0x5, 0x75, 0xd, 0x64, 0xd2, 0x1, 
	0x30, 0x2, 0x5, 0xd5, 0xe, 0x2, 0xc2, 0x2, 0x32, 0x8b, 0x32, 0x8a, 0x33, 0x89, 0x34, 0xab, 
	0x32, 0xaa, 0x33, 0xa9, 0x34, 0x12, 0x3, 0x7a, 0x60, 0x13, 0x5, 0x34, 0xe5, 0x34, 0x70, 0x2, 
	0x5, 0x33, 0x14, 0xf9, 0x12, 0x3, 0x7a, 0xff, 0x12, 0x1d, 0x92, 0x80, 0xe2, 0x22, 0x78, 0x56, 
	0x12, 0x5, 0x6, 0xad, 0x7, 0x22, 0x78, 0x5a, 0x12, 0x5, 0x6, 0xab, 0x7, 0x22, 0x78, 0x52, 
	0x12, 0x5, 0x6, 0x78, 0x5, 0x2, 0x4, 0xf3, 0xa9, 0x1a, 0x7a, 00, 0x7b, 00, 0x75, 0x26, 
	0xff, 0x22, 0x8f, 0x29, 0xe4, 0xf5, 0x2a, 0xe5, 0x29, 0x54, 0x80, 0xff, 0x12, 0x1d, 0xda, 0xe5, 
	0x29, 0x25, 0xe0, 0xf5, 0x29, 0x12, 0x1f, 0x95, 0x12, 0x1f, 0x9d, 0x5, 0x2a, 0xe5, 0x2a, 0xc3, 
	0x94, 0x8, 0x40, 0xe3, 0x22, 0x75, 0x3d, 0x60, 0x75, 0x3e, 00, 0x7b, 00, 0x7a, 0x20, 0x7d, 
	0x14, 0x22, 0x75, 0x3d, 0x2, 0x75, 0x3e, 00, 0x7b, 00, 0x7a, 0x2, 0x22, 0x75, 0x3d, 0x2, 
	0x75, 0x3e, 00, 0xe4, 0xfb, 0xfa, 0x22, 0x8f, 0x35, 0x7d, 0x80, 0x7c, 00, 0x12, 0x10, 0x67, 
	0xe4, 0xf5, 0x36, 0xe4, 0x25, 0x35, 0x12, 0x1b, 0x66, 0x12, 0x15, 0xa8, 0x5, 0x36, 0xe5, 0x36, 
	0xc3, 0x94, 0x20, 0x40, 0xee, 0x2, 0x16, 0x94, 0x7b, 0x18, 0xe4, 0xfd, 0x7f, 0xe0, 0x12, 0xa, 
	0x16, 0xec, 0x54, 0xf, 0xfc, 0x7b, 0xe9, 0x7a, 0x50, 0x79, 0x39, 0x78, 0x7, 0xc3, 0x12, 0x4, 
	0xcf, 0x60, 0x3, 0x7f, 0x1, 0x22, 0x7f, 00, 0x22, 0x8f, 0x2c, 0x7f, 0x32, 0x12, 0x1f, 0xa5, 
	0x7b, 0xc, 0xe4, 0xfd, 0x7f, 0xa0, 0x12, 0xa, 0x16, 0xe5, 0x2c, 0x5f, 0xb5, 0x2c, 0x3, 0x7f, 
	0x1, 0x22, 0x12, 0x1f, 0xdd, 0x50, 0xe9, 0x7f, 00, 0x22, 0x8f, 0x32, 0xe5, 0x32, 0xd3, 0x94, 
	0x9, 0x40, 0xb, 0xe5, 0x32, 0x75, 0xf0, 0xa, 0x84, 0x12, 0x1f, 0x43, 0x80, 0x2, 0x7f, 0x20, 
	0x12, 0x1d, 0x92, 0xe5, 0x32, 0x12, 0x1f, 0x43, 0x2, 0x1d, 0x92, 0xe4, 0xf5, 0x4e, 0x7f, 0x3, 
	0x12, 0x1f, 0x1b, 0xe4, 0xf5, 0x22, 0xaf, 0x22, 0x12, 0x1e, 0xa0, 0x50, 0x5, 0xaf, 0x22, 0x12, 
	0xa, 0xe0, 0x5, 0x22, 0xe5, 0x22, 0xc3, 0x94, 0x7, 0x40, 0xeb, 0x22, 0xe4, 0xf5, 0x29, 0xe5, 
	0x2a, 0x25, 0xe0, 0xf5, 0x2a, 0x12, 0x1f, 0x95, 0x12, 0x1d, 0xec, 0xef, 0x42, 0x2a, 0x12, 0x1f, 
	0x9d, 0x5, 0x29, 0xe5, 0x29, 0xc3, 0x94, 0x8, 0x40, 0xe5, 0xaf, 0x2a, 0x22, 0x8f, 0x47, 0x7f, 
	0x5, 0x12, 0x1f, 0xa5, 0xe4, 0xfb, 0xad, 0x47, 0x7f, 0x60, 0x12, 0xa, 0x16, 0xef, 0x54, 0xb, 
	0x60, 0xa, 0x12, 0x1f, 0xdd, 0x50, 0xed, 0x7f, 0x2, 0x12, 0x1f, 0xe5, 0x22, 0x7f, 00, 0x7e, 
	00, 0x7d, 00, 0x7c, 00, 0x22, 0xff, 0xe4, 0x34, 0x56, 0xfe, 0xe4, 0x34, 0x34, 0xfd, 0xe4, 
	0x34, 0x12, 0xfc, 0x22, 0xff, 0x74, 0x3, 0xa8, 0x6, 0x8, 0x22, 0xad, 0x7, 0xe4, 0xf5, 0x39, 
	0xf5, 0x38, 0x75, 0x37, 0x1, 0x75, 0x36, 0x10, 0xf5, 0x3d, 0xf5, 0x3c, 0x75, 0x3b, 0x1, 0x75, 
	0x3a, 0x18, 0xfb, 0x7f, 0x20, 0x2, 0x11, 0x3a, 0x8f, 0x25, 0x12, 0x1f, 0xc1, 0xe4, 0xfd, 0x12, 
	0xa, 0x6f, 0xaf, 0x25, 0x12, 0x1e, 0xa0, 0x50, 0x5, 0xaf, 0x25, 0x12, 0x1f, 0x61, 0x7d, 0xff, 
	0xaf, 0x25, 0x2, 0x1d, 0xa4, 0xbf, 0x6, 0x3, 0x7f, 0x12, 0x22, 0x7d, 0x1c, 0x12, 0x13, 0x27, 
	0xef, 0x13, 0x13, 0x13, 0x54, 0x3, 0xfe, 0xef, 0x30, 0xe5, 0x3, 0x43, 0x6, 0x10, 0xaf, 0x6, 
	0x22, 0x8f, 0x2b, 0x12, 0x1e, 0x76, 0x7d, 0x1, 0xaf, 0x2b, 0x12, 0x1e, 0xca, 0xaf, 0x2b, 0x12, 
	0xa, 0x68, 0x12, 0x1a, 0xb9, 0xe4, 0xfd, 0xaf, 0x2b, 0x2, 0x1e, 0xca, 0xe4, 0xf5, 0x22, 0xaf, 
	0x22, 0x12, 0x1e, 0xa0, 0x50, 0x5, 0xaf, 0x22, 0x12, 0xd, 0x24, 0x5, 0x22, 0xe5, 0x22, 0xc3, 
	0x94, 0x7, 0x40, 0xeb, 0xc2, 0x9, 0x22, 0x8f, 0x2f, 0xe5, 0x2f, 0xc4, 0x54, 0xf, 0xff, 0x12, 
	0x1e, 0x67, 0x12, 0x1d, 0x92, 0xe5, 0x2f, 0x54, 0xf, 0xff, 0x12, 0x1e, 0x67, 0x2, 0x1d, 0x92, 
	0x8f, 0x28, 0x8e, 0x27, 0x8d, 0x26, 0x8c, 0x25, 0x78, 0x10, 0x12, 0x4, 0xe0, 0x12, 0x1e, 0x1c, 
	0xe5, 0x28, 0xff, 0xe5, 0x27, 0xfe, 0x2, 0x1e, 0x1c, 0xe4, 0xf5, 0x22, 0xaf, 0x22, 0x12, 0x1f, 
	0x10, 0x74, 0xa0, 0x25, 0x22, 0xf8, 0xa6, 0x7, 0x5, 0x22, 0xe5, 0x22, 0xc3, 0x94, 0x7, 0x40, 
	0xeb, 0x22, 0xd3, 0xef, 0x64, 0x80, 0xf8, 0x74, 0x61, 0x64, 0x80, 0x98, 0x50, 0xc, 0xef, 0x64, 
	0x80, 0x94, 0xfa, 0x50, 0x5, 0xef, 0x24, 0xe0, 0xff, 0x22, 0x22, 0x8f, 0x25, 0x12, 0x1b, 0xb5, 
	0x8f, 0x26, 0x7d, 0x5, 0xaf, 0x25, 0x12, 0x13, 0x27, 0xee, 0x30, 0xe2, 0x3, 0x43, 0x26, 0x20, 
	0xaf, 0x26, 0x22, 0x75, 0xcb, 0xff, 0x75, 0xca, 0x2, 0x75, 0xcd, 0xff, 0x75, 0xcc, 0x2, 0x75, 
	0xc8, 0x34, 0x75, 0x98, 0x52, 0xd2, 0xbc, 0xd2, 0xac, 0x22, 0x12, 0x1a, 0x98, 0xef, 0x60, 0x5, 
	0x7f, 0x5, 0x12, 0x1f, 0xe5, 0x12, 0x1c, 0xdf, 0xef, 0x60, 0x5, 0x7f, 0x6, 0x12, 0x1f, 0xe5, 
	0x22, 0x8f, 0x37, 0x12, 0x1f, 0x7c, 0x75, 0x3d, 0x3, 0x12, 0x1a, 0x65, 0x7d, 0xc, 0xaf, 0x37, 
	0x12, 0x19, 0x52, 0xaf, 0x37, 0x2, 0x1f, 0x85, 0xe4, 0xff, 0x12, 0x1d, 0xc8, 0x7f, 0x5, 0x12, 
	0x1a, 0x32, 0x12, 0x1b, 0x1c, 0x8f, 0x28, 0x7f, 0x1, 0x12, 0x1d, 0xc8, 0xaf, 0x28, 0x22, 0xe4, 
	0xf5, 0x23, 0xf5, 0x22, 0xaf, 0x22, 0x12, 0x1e, 0xa0, 0x5, 0x22, 0xe5, 0x22, 0xc3, 0x94, 0x7, 
	0x40, 0xf2, 0xaf, 0x23, 0x22, 0xe4, 0xf5, 0xff, 0xf5, 0xfe, 0xf5, 0xfd, 0x75, 0xfc, 0x1, 0x75, 
	0xf9, 0xe0, 0x75, 0xfb, 0x14, 0xd2, 0xf8, 0x80, 0xfe, 0xe5, 0x8, 0xfd, 0xe5, 0x9, 0xfb, 0x7f, 
	0x80, 0x12, 0xa, 0x16, 0x5, 0x9, 0xe5, 0x9, 0x70, 0x2, 0x5, 0x8, 0x22, 0xe4, 0xfb, 0x7d, 
	0x4, 0x7f, 0x80, 0x12, 0xa, 0x2a, 0x7f, 00, 0x7e, 00, 0x7d, 00, 0x7c, 00, 0x2, 0xa, 
	0x31, 0x8f, 0x32, 0xaf, 0x32, 0x15, 0x32, 0xef, 0xd3, 0x94, 00, 0x40, 0x7, 0x7f, 0x20, 0x12, 
	0x1d, 0x92, 0x80, 0xef, 0x22, 0x8f, 0x51, 0x7b, 0x35, 0xe4, 0xfd, 0x7f, 0xe0, 0x12, 0xa, 0x2a, 
	0xaf, 0x51, 0xe4, 0xfc, 0xfd, 0xfe, 0x2, 0xa, 0x31, 0x53, 0x89, 0xf, 0x43, 0x89, 0x10, 0x75, 
	0x8b, 0x92, 0x75, 0x8d, 0xe6, 0xc2, 0xbb, 0xd2, 0xab, 0xd2, 0x8e, 0x22, 0xc2, 0x3, 0xae, 0x7, 
	0x1f, 0xee, 0xd3, 0x94, 00, 0x40, 0x7, 0x30, 0x3, 0xfd, 0xc2, 0x3, 0x80, 0xf0, 0x22, 0x7b, 
	0x14, 0xe4, 0xfd, 0x7f, 0xe0, 0x12, 0xa, 0x2a, 0x7f, 0x1, 0x12, 0x1f, 0xba, 0x12, 0xa, 0x31, 
	0x80, 0xfe, 0xc2, 0x6, 0x8f, 0x99, 0x7f, 0x2, 0x12, 0x1f, 0xa5, 0x20, 0x6, 0x5, 0x12, 0x1f, 
	0xdd, 0x50, 0xf8, 0x22, 0x8f, 0x26, 0x8d, 0x27, 0xe5, 0x27, 0x7c, 00, 0x12, 0x1d, 0xb6, 0xad, 
	0x27, 0xaf, 0x26, 0x2, 0x17, 00, 0x8f, 0x28, 0x8c, 0x29, 0x8d, 0x2a, 0x12, 0x1b, 0xd1, 0xad, 
	0x2a, 0xac, 0x29, 0xaf, 0x28, 0x2, 0x8, 0x5c, 0xef, 0x60, 0x6, 0xe5, 0x51, 0x44, 0x2, 0x80, 
	0x4, 0xe5, 0x51, 0x54, 0xfd, 0xff, 0x12, 0x1d, 0x45, 0x22, 0xef, 0x60, 0x6, 0xe5, 0x51, 0x44, 
	0x1, 0x80, 0x4, 0xe5, 0x51, 0x54, 0xfe, 0xff, 0x12, 0x1d, 0x45, 0x22, 0x7b, 0x35, 0xe4, 0xfd, 
	0x7f, 0xe0, 0x12, 0xa, 0x16, 0xef, 0x54, 0x10, 0xc4, 0x54, 0xf, 0xff, 0x22, 0x5, 0xf, 0xe5, 
	0xf, 0xc3, 0x94, 0x96, 0x40, 0x6, 0xe4, 0xf5, 0xf, 0x12, 0x1f, 0xac, 0x22, 0x74, 0x80, 0x25, 
	0x1f, 0xf8, 0xe6, 0xff, 0xe5, 0x1f, 0x4, 0x54, 0x1f, 0xf5, 0x1f, 0x22, 0x8e, 0x29, 0x8f, 0x2a, 
	0xe5, 0x29, 0xff, 0x12, 0x1c, 0x7, 0xaf, 0x2a, 0x2, 0x1c, 0x7, 0x5, 0x4d, 0xe5, 0x4d, 0xc3, 
	0x94, 0xa, 0x40, 0x5, 0xe4, 0xf5, 0x4d, 0xd2, 0x9, 0x22, 0x75, 0x3d, 0x12, 0x75, 0x3e, 00, 
	0x7b, 00, 0x7a, 0x12, 0xe4, 0xfd, 0x2, 0x19, 0x52, 0x4c, 0x75, 0x74, 0x6f, 0x6e, 0x20, 0x32, 
	0x2e, 0x32, 0x39, 00, 00, 00, 00, 00, 0xe4, 0xff, 0x12, 0x1d, 0xc8, 0x7f, 0x6, 0x12, 
	0x1a, 0x32, 0x7f, 0x1, 0x2, 0x1d, 0xc8, 0xef, 0xc3, 0x94, 0xa, 0xef, 0x50, 0x4, 0x24, 0x30, 
	0xff, 0x22, 0x24, 0x57, 0xff, 0x22, 0xad, 0x7, 0x75, 0x3c, 0x10, 0xe4, 0xf5, 0x3d, 0xfb, 0x7f, 
	0x20, 0x2, 0x17, 0x9d, 0x7d, 0x25, 0x12, 0x18, 0xb5, 0x7b, 00, 0x7a, 00, 0x79, 0x25, 0x2, 
	0x17, 0x69, 0x45, 0x52, 0x41, 0x53, 0x45, 00, 0x50, 0x52, 0x4f, 0x47, 0x52, 0x41, 0x4d, 00, 
	0xef, 0x90, 0x1e, 0xb5, 0x93, 0xd3, 0x94, 0x1, 0x50, 0x2, 0x80, 0x1, 0xc3, 0x22, 00, 0x1, 
	0x2, 0x3, 0x4, 0xff, 0xff, 00, 00, 00, 00, 00, 0xff, 0xff, 0x7d, 0x1, 0x12, 0x13, 
	0x27, 0xef, 0x30, 0xe2, 0x3, 0xd3, 0x80, 0x1, 0xc3, 0x22, 0x8f, 0x3c, 0x8d, 0x3d, 0x7b, 0xe, 
	0xe4, 0xfd, 0x7f, 0xa0, 0x2, 0x17, 0x9d, 0x8f, 0x3c, 0x8d, 0x3d, 0x7b, 0x10, 0xe4, 0xfd, 0x7f, 
	0x40, 0x2, 0x17, 0x9d, 0xe5, 0x1e, 0x65, 0x1f, 0x60, 0x3, 0xd3, 0x80, 0x1, 0xc3, 0x22, 0xe5, 
	0xaa, 0x54, 0x3, 0x60, 0x3, 0xd3, 0x80, 0x1, 0xc3, 0x22, 0x7f, 0x3, 0x7e, 00, 0x7d, 00, 
	0x7c, 00, 0x2, 0x18, 0x8a, 0xa8, 0x1a, 0xe6, 0xb4, 0x20, 0x4, 0x5, 0x1a, 0x80, 0xf6, 0x22, 
	0x12, 0x1e, 0xa0, 0x50, 0x3, 0x7f, 0x1, 0x22, 0x7f, 00, 0x22, 0xc2, 0x2, 0x8f, 0xe, 0xd2, 
	0x2, 0x20, 0x2, 0xfd, 0x22, 0xeb, 0x75, 0xf0, 0x20, 0xa4, 0x4d, 0xac, 0xf0, 0xfd, 0x22, 0x7f, 
	0x30, 0x12, 0x1d, 0x92, 0x7f, 0x78, 0x2, 0x1d, 0x92, 0x7f, 0xd, 0x12, 0x1d, 0x92, 0x7f, 0xa, 
	0x2, 0x1d, 0x92, 0x75, 0xf0, 0xa, 0x84, 0xe5, 0xf0, 0x24, 0x30, 0xff, 0x22, 0x8f, 0x35, 0x12, 
	0x18, 0x2f, 0xaf, 0x35, 0x2, 0xa, 0xe0, 0xd3, 0xe5, 0x26, 0x94, 00, 0xe5, 0x25, 0x94, 00, 
	0x22, 0xef, 0x64, 0x6, 0x60, 0x3, 0x12, 0x1c, 0xb1, 0x22, 0x7b, 0x1, 0x7a, 00, 0x7d, 0x1f, 
	0x2, 0x17, 0xff, 0x7b, 0xb5, 0x7a, 0x52, 0x7d, 0x1f, 0x2, 0x17, 0xff, 0x7b, 0x30, 0x7a, 0x2a, 
	0x7d, 0x1f, 0x2, 0x17, 0xff, 0xe4, 0xfb, 0xfa, 0x7d, 0x1f, 0x2, 0x17, 0xff, 0x7d, 0x32, 0x12, 
	0x18, 0xb5, 0xaf, 0x37, 0x22, 0xe5, 0x51, 0x44, 0x4, 0xff, 0x2, 0x1d, 0x45, 0xe5, 0x51, 0x54, 
	0xfb, 0xff, 0x2, 0x1d, 0x45, 0xc2, 0x2, 0x8f, 0xe, 0xd2, 0x2, 0x22, 0x7f, 0x3, 0x7e, 00, 
	0x2, 0x16, 0xcb, 0x7f, 0x5, 0x7e, 00, 0x2, 0x16, 0xcb, 0x7e, 00, 0x7d, 00, 0x7c, 00, 
	0x22, 0x7b, 00, 0x7a, 00, 0x79, 0x4e, 0x22, 0xef, 0x90, 0x1e, 0xb5, 0x93, 0xff, 0x22, 0xef, 
	0x90, 0x1e, 0xae, 0x93, 0xff, 0x22, 00, 0x1, 0x2, 0x3, 0x4, 0x6, 0x5, 0xa2, 0x2, 0xb3, 
	0x22, 0x12, 0x12, 0x7, 0x22, 0xd2, 0x7, 0x22, 0x2, 0x1d, 0x92, 0x2, 0x15, 0xe4, 0x22, 0x22, 
};



/*---------------------------------------------------------------------*/
/* Prototypes of External Functions                                    */
/*---------------------------------------------------------------------*/



/*---------------------------------------------------------------------*/
/* Prototypes of Functions Used                                        */
/*---------------------------------------------------------------------*/
static inline void spi_chip_select(u8 enable);
static void spi_write(u8 data);								
static u8 spi_read(void);

static void spi_master_init(void);
static u8 spi_eeprom_read(u16 address, u16 nbytes, u8 *dest);
static u8 spi_eeprom_write(u16 address, u16 nbytes, u8 *src);

int rt2880_spi_toolkit(cmd_tbl_t *cmdtp, int flag, int argc, char *argv[]);

/*---------------------------------------------------------------------*/
/* External Variable Definitions                                       */
/*---------------------------------------------------------------------*/




/*----------------------------------------------------------------------*/
/*   Function                                                           */
/*           	spi_chip_select                                         */
/*    INPUTS: ENABLE or DISABLE                                         */
/*   RETURNS: None                                                      */
/*   OUTPUTS: None                                                      */
/*   NOTE(S): Pull on or Pull down #CS                                  */
/*                                                                      */
/*----------------------------------------------------------------------*/

static inline void spi_chip_select(u8 enable)
{
	int i;

	for (i=0; i<spi_busy_loop; i++) {
		if (!IS_BUSY) {
			/* low active */
			if (enable) {		
				RT2880_REG(RT2880_SPICTL_REG) =	SPICTL_SPIENA_LOW;
			} else  {
				RT2880_REG(RT2880_SPICTL_REG) = SPICTL_SPIENA_HIGH;
			}		
			break;
		}
	}

#ifdef DBG
	if (i == spi_busy_loop) {
		printf("warning : spi_transfer (spi_chip_select) busy !\n");
	}
#endif
}

/*----------------------------------------------------------------------*/
/*   Function                                                           */
/*           	spi_master_init                                         */
/*    INPUTS: None                                                      */
/*   RETURNS: None                                                      */
/*   OUTPUTS: None                                                      */
/*   NOTE(S): Initialize SPI block to desired state                     */
/*                                                                      */
/*----------------------------------------------------------------------*/
static void spi_master_init(void)
{
	/* try to reset again from Reset Control Register */
	RT2880_REG(RT2880_RSTCTRL_REG) = RSTCTRL_SPI_RESET;
	RT2880_REG(RT2880_RSTCTRL_REG) = 0;
	udelay(500);

	RT2880_REG(RT2880_SPICFG_REG) = SPICFG_MSBFIRST | 
									SPICFG_RXCLKEDGE_FALLING |
									SPICFG_TXCLKEDGE_FALLING |
									SPICFG_SPICLK_DIV8;
	spi_chip_select(DISABLE);

#ifdef DBG
/*        printf("SPICFG = %08x\n", RT2880_REG(RT2880_SPICFG_REG));*/
/*        printf("is busy %d\n", IS_BUSY);*/
	if (IS_BUSY) printf("spi_master_init: is busy\n");
#endif		 	
}



/*----------------------------------------------------------------------*/
/*   Function                                                           */
/*           	spi_write                                               */
/*    INPUTS: 8-bit data                                                */
/*   RETURNS: None                                                      */
/*   OUTPUTS: None                                                      */
/*   NOTE(S): transfer 8-bit data to SPI                                */
/*                                                                      */
/*----------------------------------------------------------------------*/
static void spi_write(u8 data)
{
	int i;

	for (i=0; i<spi_busy_loop; i++) {
		if (!IS_BUSY) {
			RT2880_REG(RT2880_SPIDATA_REG) = data;
			/* start write transfer */
			RT2880_REG(RT2880_SPICTL_REG) = SPICTL_HIZSDO |  SPICTL_STARTWR | 
											SPICTL_SPIENA_LOW;
			break;
		}
	}

#ifdef DBG
	if (i == spi_busy_loop) {
		printf("warning : spi_transfer (write %02x) busy !\n", data);
	}
#endif
}

//write32 MSB first
static void spi_write32(u32 data)
{
	u8 d0, d1, d2, d3;

	d0 = (u8)((data >> 24) & 0xff);
	d1 = (u8)((data >> 16) & 0xff);
	d2 = (u8)((data >> 8) & 0xff);
	d3 = (u8)(data & 0xff);

	spi_write(d0);
	spi_write(d1);
	spi_write(d2);
	spi_write(d3);
}

/*----------------------------------------------------------------------*/
/*   Function                                                           */
/*           	spi_read                                                */
/*    INPUTS: None                                                      */
/*   RETURNS: 8-bit data                                                */
/*   OUTPUTS: None                                                      */
/*   NOTE(S): get 8-bit data from SPI                                   */
/*                                                                      */
/*----------------------------------------------------------------------*/
static u8 spi_read(void) 
{
	int i;

	/* 
	 * poll busy bit until it is 0 
	 * then start read transfer
	 */
	for (i=0; i<spi_busy_loop; i++) {
		if (!IS_BUSY) {
			RT2880_REG(RT2880_SPIDATA_REG) = 0;
			/* start read transfer */
			RT2880_REG(RT2880_SPICTL_REG) = SPICTL_HIZSDO | SPICTL_STARTRD |
											SPICTL_SPIENA_LOW;
			break;
		}
	}

	/* 
	 * poll busy bit until it is 0 
	 * then get data 
	 */
	for (i=0; i<spi_busy_loop; i++) {
		if (!IS_BUSY) {
			break;
		}
	}
	
#ifdef DBG		
	if (i == spi_busy_loop) {
		printf("warning : spi_transfer busy !\n");
	} 
#endif

	return ((u8)RT2880_REG(RT2880_SPIDATA_REG));
}

//read32 MSB first
static u32 spi_read32(void)
{
	u8 d0, d1, d2, d3;
	u32 ret;

	d0 = spi_read();
	d1 = spi_read();
	d2 = spi_read();
	d3 = spi_read();
	ret = (d0 << 24) | (d1 << 16) | (d2 << 8) | d3;

	return ret;
}

/*----------------------------------------------------------------------*/
/*   Function                                                           */
/*           	eeprom_get_status_reg                                   */
/*    INPUTS: pointer to status                                         */
/*   RETURNS: None                                                      */
/*   OUTPUTS: status                                                    */
/*   NOTE(S): get the status of eeprom (AT25xxxx)                       */
/*                                                                      */
/*----------------------------------------------------------------------*/
static void eeprom_get_status_reg(u8 *status) 
{
	spi_chip_select(ENABLE);
	spi_write(RDSR_CMD);
	*status = spi_read();		
	spi_chip_select(DISABLE);
}


/*----------------------------------------------------------------------*/
/*   Function                                                           */
/*           	spi_eeprom_read                                         */
/*    INPUTS: address - start address to be read                        */ 
/*            nbytes  - number of bytes to be read                      */
/*            dest    - pointer to read buffer                          */
/*   RETURNS: 0 - successful                                            */
/*            or eeprom status                                          */
/*   OUTPUTS: read buffer                                               */
/*   NOTE(S): If the eeprom is busy , the function returns with status  */
/*            register of eeprom                                        */
/*----------------------------------------------------------------------*/
static u8 spi_eeprom_read(u16 address, u16 nbytes, u8 *dest)
{
	u8	status;
	u16	cnt = 0;
	int i = 0;

	do {
		i++;
		eeprom_get_status_reg(&status);		
	}
	while((status & (1<<RDY)) && (i < max_ee_busy_loop));

	if (i == max_ee_busy_loop)
		return (status);

	/* eeprom ready */
	if (!(status & (1<<RDY))) {	
		spi_chip_select(ENABLE);
		/* read op */
		spi_write(READ_CMD);
		spi_write((u8)(address >> 8));		/* MSB byte First */
		spi_write((u8)(address & 0x00FF));	/* LSB byte */

		while (cnt < nbytes) {
			*(dest++) = spi_read();
			cnt++;
		}
		status = 0;
		/* deassert cs */
		spi_chip_select(DISABLE);
	} 
	return (status);	
}

/*----------------------------------------------------------------------*/
/*   Function                                                           */
/*           	eeprom_write_enable                                     */
/*    INPUTS: None                                                      */
/*   RETURNS: None                                                      */
/*   OUTPUTS: None                                                      */
/*   NOTE(S): always perform write enable  before any write operation   */
/*                                                                      */
/*----------------------------------------------------------------------*/
static void eeprom_write_enable(void)
{
	u8	status;
	int i = 0;

	do {
		i++;
		eeprom_get_status_reg(&status);		
	}
	while((status & (1<<RDY)) && (i < max_ee_busy_loop));

	if (i == max_ee_busy_loop)
		return;

	/* eeprom ready */
	if (!(status & (1<<RDY))) 
	{	
		spi_chip_select(ENABLE);
		/* always write enable  before any write operation */
		spi_write(WREN_CMD);

		spi_chip_select(DISABLE);
		
		/* wait for write enable */
		do {
			eeprom_get_status_reg(&status);
		} while((status & (1<<RDY)) || !(status & (1<<WEN)));

	}

}

/*----------------------------------------------------------------------*/
/*   Function                                                           */
/*           	spi_eeprom_write                                        */
/*    INPUTS: address - the first byte address to be written            */
/*            nbytes  - the number of bytes to be written               */
/*            src     - the pointer to source buffer                    */     
/*   RETURNS: 0  - successful                                           */
/*            or eeprom buy status                                      */
/*   OUTPUTS: None                                                      */
/*   NOTE(S): The different eeprom has various write page size          */
/*            The function don't care write page size so the caller     */
/*            must check the page size of eeprom                        */
/*                                                                      */
/*----------------------------------------------------------------------*/
u8 spi_eeprom_write(u16 address, u16 nbytes, u8 *src)
{
	u8	status;
	u16	cnt = 0;
	int i = 0;

	do {
		i++;
		eeprom_get_status_reg(&status);		
	}
	while((status & (1<<RDY)) && (i < max_ee_busy_loop));

	if (i == max_ee_busy_loop)
		goto done;


	/* eeprom ready */
	if (!(status & (1<<RDY))) {			
		/* always write enable  before any write operation */		
		eeprom_write_enable();

		spi_chip_select(ENABLE);
		spi_write(WRITE_CMD);
		spi_write((u8)(address >> 8));		/* MSB byte First */
		spi_write((u8)(address & 0x00FF));	/* LSB byte */

		while (cnt < nbytes) {
			spi_write(src[cnt]);
			cnt++;
		}
		status = 0;
		/* last byte sent then pull #cs high  */
		spi_chip_select(DISABLE);
	} 

	i = 0;
	do {
		i++;
		eeprom_get_status_reg(&status);		
	}
	while((status & (1<<RDY)) && (i < max_ee_busy_loop));

done:
	return (status);
}

void spi_eeprom_dump(void)
{
	printf("spi_eeprom_dump()... called!\n");
}

void spi_eeprom_cmd(cmd_tbl_t *cmdtp, int argc, char *argv[])
{
	int chk_match, size;
	ulong addr, value;
	u16 address;

	/* We use the last specified parameters, unless new ones are entered */
	switch (argc) {
		case RT2880_SPI_DUMP:
        		chk_match = strcmp(argv[1], RT2880_SPI_DUMP_STR);
			if ( chk_match != 0) {
				printf("Usage:\n%s\n", cmdtp->usage);
				return;
			}
			spi_eeprom_dump();
        		break;
		case RT2880_SPI_READ:
        		chk_match = strcmp(argv[1], RT2880_SPI_READ_STR);
        		if ( chk_match != 0) {
				printf("Usage:\n%s\n", cmdtp->usage);
                		return;
        		}
        		addr = simple_strtoul(argv[2], NULL, 16);
			address = addr;
			spi_master_init();
			spi_eeprom_read(addr, 4, (u8*)&value);
        		printf("0x%04x : 0x%04x\n", address, value);
        		break;
		case RT2880_SPI_WRITE:
			chk_match = strcmp(argv[1], RT2880_SPI_WRITE_STR);
			if ( chk_match != 0) {
				printf("Usage:\n%s\n", cmdtp->usage);
				return;
			}
			size = simple_strtoul(argv[2], NULL, 16);
			addr = simple_strtoul(argv[3], NULL, 16);
			value = simple_strtoul(argv[4], NULL, 16);
			spi_master_init();
			address = addr;
			spi_eeprom_write(address, size, (u8*)&value);
			printf("0x%08x: 0x%08x in %d bytes\n", address, value, size);
			break;
		default:
			printf("Usage:\n%s\n use \"help spicmd\" to get more detail!\n", cmdtp->usage);
	}
}

#if 0
void spi_vtss_read(u8 blk, u8 subblk, u8 addr, u32 *value)
{
	u8 cmd;

	spi_master_init();
	cmd = (u8)((blk << 5) | subblk);
	spi_write(cmd);
	spi_write(addr);
	spi_read(); //dummy byte
	spi_read(); //dummy byte
	*value = spi_read32();
	//printf("rd %x:%x:%x = %x\n", blk, subblk, addr, *value);
	udelay(100);
}
#endif

void spi_vtss_write(u8 blk, u8 subblk, u8 addr, u32 value)
{
	u8 cmd;

	spi_master_init();
	cmd = (u8)((blk << 5) | subblk | 0x10);
	spi_write(cmd);
	spi_write(addr);
	spi_write32(value);
	//printf("wr %x:%x:%x = %x\n", blk, subblk, addr, value);
	udelay(10);
}

void vtss_init(void)
{
	int i, len;
	len = sizeof(lutonu);

	//HT_WR(SYSTEM, 0, ICPU_CTRL, (1<<7) | (1<<3) | (1<<2) | (0<<0));
	spi_vtss_write(7, 0, 0x10, (1<<7) | (1<<3) | (1<<2) | (0<<0));

	//HT_WR(SYSTEM, 0, ICPU_ADDR, 0); //clear SP_SELECT and ADDR

	for (i = 0; i < len; i++) {
		//HT_WR(SYSTEM, 0, ICPU_DATA, lutonu[i]);
		spi_vtss_write(7, 0, 0x12, lutonu[i]);
	}

	//HT_WR(SYSTEM, 0, GLORESET, (1<<0)); //MASTER_RESET
	spi_vtss_write(7, 0, 0x14, (1<<0));
	udelay(125000);

	//HT_WR(SYSTEM, 0, ICPU_CTRL, (1<<8) | (1<<3) | (1<<1) | (1<<0));
	spi_vtss_write(7, 0, 0x10, (1<<8) | (1<<3) | (1<<1) | (1<<0));
	printf(" Vitesse uploading binary codes (%d bytes) done.\n", len);
}

void spi_vtss_cmd(cmd_tbl_t *cmdtp, int argc, char *argv[])
{
	u8 blk, subblk, cmd, addr;
	u32 value;

	//insufficient arguments
	if (argc < 5)
		printf("Usage:\n%s\n use \"help spicmd\" for detail!\n", cmdtp->usage);

	blk = (u8)simple_strtoul(argv[2], NULL, 16);
	subblk = (u8)simple_strtoul(argv[3], NULL, 16);
	addr = (u8)simple_strtoul(argv[4], NULL, 16);

	//block, sub-bluck out of range
	if (blk == 0 || blk > 7 || subblk > 7)
		printf("Usage:\n%s\n use \"help spicmd\" for detail!\n", cmdtp->usage);

	cmd = (u8)((blk << 5) | subblk);

	if (!strncmp(argv[1], "read", 5)) {
		spi_master_init();
		spi_write(cmd);
		spi_write(addr);
		spi_read(); //dummy byte
		spi_read(); //dummy byte
		value = spi_read32();
		printf("read %d:%d:%x = 0x%x\n", blk, subblk, addr, value);
	}
	else if (!strncmp(argv[1], "write", 6)) {
		cmd |= 0x10; //set the write bit
		value = (u32)simple_strtoul(argv[5], NULL, 16);
		spi_master_init();
		spi_write(cmd);
		spi_write(addr);
		spi_write32(value);
		printf("write %d:%d:%x = 0x%x\n", blk, subblk, addr, value);
	}
	else
		printf("Usage:\n%s\n use \"help spicmd\" for detail!\n", cmdtp->usage);
}

int rt2880_spi_toolkit(cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{
	if (!strncmp(argv[1], "eeprom", 7)) {
		spi_eeprom_cmd(cmdtp, argc - 1, argv + 1);
	}
	else if (!strncmp(argv[1], "vtss", 5)) {
		spi_vtss_cmd(cmdtp, argc - 1, argv + 1);
	}
	else
		printf("Usage:\n%s\n use \"help spicmd\" for detail!\n", cmdtp->usage);
	return 0;
}

U_BOOT_CMD(
	spicmd,	7,	1, 	rt2880_spi_toolkit,
	"spicmd	- read/write data from/to eeprom or vtss\n",
	"spicmd usage:\n"
	"  spicmd eeprom read [address]\n"
	"  spicmd eeprom write [size] [address] [value]\n"
	"  spicmd eeprom dump\n"
	"  spicmd vtss read [block] [sub-block] [address]\n"
	"  spicmd vtss write [block] [sub-block] [address] [value]\n"
	"    NOTE: size is 1, 2, 4 bytes only, address and value are in hex\n"
);

#endif
